@@ -1,5 +1,6 @@
 /* GStreamer
  * Copyright (C) 2005 Wim Taymans <wim@fluendo.com>
+ * Copyright (C) 2006 Tim-Philipp Müller <tim centricular net>
  *
  * gstalsasink.c:
  *
@@ -19,6 +20,27 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/**
+ * SECTION:element-alsasink
+ * @short_description: play audio to an ALSA device
+ * @see_also: alsasrc, alsamixer
+ *
+ * <refsect2>
+ * <para>
+ * This element renders raw audio samples using the ALSA api.
+ * </para>
+ * <title>Example pipelines</title>
+ * <para>
+ * Play an Ogg/Vorbis file.
+ * </para>
+ * <programlisting>
+ * gst-launch -v filesrc location=sine.ogg ! oggdemux ! vorbisdec ! audioconvert ! audioresample ! alsasink
+ * </programlisting>
+ * </refsect2>
+ *
+ * Last reviewed on 2006-03-01 (0.10.4)
+ */
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -32,14 +54,20 @@
 
 #include "gstalsa.h"
 #include "gstalsasink.h"
+#include "gstalsadeviceprobe.h"
+
+#include <gst/gst-i18n-plugin.h>
 
 /* elementfactory information */
-static GstElementDetails gst_alsasink_details =
-GST_ELEMENT_DETAILS ("Audio Sink (ALSA)",
+static const GstElementDetails gst_alsasink_details =
+GST_ELEMENT_DETAILS ("Audio sink (ALSA)",
     "Sink/Audio",
     "Output to a sound card via ALSA",
     "Wim Taymans <wim@fluendo.com>");
 
+#define DEFAULT_DEVICE		"default"
+#define DEFAULT_DEVICE_NAME	""
+
 enum
 {
   PROP_0,
@@ -47,10 +75,13 @@ enum
   PROP_DEVICE_NAME
 };
 
-static void gst_alsasink_base_init (gpointer g_class);
-static void gst_alsasink_class_init (GstAlsaSinkClass * klass);
-static void gst_alsasink_init (GstAlsaSink * alsasink);
+static void gst_alsasink_init_interfaces (GType type);
+
+GST_BOILERPLATE_FULL (GstAlsaSink, gst_alsasink, GstAudioSink,
+    GST_TYPE_AUDIO_SINK, gst_alsasink_init_interfaces);
+
 static void gst_alsasink_dispose (GObject * object);
+static void gst_alsasink_finalise (GObject * object);
 static void gst_alsasink_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec);
 static void gst_alsasink_get_property (GObject * object,
@@ -68,69 +99,71 @@ static guint gst_alsasink_write (GstAudioSink * asink, gpointer data,
 static guint gst_alsasink_delay (GstAudioSink * asink);
 static void gst_alsasink_reset (GstAudioSink * asink);
 
-/* AlsaSink signals and args */
-enum
-{
-  LAST_SIGNAL
-};
+static gint output_ref;         /* 0    */
+static snd_output_t *output;    /* NULL */
+static GStaticMutex output_mutex = G_STATIC_MUTEX_INIT;
+
+
+#if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
+# define ALSA_SINK_FACTORY_ENDIANNESS	"LITTLE_ENDIAN, BIG_ENDIAN"
+#else
+# define ALSA_SINK_FACTORY_ENDIANNESS	"BIG_ENDIAN, LITTLE_ENDIAN"
+#endif
 
 static GstStaticPadTemplate alsasink_sink_factory =
     GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("audio/x-raw-int, "
-#if (G_BYTE_ORDER == G_LITTLE_ENDIAN)
-        "endianness = (int) { LITTLE_ENDIAN, BIG_ENDIAN }, "
-#else
-        "endianness = (int) { BIG_ENDIAN, LITTLE_ENDIAN }, "
-#endif
+        "endianness = (int) { " ALSA_SINK_FACTORY_ENDIANNESS " }, "
+        "signed = (boolean) { TRUE, FALSE }, "
+        "width = (int) 32, "
+        "depth = (int) 32, "
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]; "
+        "audio/x-raw-int, "
+        "endianness = (int) { " ALSA_SINK_FACTORY_ENDIANNESS " }, "
         "signed = (boolean) { TRUE, FALSE }, "
         "width = (int) 16, "
         "depth = (int) 16, "
-        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, 2 ]; "
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]; "
         "audio/x-raw-int, "
         "signed = (boolean) { TRUE, FALSE }, "
         "width = (int) 8, "
         "depth = (int) 8, "
-        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, 2 ]")
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]")
     );
 
-static GstElementClass *parent_class = NULL;
-
-/* static guint gst_alsasink_signals[LAST_SIGNAL] = { 0 }; */
-
-GType
-gst_alsasink_get_type (void)
-{
-  static GType alsasink_type = 0;
-
-  if (!alsasink_type) {
-    static const GTypeInfo alsasink_info = {
-      sizeof (GstAlsaSinkClass),
-      gst_alsasink_base_init,
-      NULL,
-      (GClassInitFunc) gst_alsasink_class_init,
-      NULL,
-      NULL,
-      sizeof (GstAlsaSink),
-      0,
-      (GInstanceInitFunc) gst_alsasink_init,
-    };
-
-    alsasink_type =
-        g_type_register_static (GST_TYPE_AUDIO_SINK, "GstAlsaSink",
-        &alsasink_info, 0);
-  }
-
-  return alsasink_type;
-}
-
 static void
 gst_alsasink_dispose (GObject * object)
 {
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
+static void
+gst_alsasink_finalise (GObject * object)
+{
+  GstAlsaSink *sink = GST_ALSA_SINK (object);
+
+  g_free (sink->device);
+  g_mutex_free (sink->alsa_lock);
+
+  g_static_mutex_lock (&output_mutex);
+  --output_ref;
+  if (output_ref == 0) {
+    snd_output_close (output);
+    output = NULL;
+  }
+  g_static_mutex_unlock (&output_mutex);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
+}
+
+static void
+gst_alsasink_init_interfaces (GType type)
+{
+  gst_alsa_type_add_device_property_probe_interface (type);
+}
+
 static void
 gst_alsasink_base_init (gpointer g_class)
 {
@@ -156,9 +189,10 @@ gst_alsasink_class_init (GstAlsaSinkClass * klass)
   gstbaseaudiosink_class = (GstBaseAudioSinkClass *) klass;
   gstaudiosink_class = (GstAudioSinkClass *) klass;
 
-  parent_class = g_type_class_ref (GST_TYPE_BASE_AUDIO_SINK);
+  parent_class = g_type_class_peek_parent (klass);
 
   gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_alsasink_dispose);
+  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_alsasink_finalise);
   gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_alsasink_get_property);
   gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_alsasink_set_property);
 
@@ -175,11 +209,12 @@ gst_alsasink_class_init (GstAlsaSinkClass * klass)
   g_object_class_install_property (gobject_class, PROP_DEVICE,
       g_param_spec_string ("device", "Device",
           "ALSA device, as defined in an asound configuration file",
-          "default", G_PARAM_READWRITE));
+          DEFAULT_DEVICE, G_PARAM_READWRITE));
 
   g_object_class_install_property (gobject_class, PROP_DEVICE_NAME,
       g_param_spec_string ("device-name", "Device name",
-          "Human-readable name of the sound device", "", G_PARAM_READABLE));
+          "Human-readable name of the sound device", DEFAULT_DEVICE_NAME,
+          G_PARAM_READABLE));
 }
 
 static void
@@ -192,9 +227,12 @@ gst_alsasink_set_property (GObject * object, guint prop_id,
 
   switch (prop_id) {
     case PROP_DEVICE:
-      if (sink->device)
-        g_free (sink->device);
-      sink->device = g_strdup (g_value_get_string (value));
+      g_free (sink->device);
+      sink->device = g_value_dup_string (value);
+      /* setting NULL restores the default device */
+      if (sink->device == NULL) {
+        sink->device = g_strdup (DEFAULT_DEVICE);
+      }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -233,18 +271,21 @@ gst_alsasink_get_property (GObject * object, guint prop_id,
 }
 
 static void
-gst_alsasink_init (GstAlsaSink * alsasink)
+gst_alsasink_init (GstAlsaSink * alsasink, GstAlsaSinkClass * g_class)
 {
-  GST_DEBUG ("initializing alsasink");
+  GST_DEBUG_OBJECT (alsasink, "initializing alsasink");
 
-  alsasink->device = g_strdup ("default");
+  alsasink->device = g_strdup (DEFAULT_DEVICE);
   alsasink->handle = NULL;
-}
+  alsasink->cached_caps = NULL;
+  alsasink->alsa_lock = g_mutex_new ();
 
-static GstCaps *
-gst_alsasink_getcaps (GstBaseSink * bsink)
-{
-  return NULL;
+  g_static_mutex_lock (&output_mutex);
+  if (output_ref == 0) {
+    snd_output_stdio_attach (&output, stdout, 0);
+    ++output_ref;
+  }
+  g_static_mutex_unlock (&output_mutex);
 }
 
 #define CHECK(call, error) \
@@ -253,17 +294,59 @@ if ((err = call) < 0)           \
   goto error;                   \
 } G_STMT_END;
 
+static GstCaps *
+gst_alsasink_getcaps (GstBaseSink * bsink)
+{
+  GstElementClass *element_class;
+  GstPadTemplate *pad_template;
+  GstAlsaSink *sink = GST_ALSA_SINK (bsink);
+  GstCaps *caps;
+
+  if (sink->handle == NULL) {
+    GST_DEBUG_OBJECT (sink, "device not open, using template caps");
+    return NULL;                /* base class will get template caps for us */
+  }
+
+  if (sink->cached_caps) {
+    GST_LOG_OBJECT (sink, "Returning cached caps");
+    return gst_caps_ref (sink->cached_caps);
+  }
+
+  element_class = GST_ELEMENT_GET_CLASS (sink);
+  pad_template = gst_element_class_get_pad_template (element_class, "sink");
+  g_return_val_if_fail (pad_template != NULL, NULL);
+
+  caps = gst_alsa_probe_supported_formats (GST_OBJECT (sink), sink->handle,
+      gst_pad_template_get_caps (pad_template));
+
+  if (caps) {
+    sink->cached_caps = gst_caps_ref (caps);
+  }
+
+  GST_INFO_OBJECT (sink, "returning caps %" GST_PTR_FORMAT, caps);
+
+  return caps;
+}
+
 static int
 set_hwparams (GstAlsaSink * alsa)
 {
   guint rrate;
   gint err, dir;
   snd_pcm_hw_params_t *params;
+  guint period_time, buffer_time;
 
   snd_pcm_hw_params_alloca (&params);
 
-  GST_DEBUG ("Negotiating to %d channels @ %d Hz", alsa->channels, alsa->rate);
+  GST_DEBUG_OBJECT (alsa, "Negotiating to %d channels @ %d Hz (format = %s)",
+      alsa->channels, alsa->rate, snd_pcm_format_name (alsa->format));
 
+  /* start with requested values, if we cannot configure alsa for those values,
+   * we set these values to -1, which will leave the default alsa values */
+  buffer_time = alsa->buffer_time;
+  period_time = alsa->period_time;
+
+retry:
   /* choose all parameters */
   CHECK (snd_pcm_hw_params_any (alsa->handle, params), no_config);
   /* set the interleaved read/write format */
@@ -282,104 +365,135 @@ set_hwparams (GstAlsaSink * alsa)
   if (rrate != alsa->rate)
     goto rate_match;
 
-  if (alsa->buffer_time != -1) {
-    /* set the buffer time */
-    CHECK (snd_pcm_hw_params_set_buffer_time_near (alsa->handle, params,
-            &alsa->buffer_time, &dir), buffer_time);
+  /* get and dump some limits */
+  {
+    guint min, max;
+
+    snd_pcm_hw_params_get_buffer_time_min (params, &min, &dir);
+    snd_pcm_hw_params_get_buffer_time_max (params, &max, &dir);
+
+    GST_DEBUG_OBJECT (alsa, "buffer time %u, min %u, max %u",
+        alsa->buffer_time, min, max);
+
+    snd_pcm_hw_params_get_period_time_min (params, &min, &dir);
+    snd_pcm_hw_params_get_period_time_max (params, &max, &dir);
+
+    GST_DEBUG_OBJECT (alsa, "period time %u, min %u, max %u",
+        alsa->period_time, min, max);
+
+    snd_pcm_hw_params_get_periods_min (params, &min, &dir);
+    snd_pcm_hw_params_get_periods_max (params, &max, &dir);
+
+    GST_DEBUG_OBJECT (alsa, "periods min %u, max %u", min, max);
   }
-  if (alsa->period_time != -1) {
+
+  /* now try to configure the buffer time and period time, if one
+   * of those fail, we fall back to the defaults and emit a warning. */
+  if (buffer_time != -1) {
+    /* set the buffer time */
+    if ((err = snd_pcm_hw_params_set_buffer_time_near (alsa->handle, params,
+                &buffer_time, &dir)) < 0) {
+      GST_ELEMENT_WARNING (alsa, RESOURCE, SETTINGS, (NULL),
+          ("Unable to set buffer time %i for playback: %s",
+              buffer_time, snd_strerror (err)));
+      /* disable buffer_time the next round */
+      buffer_time = -1;
+      goto retry;
+    }
+  }
+  if (period_time != -1) {
     /* set the period time */
-    CHECK (snd_pcm_hw_params_set_period_time_near (alsa->handle, params,
-            &alsa->period_time, &dir), period_time);
+    if ((err = snd_pcm_hw_params_set_period_time_near (alsa->handle, params,
+                &period_time, &dir)) < 0) {
+      GST_ELEMENT_WARNING (alsa, RESOURCE, SETTINGS, (NULL),
+          ("Unable to set period time %i for playback: %s",
+              period_time, snd_strerror (err)));
+      /* disable period_time the next round */
+      period_time = -1;
+      goto retry;
+    }
   }
 
   /* write the parameters to device */
   CHECK (snd_pcm_hw_params (alsa->handle, params), set_hw_params);
 
+  /* now get the configured values */
   CHECK (snd_pcm_hw_params_get_buffer_size (params, &alsa->buffer_size),
       buffer_size);
-
   CHECK (snd_pcm_hw_params_get_period_size (params, &alsa->period_size, &dir),
       period_size);
 
+  GST_DEBUG_OBJECT (alsa, "buffer size %lu, period size %lu", alsa->buffer_size,
+      alsa->period_size);
+
   return 0;
 
   /* ERRORS */
 no_config:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
         ("Broken configuration for playback: no configurations available: %s",
-            snd_strerror (err)), (NULL));
+            snd_strerror (err)));
     return err;
   }
 wrong_access:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Access type not available for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Access type not available for playback: %s", snd_strerror (err)));
     return err;
   }
 no_sample_format:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Sample format not available for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Sample format not available for playback: %s", snd_strerror (err)));
     return err;
   }
 no_channels:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Channels count (%i) not available for playbacks: %s",
-            alsa->channels, snd_strerror (err)), (NULL));
+    gchar *msg = NULL;
+
+    if ((alsa->channels) == 1)
+      msg = g_strdup (_("Could not open device for playback in mono mode."));
+    if ((alsa->channels) == 2)
+      msg = g_strdup (_("Could not open device for playback in stereo mode."));
+    if ((alsa->channels) > 2)
+      msg =
+          g_strdup_printf (_
+          ("Could not open device for playback in %d-channel mode."),
+          alsa->channels);
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (msg), (snd_strerror (err)));
+    g_free (msg);
     return err;
   }
 no_rate:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
         ("Rate %iHz not available for playback: %s",
-            alsa->rate, snd_strerror (err)), (NULL));
+            alsa->rate, snd_strerror (err)));
     return err;
   }
 rate_match:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Rate doesn't match (requested %iHz, get %iHz)",
-            alsa->rate, err), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Rate doesn't match (requested %iHz, get %iHz)", alsa->rate, err));
     return -EINVAL;
   }
-buffer_time:
-  {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set buffer time %i for playback: %s",
-            alsa->buffer_time, snd_strerror (err)), (NULL));
-    return err;
-  }
 buffer_size:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to get buffer size for playback: %s", snd_strerror (err)),
-        (NULL));
-    return err;
-  }
-period_time:
-  {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set period time %i for playback: %s", alsa->period_time,
-            snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to get buffer size for playback: %s", snd_strerror (err)));
     return err;
   }
 period_size:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to get period size for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to get period size for playback: %s", snd_strerror (err)));
     return err;
   }
 set_hw_params:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set hw params for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to set hw params for playback: %s", snd_strerror (err)));
     return err;
   }
 }
@@ -414,37 +528,34 @@ set_swparams (GstAlsaSink * alsa)
   /* ERRORS */
 no_config:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
         ("Unable to determine current swparams for playback: %s",
-            snd_strerror (err)), (NULL));
+            snd_strerror (err)));
     return err;
   }
 start_threshold:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
         ("Unable to set start threshold mode for playback: %s",
-            snd_strerror (err)), (NULL));
+            snd_strerror (err)));
     return err;
   }
 set_avail:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set avail min for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to set avail min for playback: %s", snd_strerror (err)));
     return err;
   }
 set_align:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set transfer align for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to set transfer align for playback: %s", snd_strerror (err)));
     return err;
   }
 set_sw_params:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Unable to set sw params for playback: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Unable to set sw params for playback: %s", snd_strerror (err)));
     return err;
   }
 }
@@ -454,6 +565,10 @@ alsasink_parse_spec (GstAlsaSink * alsa, GstRingBufferSpec * spec)
 {
   switch (spec->type) {
     case GST_BUFTYPE_LINEAR:
+      GST_DEBUG_OBJECT (alsa,
+          "Linear format : depth=%d, width=%d, sign=%d, bigend=%d", spec->depth,
+          spec->width, spec->sign, spec->bigend);
+
       alsa->format = snd_pcm_build_linear_format (spec->depth, spec->width,
           spec->sign ? 0 : 1, spec->bigend ? 1 : 0);
       break;
@@ -517,10 +632,12 @@ gst_alsasink_open (GstAudioSink * asink)
 open_error:
   {
     if (err == -EBUSY) {
-      GST_ELEMENT_ERROR (alsa, RESOURCE, BUSY, (NULL), (NULL));
+      GST_ELEMENT_ERROR (alsa, RESOURCE, BUSY, (NULL), ("Device '%s' is busy",
+              alsa->device));
     } else {
       GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_WRITE,
-          (NULL), ("Playback open error: %s", snd_strerror (err)));
+          (NULL), ("Playback open error on device '%s': %s", alsa->device,
+              snd_strerror (err)));
     }
     return FALSE;
   }
@@ -545,36 +662,32 @@ gst_alsasink_prepare (GstAudioSink * asink, GstRingBufferSpec * spec)
   alsa->bytes_per_sample = spec->bytes_per_sample;
   spec->segsize = alsa->period_size * spec->bytes_per_sample;
   spec->segtotal = alsa->buffer_size / alsa->period_size;
-  spec->silence_sample[0] = 0;
-  spec->silence_sample[1] = 0;
-  spec->silence_sample[2] = 0;
-  spec->silence_sample[3] = 0;
 
   return TRUE;
 
   /* ERRORS */
 spec_parse:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Error parsing spec"), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Error parsing spec"));
     return FALSE;
   }
 non_block:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Could not set device to blocking: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Could not set device to blocking: %s", snd_strerror (err)));
     return FALSE;
   }
 hw_params_failed:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Setting of hwparams failed: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Setting of hwparams failed: %s", snd_strerror (err)));
     return FALSE;
   }
 sw_params_failed:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Setting of swparams failed: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Setting of swparams failed: %s", snd_strerror (err)));
     return FALSE;
   }
 }
@@ -598,21 +711,20 @@ gst_alsasink_unprepare (GstAudioSink * asink)
   /* ERRORS */
 drop:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Could not drop samples: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Could not drop samples: %s", snd_strerror (err)));
     return FALSE;
   }
 hw_free:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Could not free hw params: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Could not free hw params: %s", snd_strerror (err)));
     return FALSE;
   }
 non_block:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-        ("Could not set device to nonblocking: %s", snd_strerror (err)),
-        (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS, (NULL),
+        ("Could not set device to nonblocking: %s", snd_strerror (err)));
     return FALSE;
   }
 }
@@ -625,14 +737,15 @@ gst_alsasink_close (GstAudioSink * asink)
 
   CHECK (snd_pcm_close (alsa->handle), close_error);
   alsa->handle = NULL;
+  gst_caps_replace (&alsa->cached_caps, NULL);
 
   return TRUE;
 
   /* ERRORS */
 close_error:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, CLOSE,
-        ("Playback close error: %s", snd_strerror (err)), (NULL));
+    GST_ELEMENT_ERROR (alsa, RESOURCE, CLOSE, (NULL),
+        ("Playback close error: %s", snd_strerror (err)));
     return FALSE;
   }
 }
@@ -642,14 +755,15 @@ close_error:
  *   Underrun and suspend recovery
  */
 static gint
-xrun_recovery (snd_pcm_t * handle, gint err)
+xrun_recovery (GstAlsaSink * alsa, snd_pcm_t * handle, gint err)
 {
-  GST_DEBUG ("xrun recovery %d", err);
+  GST_DEBUG_OBJECT (alsa, "xrun recovery %d", err);
 
   if (err == -EPIPE) {          /* under-run */
     err = snd_pcm_prepare (handle);
     if (err < 0)
-      GST_WARNING ("Can't recovery from underrun, prepare failed: %s",
+      GST_WARNING_OBJECT (alsa,
+          "Can't recovery from underrun, prepare failed: %s",
           snd_strerror (err));
     return 0;
   } else if (err == -ESTRPIPE) {
@@ -659,7 +773,8 @@ xrun_recovery (snd_pcm_t * handle, gint err)
     if (err < 0) {
       err = snd_pcm_prepare (handle);
       if (err < 0)
-        GST_WARNING ("Can't recovery from suspend, prepare failed: %s",
+        GST_WARNING_OBJECT (alsa,
+            "Can't recovery from suspend, prepare failed: %s",
             snd_strerror (err));
     }
     return 0;
@@ -680,14 +795,16 @@ gst_alsasink_write (GstAudioSink * asink, gpointer data, guint length)
   cptr = length / alsa->bytes_per_sample;
   ptr = data;
 
+  GST_ALSA_LOCK (asink);
   while (cptr > 0) {
     err = snd_pcm_writei (alsa->handle, ptr, cptr);
 
+    GST_DEBUG_OBJECT (asink, "written %d result %d", cptr, err);
     if (err < 0) {
+      GST_DEBUG_OBJECT (asink, "Write error: %s", snd_strerror (err));
       if (err == -EAGAIN) {
-        GST_DEBUG ("Write error: %s", snd_strerror (err));
         continue;
-      } else if (xrun_recovery (alsa->handle, err) < 0) {
+      } else if (xrun_recovery (alsa, alsa->handle, err) < 0) {
         goto write_error;
       }
       continue;
@@ -696,11 +813,13 @@ gst_alsasink_write (GstAudioSink * asink, gpointer data, guint length)
     ptr += err * alsa->channels;
     cptr -= err;
   }
+  GST_ALSA_UNLOCK (asink);
 
   return length - cptr;
 
 write_error:
   {
+    GST_ALSA_UNLOCK (asink);
     return length;              /* skip one period */
   }
 }
@@ -721,29 +840,34 @@ gst_alsasink_delay (GstAudioSink * asink)
 static void
 gst_alsasink_reset (GstAudioSink * asink)
 {
-#if 0
   GstAlsaSink *alsa;
   gint err;
 
   alsa = GST_ALSA_SINK (asink);
 
+  GST_ALSA_LOCK (asink);
+  GST_DEBUG_OBJECT (alsa, "drop");
   CHECK (snd_pcm_drop (alsa->handle), drop_error);
+  GST_DEBUG_OBJECT (alsa, "prepare");
   CHECK (snd_pcm_prepare (alsa->handle), prepare_error);
+  GST_DEBUG_OBJECT (alsa, "reset done");
+  GST_ALSA_UNLOCK (asink);
 
   return;
 
   /* ERRORS */
 drop_error:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS,
         ("alsa-reset: pcm drop error: %s", snd_strerror (err)), (NULL));
+    GST_ALSA_UNLOCK (asink);
     return;
   }
 prepare_error:
   {
-    GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
+    GST_ELEMENT_ERROR (alsa, RESOURCE, SETTINGS,
         ("alsa-reset: pcm prepare error: %s", snd_strerror (err)), (NULL));
+    GST_ALSA_UNLOCK (asink);
     return;
   }
-#endif
 }
