@@ -52,12 +52,13 @@
 #include <alsa/asoundlib.h>
 
 #include "gstalsasrc.h"
+#include "gstalsadeviceprobe.h"
 
 #include <gst/gst-i18n-plugin.h>
 
 /* elementfactory information */
-static GstElementDetails gst_alsasrc_details =
-GST_ELEMENT_DETAILS ("Audio Src (ALSA)",
+static const GstElementDetails gst_alsasrc_details =
+GST_ELEMENT_DETAILS ("Audio source (ALSA)",
     "Source/Audio",
     "Read from a sound card via ALSA",
     "Wim Taymans <wim@fluendo.com>");
@@ -72,8 +73,10 @@ enum
   PROP_DEVICE_NAME,
 };
 
-GST_BOILERPLATE_WITH_INTERFACE (GstAlsaSrc, gst_alsasrc, GstAudioSrc,
-    GST_TYPE_AUDIO_SRC, GstMixer, GST_TYPE_MIXER, gst_alsasrc_mixer);
+static void gst_alsasrc_init_interfaces (GType type);
+
+GST_BOILERPLATE_FULL (GstAlsaSrc, gst_alsasrc, GstAudioSrc,
+    GST_TYPE_AUDIO_SRC, gst_alsasrc_init_interfaces);
 
 GST_IMPLEMENT_ALSA_MIXER_METHODS (GstAlsaSrc, gst_alsasrc_mixer);
 
@@ -111,24 +114,71 @@ static GstStaticPadTemplate alsasrc_src_factory =
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("audio/x-raw-int, "
+        "endianness = (int) { " ALSA_SRC_FACTORY_ENDIANNESS " }, "
+        "signed = (boolean) { TRUE, FALSE }, "
+        "width = (int) 32, "
+        "depth = (int) 32, "
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]; "
+        "audio/x-raw-int, "
         "endianness = (int) { " ALSA_SRC_FACTORY_ENDIANNESS " }, "
         "signed = (boolean) { TRUE, FALSE }, "
         "width = (int) 16, "
         "depth = (int) 16, "
-        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, 2 ]; "
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]; "
         "audio/x-raw-int, "
         "signed = (boolean) { TRUE, FALSE }, "
         "width = (int) 8, "
         "depth = (int) 8, "
-        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, 2 ]")
+        "rate = (int) [ 1, MAX ], " "channels = (int) [ 1, MAX ]")
     );
 
 static void
 gst_alsasrc_dispose (GObject * object)
 {
+  GstAlsaSrc *src = GST_ALSA_SRC (object);
+
+  g_free (src->device);
+  src->device = NULL;
+
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
+static gboolean
+gst_alsasrc_interface_supported (GstAlsaSrc * this, GType interface_type)
+{
+  /* only support this one interface (wrapped by GstImplementsInterface) */
+  g_assert (interface_type == GST_TYPE_MIXER);
+
+  return gst_alsasrc_mixer_supported (this, interface_type);
+}
+
+static void
+gst_implements_interface_init (GstImplementsInterfaceClass * klass)
+{
+  klass->supported = (gpointer) gst_alsasrc_interface_supported;
+}
+
+static void
+gst_alsasrc_init_interfaces (GType type)
+{
+  static const GInterfaceInfo implements_iface_info = {
+    (GInterfaceInitFunc) gst_implements_interface_init,
+    NULL,
+    NULL,
+  };
+  static const GInterfaceInfo mixer_iface_info = {
+    (GInterfaceInitFunc) gst_alsasrc_mixer_interface_init,
+    NULL,
+    NULL,
+  };
+
+  g_type_add_interface_static (type, GST_TYPE_IMPLEMENTS_INTERFACE,
+      &implements_iface_info);
+  g_type_add_interface_static (type, GST_TYPE_MIXER, &mixer_iface_info);
+
+  gst_alsa_type_add_device_property_probe_interface (type);
+}
+
 static void
 gst_alsasrc_base_init (gpointer g_class)
 {
@@ -190,9 +240,11 @@ gst_alsasrc_set_property (GObject * object, guint prop_id,
 
   switch (prop_id) {
     case PROP_DEVICE:
-      if (src->device)
-        g_free (src->device);
-      src->device = g_strdup (g_value_get_string (value));
+      g_free (src->device);
+      src->device = g_value_dup_string (value);
+      if (src->device == NULL) {
+        src->device = g_strdup (DEFAULT_PROP_DEVICE);
+      }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -236,12 +288,7 @@ gst_alsasrc_init (GstAlsaSrc * alsasrc, GstAlsaSrcClass * g_class)
   GST_DEBUG_OBJECT (alsasrc, "initializing");
 
   alsasrc->device = g_strdup (DEFAULT_PROP_DEVICE);
-}
-
-static GstCaps *
-gst_alsasrc_getcaps (GstBaseSrc * bsrc)
-{
-  return NULL;
+  alsasrc->cached_caps = NULL;
 }
 
 #define CHECK(call, error) \
@@ -250,6 +297,43 @@ if ((err = call) < 0)           \
   goto error;                   \
 } G_STMT_END;
 
+
+static GstCaps *
+gst_alsasrc_getcaps (GstBaseSrc * bsrc)
+{
+  GstElementClass *element_class;
+  GstPadTemplate *pad_template;
+  GstAlsaSrc *src;
+  GstCaps *caps;
+
+  src = GST_ALSA_SRC (bsrc);
+
+  if (src->handle == NULL) {
+    GST_DEBUG_OBJECT (src, "device not open, using template caps");
+    return NULL;                /* base class will get template caps for us */
+  }
+
+  if (src->cached_caps) {
+    GST_LOG_OBJECT (src, "Returning cached caps");
+    return gst_caps_ref (src->cached_caps);
+  }
+
+  element_class = GST_ELEMENT_GET_CLASS (src);
+  pad_template = gst_element_class_get_pad_template (element_class, "src");
+  g_return_val_if_fail (pad_template != NULL, NULL);
+
+  caps = gst_alsa_probe_supported_formats (GST_OBJECT (src), src->handle,
+      gst_pad_template_get_caps (pad_template));
+
+  if (caps) {
+    src->cached_caps = gst_caps_ref (caps);
+  }
+
+  GST_INFO_OBJECT (src, "returning caps %" GST_PTR_FORMAT, caps);
+
+  return caps;
+}
+
 static int
 set_hwparams (GstAlsaSrc * alsa)
 {
@@ -520,10 +604,12 @@ gst_alsasrc_open (GstAudioSrc * asrc)
 open_error:
   {
     if (err == -EBUSY) {
-      GST_ELEMENT_ERROR (alsa, RESOURCE, BUSY, (NULL), (NULL));
+      GST_ELEMENT_ERROR (alsa, RESOURCE, BUSY, (NULL), ("Device '%s' is busy",
+              alsa->device));
     } else {
-      GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_READ,
-          (NULL), ("Recording open error: %s", snd_strerror (err)));
+      GST_ELEMENT_ERROR (alsa, RESOURCE, OPEN_WRITE,
+          (NULL), ("Recording open error on device '%s': %s", alsa->device,
+              snd_strerror (err)));
     }
     return FALSE;
   }
@@ -638,6 +724,8 @@ gst_alsasrc_close (GstAudioSrc * asrc)
     alsa->mixer = NULL;
   }
 
+  gst_caps_replace (&alsa->cached_caps, NULL);
+
   return TRUE;
 }
 
@@ -645,14 +733,15 @@ gst_alsasrc_close (GstAudioSrc * asrc)
  *   Underrun and suspend recovery
  */
 static gint
-xrun_recovery (snd_pcm_t * handle, gint err)
+xrun_recovery (GstAlsaSrc * alsa, snd_pcm_t * handle, gint err)
 {
-  GST_DEBUG ("xrun recovery %d", err);
+  GST_DEBUG_OBJECT (alsa, "xrun recovery %d", err);
 
   if (err == -EPIPE) {          /* under-run */
     err = snd_pcm_prepare (handle);
     if (err < 0)
-      GST_WARNING ("Can't recovery from underrun, prepare failed: %s",
+      GST_WARNING_OBJECT (alsa,
+          "Can't recovery from underrun, prepare failed: %s",
           snd_strerror (err));
     return 0;
   } else if (err == -ESTRPIPE) {
@@ -662,7 +751,8 @@ xrun_recovery (snd_pcm_t * handle, gint err)
     if (err < 0) {
       err = snd_pcm_prepare (handle);
       if (err < 0)
-        GST_WARNING ("Can't recovery from suspend, prepare failed: %s",
+        GST_WARNING_OBJECT (alsa,
+            "Can't recovery from suspend, prepare failed: %s",
             snd_strerror (err));
     }
     return 0;
@@ -688,7 +778,7 @@ gst_alsasrc_read (GstAudioSrc * asrc, gpointer data, guint length)
       if (err == -EAGAIN) {
         GST_DEBUG_OBJECT (asrc, "Read error: %s", snd_strerror (err));
         continue;
-      } else if (xrun_recovery (alsa->handle, err) < 0) {
+      } else if (xrun_recovery (alsa, alsa->handle, err) < 0) {
         goto read_error;
       }
       continue;
