@@ -26,10 +26,11 @@
 #include <gst/base/gstcollectpads.h>
 
 #include <ogg/ogg.h>
-/* memcpy - if someone knows a way to get rid of it, please speak up 
+/* memcpy - if someone knows a way to get rid of it, please speak up
  * note: the ogg docs even say you need this... */
 #include <string.h>
 #include <time.h>
+#include <stdlib.h>             /* rand, srand, atoi */
 
 GST_DEBUG_CATEGORY_STATIC (gst_ogg_mux_debug);
 #define GST_CAT_DEFAULT gst_ogg_mux_debug
@@ -38,7 +39,7 @@ GST_DEBUG_CATEGORY_STATIC (gst_ogg_mux_debug);
 #define GST_OGG_MUX(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_OGG_MUX, GstOggMux))
 #define GST_OGG_MUX_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_OGG_MUX, GstOggMux))
 #define GST_IS_OGG_MUX(obj) (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_OGG_MUX))
-#define GST_IS_OGG_MUX_CLASS(obj) (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_OGG_MUX))
+#define GST_IS_OGG_MUX_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_OGG_MUX))
 
 /* This isn't generally what you'd want with an end-time macro, because
    technically the end time of a buffer with invalid duration is invalid. But
@@ -65,7 +66,10 @@ typedef struct
 {
   GstCollectData collect;       /* we extend the CollectData */
 
-  GstBuffer *buffer;            /* the queued buffer for this pad */
+  /* These two buffers make a very simple queue - they enter as 'next_buffer'
+   * and (usually) leave as 'buffer', except at EOS, when buffer will be NULL */
+  GstBuffer *buffer;            /* the first waiting buffer for the pad */
+  GstBuffer *next_buffer;       /* the second waiting buffer for the pad */
 
   gint serial;
   ogg_stream_state stream;
@@ -104,12 +108,18 @@ struct _GstOggMux
   /* sinkpads */
   GstCollectPads *collect;
 
+  /* number of pads which have not received EOS */
+  gint active_pads;
+
   /* the pad we are currently using to fill a page */
   GstOggPad *pulling;
 
   /* next timestamp for the page */
   GstClockTime next_ts;
 
+  /* Last timestamp actually output on src pad */
+  GstClockTime last_ts;
+
   /* offset in stream */
   guint64 offset;
 
@@ -138,7 +148,8 @@ struct _GstOggMuxClass
 };
 
 /* elementfactory information */
-static GstElementDetails gst_ogg_mux_details = GST_ELEMENT_DETAILS ("ogg muxer",
+static const GstElementDetails gst_ogg_mux_details =
+GST_ELEMENT_DETAILS ("Ogg muxer",
     "Codec/Muxer",
     "mux ogg streams (info about ogg: http://xiph.org)",
     "Wim Taymans <wim@fluendo.com>");
@@ -171,7 +182,8 @@ static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink_%d",
     GST_PAD_REQUEST,
     GST_STATIC_CAPS ("video/x-theora; "
         "audio/x-vorbis; audio/x-flac; audio/x-speex; "
-        "application/x-ogm-video; application/x-ogm-audio; video/x-dirac")
+        "application/x-ogm-video; application/x-ogm-audio; video/x-dirac; "
+        "video/x-smoke; text/x-cmml, encoded = (boolean) TRUE")
     );
 
 static void gst_ogg_mux_base_init (gpointer g_class);
@@ -202,7 +214,7 @@ gst_ogg_mux_get_type (void)
 {
   static GType ogg_mux_type = 0;
 
-  if (!ogg_mux_type) {
+  if (G_UNLIKELY (ogg_mux_type == 0)) {
     static const GTypeInfo ogg_mux_info = {
       sizeof (GstOggMuxClass),
       gst_ogg_mux_base_init,
@@ -244,7 +256,7 @@ gst_ogg_mux_class_init (GstOggMuxClass * klass)
   gobject_class = (GObjectClass *) klass;
   gstelement_class = (GstElementClass *) klass;
 
-  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);
+  parent_class = g_type_class_peek_parent (klass);
 
   gobject_class->finalize = gst_ogg_mux_finalize;
   gobject_class->get_property = gst_ogg_mux_get_property;
@@ -288,6 +300,9 @@ gst_ogg_mux_clear (GstOggMux * ogg_mux)
   ogg_mux->max_delay = DEFAULT_MAX_DELAY;
   ogg_mux->max_page_delay = DEFAULT_MAX_PAGE_DELAY;
   ogg_mux->delta_pad = NULL;
+  ogg_mux->offset = 0;
+  ogg_mux->next_ts = 0;
+  ogg_mux->last_ts = GST_CLOCK_TIME_NONE;
 }
 
 static void
@@ -308,7 +323,8 @@ gst_ogg_mux_init (GstOggMux * ogg_mux)
 
   ogg_mux->collect = gst_collect_pads_new ();
   gst_collect_pads_set_function (ogg_mux->collect,
-      (GstCollectPadsFunction) gst_ogg_mux_collected, ogg_mux);
+      (GstCollectPadsFunction) GST_DEBUG_FUNCPTR (gst_ogg_mux_collected),
+      ogg_mux);
 
   gst_ogg_mux_clear (ogg_mux);
 }
@@ -387,6 +403,7 @@ gst_ogg_mux_request_new_pad (GstElement * element,
       oggpad = (GstOggPad *)
           gst_collect_pads_add_pad (ogg_mux->collect, newpad,
           sizeof (GstOggPad));
+      ogg_mux->active_pads++;
 
       oggpad->serial = serial;
       ogg_stream_init (&oggpad->stream, serial);
@@ -430,34 +447,29 @@ gst_ogg_mux_release_pad (GstElement * element, GstPad * pad)
 
   ogg_mux = GST_OGG_MUX (gst_pad_get_parent (pad));
 
-  /* FIXME: When a request pad is released while paused or playing, 
-   * we probably need to do something to finalise its stream in the
-   * ogg data we're producing, but I'm not sure what */
-
   /* Find out GstOggPad in the collect pads info and clean it up */
 
   GST_OBJECT_LOCK (ogg_mux->collect);
-  walk = ogg_mux->collect->data;
-  while (walk) {
+  for (walk = ogg_mux->collect->data; walk; walk = g_slist_next (walk)) {
     GstOggPad *oggpad = (GstOggPad *) walk->data;
     GstCollectData *cdata = (GstCollectData *) walk->data;
     GstBuffer *buf;
 
     if (cdata->pad == pad) {
-      /* FIXME: clear the ogg stream stuff? - 
-       *    ogg_stream_clear (&oggpad->stream); */
+      ogg_stream_clear (&oggpad->stream);
 
       while ((buf = g_queue_pop_head (oggpad->pagebuffers)) != NULL) {
         gst_buffer_unref (buf);
       }
 
       g_queue_free (oggpad->pagebuffers);
+      oggpad->pagebuffers = NULL;
     }
-    walk = g_slist_next (walk);
   }
   GST_OBJECT_UNLOCK (ogg_mux->collect);
 
   gst_collect_pads_remove_pad (ogg_mux->collect, pad);
+  gst_element_remove_pad (element, pad);
 }
 
 /* handle events */
@@ -498,7 +510,7 @@ gst_ogg_mux_buffer_from_page (GstOggMux * mux, ogg_page * page, gboolean delta)
     GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DELTA_UNIT);
 
   GST_LOG_OBJECT (mux, GST_GP_FORMAT
-      " created buffer %p from ogg page", ogg_page_granulepos (page));
+      " created buffer %p from ogg page", ogg_page_granulepos (page), buffer);
 
   return buffer;
 }
@@ -506,11 +518,26 @@ gst_ogg_mux_buffer_from_page (GstOggMux * mux, ogg_page * page, gboolean delta)
 static GstFlowReturn
 gst_ogg_mux_push_buffer (GstOggMux * mux, GstBuffer * buffer)
 {
+  GstCaps *caps;
+
   /* fix up OFFSET and OFFSET_END again */
   GST_BUFFER_OFFSET (buffer) = mux->offset;
   mux->offset += GST_BUFFER_SIZE (buffer);
   GST_BUFFER_OFFSET_END (buffer) = mux->offset;
 
+  /* Ensure we have monotonically increasing timestamps in the output. */
+  if (GST_BUFFER_TIMESTAMP_IS_VALID (buffer)) {
+    if (mux->last_ts != GST_CLOCK_TIME_NONE &&
+        GST_BUFFER_TIMESTAMP (buffer) < mux->last_ts)
+      GST_BUFFER_TIMESTAMP (buffer) = mux->last_ts;
+    else
+      mux->last_ts = GST_BUFFER_TIMESTAMP (buffer);
+  }
+
+  caps = gst_pad_get_negotiated_caps (mux->srcpad);
+  gst_buffer_set_caps (buffer, caps);
+  gst_caps_unref (caps);
+
   return gst_pad_push (mux->srcpad, buffer);
 }
 
@@ -682,18 +709,24 @@ gst_ogg_mux_compare_pads (GstOggMux * ogg_mux, GstOggPad * first,
 
   /* if the first pad doesn't contain anything or is even NULL, return
    * the second pad as best candidate and vice versa */
-  if (first == NULL || first->buffer == NULL)
+  if (first == NULL || (first->buffer == NULL && first->next_buffer == NULL))
     return 1;
-  if (second == NULL || second->buffer == NULL)
+  if (second == NULL || (second->buffer == NULL && second->next_buffer == NULL))
     return -1;
 
   /* no timestamp on first buffer, it must go first */
-  firsttime = GST_BUFFER_TIMESTAMP (first->buffer);
+  if (first->buffer)
+    firsttime = GST_BUFFER_TIMESTAMP (first->buffer);
+  else
+    firsttime = GST_BUFFER_TIMESTAMP (first->next_buffer);
   if (firsttime == GST_CLOCK_TIME_NONE)
     return -1;
 
   /* no timestamp on second buffer, it must go first */
-  secondtime = GST_BUFFER_TIMESTAMP (second->buffer);
+  if (second->buffer)
+    secondtime = GST_BUFFER_TIMESTAMP (second->buffer);
+  else
+    secondtime = GST_BUFFER_TIMESTAMP (second->next_buffer);
   if (secondtime == GST_CLOCK_TIME_NONE)
     return 1;
 
@@ -716,9 +749,19 @@ gst_ogg_mux_compare_pads (GstOggMux * ogg_mux, GstOggPad * first,
   return 0;
 }
 
-/* make sure a buffer is queued on all pads, returns a pointer to an oggpad
- * that holds the best buffer or NULL when no pad was usable.
- * "best" means the buffer marked with the lowest timestamp */
+/* make sure at least one buffer is queued on all pads, two if possible
+ * 
+ * if pad->buffer == NULL, pad->next_buffer !=  NULL, then
+ *   we do not know if the buffer is the last or not
+ * if pad->buffer != NULL, pad->next_buffer != NULL, then
+ *   pad->buffer is not the last buffer for the pad
+ * if pad->buffer != NULL, pad->next_buffer == NULL, then
+ *   pad->buffer if the last buffer for the pad
+ * 
+ * returns a pointer to an oggpad that holds the best buffer, or
+ * NULL when no pad was usable. "best" means the buffer marked
+ * with the lowest timestamp. If best->buffer == NULL then nothing
+ * should be done until more data arrives */
 static GstOggPad *
 gst_ogg_mux_queue_pads (GstOggMux * ogg_mux)
 {
@@ -743,6 +786,14 @@ gst_ogg_mux_queue_pads (GstOggMux * ogg_mux)
       GstBuffer *buf;
       gboolean incaps;
 
+      /* shift the buffer along if needed (it's okay if next_buffer is NULL) */
+      if (pad->buffer == NULL) {
+        GST_LOG_OBJECT (data->pad, "shifting buffer %" GST_PTR_FORMAT,
+            pad->next_buffer);
+        pad->buffer = pad->next_buffer;
+        pad->next_buffer = NULL;
+      }
+
       buf = gst_collect_pads_pop (ogg_mux->collect, data);
       GST_LOG_OBJECT (data->pad, "popped buffer %" GST_PTR_FORMAT, buf);
 
@@ -767,17 +818,38 @@ gst_ogg_mux_queue_pads (GstOggMux * ogg_mux)
         }
       } else {
         GST_DEBUG_OBJECT (data->pad, "EOS on pad");
-        pad->eos = TRUE;
+        if (!pad->eos) {
+          ogg_page page;
+          GstFlowReturn ret;
+
+          /* it's no longer active */
+          ogg_mux->active_pads--;
+
+          /* Just gone to EOS. Flush existing page(s) */
+          pad->eos = TRUE;
+
+          while (ogg_stream_flush (&pad->stream, &page)) {
+            /* Place page into the per-pad queue */
+            ret = gst_ogg_mux_pad_queue_page (ogg_mux, pad, &page,
+                pad->first_delta);
+            /* increment the page number counter */
+            pad->pageno++;
+            /* mark other pages as delta */
+            pad->first_delta = TRUE;
+          }
+        }
       }
 
-      pad->buffer = buf;
+      pad->next_buffer = buf;
     }
 
     /* we should have a buffer now, see if it is the best pad to
      * pull on */
-    if (pad->buffer) {
+    if (pad->buffer || pad->next_buffer) {
       if (gst_ogg_mux_compare_pads (ogg_mux, bestpad, pad) > 0) {
-        GST_LOG_OBJECT (data->pad, "new best pad");
+        GST_LOG_OBJECT (data->pad,
+            "new best pad, with buffers %" GST_PTR_FORMAT
+            " and %" GST_PTR_FORMAT, pad->buffer, pad->next_buffer);
 
         bestpad = pad;
       }
@@ -800,7 +872,7 @@ gst_ogg_mux_get_headers (GstOggPad * pad)
   GList *res = NULL;
   GstOggMux *ogg_mux;
   GstStructure *structure;
-  const GstCaps *caps;
+  GstCaps *caps;
   GstPad *thepad;
 
   thepad = pad->collect.pad;
@@ -814,34 +886,51 @@ gst_ogg_mux_get_headers (GstOggPad * pad)
     const GValue *streamheader;
 
     structure = gst_caps_get_structure (caps, 0);
-    streamheader = gst_structure_get_value (structure, "streamheader");
-    if (streamheader != NULL) {
-      GST_LOG_OBJECT (thepad, "got header");
-      if (G_VALUE_TYPE (streamheader) == GST_TYPE_ARRAY) {
-        GArray *bufarr = g_value_peek_pointer (streamheader);
-        gint i;
+    if (strcmp (gst_structure_get_name (structure), "video/x-dirac") == 0) {
+      GstBuffer *buf = gst_buffer_new_and_alloc (16);
+      int fps_n = 12;
+      int fps_d = 1;
 
-        GST_LOG_OBJECT (thepad, "got fixed list");
+      gst_structure_get_fraction (structure, "framerate", &fps_n, &fps_d);
 
-        for (i = 0; i < bufarr->len; i++) {
-          GValue *bufval = &g_array_index (bufarr, GValue, i);
+      memcpy (GST_BUFFER_DATA (buf), "KW-DIRAC", 8);
+      GST_WRITE_UINT32_BE (GST_BUFFER_DATA (buf) + 8, fps_n);
+      GST_WRITE_UINT32_BE (GST_BUFFER_DATA (buf) + 12, fps_d);
 
-          GST_LOG_OBJECT (thepad, "item %d", i);
-          if (G_VALUE_TYPE (bufval) == GST_TYPE_BUFFER) {
-            GstBuffer *buf = g_value_peek_pointer (bufval);
+      res = g_list_append (res, buf);
 
-            GST_LOG_OBJECT (thepad, "adding item %d to header list", i);
+      //res = g_list_append (res, gst_buffer_ref(pad->buffer));
+    } else {
+      streamheader = gst_structure_get_value (structure, "streamheader");
+      if (streamheader != NULL) {
+        GST_LOG_OBJECT (thepad, "got header");
+        if (G_VALUE_TYPE (streamheader) == GST_TYPE_ARRAY) {
+          GArray *bufarr = g_value_peek_pointer (streamheader);
+          gint i;
 
-            gst_buffer_ref (buf);
-            res = g_list_append (res, buf);
+          GST_LOG_OBJECT (thepad, "got fixed list");
+
+          for (i = 0; i < bufarr->len; i++) {
+            GValue *bufval = &g_array_index (bufarr, GValue, i);
+
+            GST_LOG_OBJECT (thepad, "item %d", i);
+            if (G_VALUE_TYPE (bufval) == GST_TYPE_BUFFER) {
+              GstBuffer *buf = g_value_peek_pointer (bufval);
+
+              GST_LOG_OBJECT (thepad, "adding item %d to header list", i);
+
+              gst_buffer_ref (buf);
+              res = g_list_append (res, buf);
+            }
           }
+        } else {
+          GST_LOG_OBJECT (thepad, "streamheader is not fixed list");
         }
       } else {
-        GST_LOG_OBJECT (thepad, "streamheader is not fixed list");
+        GST_LOG_OBJECT (thepad, "caps don't have streamheader");
       }
-    } else {
-      GST_LOG_OBJECT (thepad, "caps done have streamheader");
     }
+    gst_caps_unref (caps);
   } else {
     GST_LOG_OBJECT (thepad, "got empty caps as negotiated format");
   }
@@ -864,6 +953,7 @@ gst_ogg_mux_set_header_on_caps (GstCaps * caps, GList * buffers)
 
   while (walk) {
     GstBuffer *buf = GST_BUFFER (walk->data);
+    GstBuffer *copy;
     GValue value = { 0 };
 
     walk = walk->next;
@@ -873,7 +963,9 @@ gst_ogg_mux_set_header_on_caps (GstCaps * caps, GList * buffers)
     GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_IN_CAPS);
 
     g_value_init (&value, GST_TYPE_BUFFER);
-    gst_value_set_buffer (&value, buf);
+    copy = gst_buffer_copy (buf);
+    gst_value_set_buffer (&value, copy);
+    gst_buffer_unref (copy);
     gst_value_array_append_value (&array, &value);
     g_value_unset (&value);
   }
@@ -917,7 +1009,7 @@ gst_ogg_mux_send_headers (GstOggMux * mux)
     GST_LOG_OBJECT (mux, "looking at pad %s:%s", GST_DEBUG_PAD_NAME (thepad));
 
     /* if the pad has no buffer, we don't care */
-    if (pad->buffer == NULL)
+    if (pad->buffer == NULL && pad->next_buffer == NULL)
       continue;
 
     /* now figure out the headers */
@@ -953,6 +1045,9 @@ gst_ogg_mux_send_headers (GstOggMux * mux)
     } else if (pad->buffer) {
       buf = pad->buffer;
       gst_buffer_ref (buf);
+    } else if (pad->next_buffer) {
+      buf = pad->next_buffer;
+      gst_buffer_ref (buf);
     } else {
       /* fixme -- should be caught in the previous list traversal. */
       GST_OBJECT_LOCK (pad);
@@ -1040,8 +1135,8 @@ gst_ogg_mux_send_headers (GstOggMux * mux)
       /* if last header, flush page */
       if (hwalk == NULL) {
         GST_LOG_OBJECT (mux,
-            "flushing page as packet %d is first or last packet",
-            pad->packetno);
+            "flushing page as packet %" G_GUINT64_FORMAT " is first or "
+            "last packet", pad->packetno);
         while (ogg_stream_flush (&pad->stream, &page)) {
           GstBuffer *hbuf = gst_ogg_mux_buffer_from_page (mux, &page, FALSE);
 
@@ -1086,59 +1181,44 @@ gst_ogg_mux_send_headers (GstOggMux * mux)
   return ret;
 }
 
-/* this function is called when there is data on all pads.
+/* this function is called to process data on the best pending pad.
  *
  * basic idea:
  *
- * 1) find a pad to pull on, this is done by looking at the buffers
- *    to decide which one to use, we use the 'oldest' one first.
- * 2) store the selected pad and keep on pulling until we fill a
+ * 1) store the selected pad and keep on pulling until we fill a
  *    complete ogg page or the ogg page is filled above the max-delay
  *    threshold. This is needed because the ogg spec says that
  *    you should fill a complete page with data from the same logical
  *    stream. When the page is filled, go back to 1).
- * 3) before filling a page, read ahead one more buffer to see if this
+ * 2) before filling a page, read ahead one more buffer to see if this
  *    packet is the last of the stream. We need to do this because the ogg
  *    spec mandates that the last packet should have the EOS flag set before
- *    sending it to ogg. FIXME: Apparently we're allowed to send empty 'nil'
- *    pages with the EOS flag set for EOS, so we could do this. Not sure how
- *    that works, though. TODO: 'read ahead one more buffer' is a bit funky
- *    with collectpads. Rethink this.
- * 4) pages get queued on a per-pad queue. Every time a page is queued, a
+ *    sending it to ogg. if pad->buffer is NULL we need to wait to find out
+ *    whether there are any more buffers.
+ * 3) pages get queued on a per-pad queue. Every time a page is queued, a
  *    dequeue is called, which will dequeue the oldest page on any pad, provided
  *    that ALL pads have at least one marked page in the queue (or remaining
- *    pad are at EOS)
+ *    pads are at EOS)
  */
 static GstFlowReturn
-gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
+gst_ogg_mux_process_best_pad (GstOggMux * ogg_mux, GstOggPad * best)
 {
-  GstOggPad *best;
   gboolean delta_unit;
   GstFlowReturn ret;
   gint64 granulepos = 0;
   GstClockTime timestamp, gp_time;
 
-  GST_LOG_OBJECT (ogg_mux, "collected");
-
-  /* queue buffers on all pads; find a buffer with the lowest timestamp */
-  best = gst_ogg_mux_queue_pads (ogg_mux);
-  if (best && !best->buffer) {
-    GST_DEBUG_OBJECT (ogg_mux, "No buffer available on best pad");
-    return GST_FLOW_OK;
-  }
-
-  if (!best) {
-    /* EOS : FIXME !! We need to handle EOS correctly, and set EOS
-       flags on the ogg pages. */
-    GST_DEBUG_OBJECT (ogg_mux, "Pushing EOS");
-    gst_pad_push_event (ogg_mux->srcpad, gst_event_new_eos ());
-    return GST_FLOW_WRONG_STATE;
-  }
-
   GST_LOG_OBJECT (ogg_mux, "best pad %" GST_PTR_FORMAT
       ", currently pulling from %" GST_PTR_FORMAT, best->collect.pad,
       ogg_mux->pulling);
 
+  /* best->buffer is non-NULL, either the pad is EOS's or there is a next 
+   * buffer */
+  if (best->next_buffer == NULL && !best->eos) {
+    GST_WARNING_OBJECT (ogg_mux, "no subsequent buffer and EOS not reached");
+    return GST_FLOW_WRONG_STATE;
+  }
+
   /* if we were already pulling from one pad, but the new "best" buffer is
    * from another pad, we need to check if we have reason to flush a page
    * for the pad we were pulling from before */
@@ -1240,7 +1320,7 @@ gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
         " packet %" G_GINT64_FORMAT " (%ld bytes) created from buffer",
         packet.granulepos, packet.packetno, packet.bytes);
 
-    packet.e_o_s = 0;
+    packet.e_o_s = (pad->eos ? 1 : 0);
     tmpbuf = NULL;
 
     /* we flush when we see a new keyframe */
@@ -1254,6 +1334,13 @@ gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
       }
     }
 
+    if (GST_BUFFER_IS_DISCONT (buf)) {
+      packet.packetno++;
+      /* No public API for this; hack things in */
+      pad->stream.pageno++;
+      force_flush = TRUE;
+    }
+
     /* flush the currently built page if necessary */
     if (force_flush) {
       GST_LOG_OBJECT (pad->collect.pad,
@@ -1337,7 +1424,7 @@ gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
           GST_TIME_FORMAT ") caused new page",
           granulepos, packet.packetno, GST_TIME_ARGS (timestamp));
       GST_LOG_OBJECT (pad->collect.pad,
-          GST_GP_FORMAT " new page %d", ogg_page_granulepos (&page),
+          GST_GP_FORMAT " new page %ld", ogg_page_granulepos (&page),
           pad->stream.pageno);
 
       if (ogg_page_granulepos (&page) == granulepos) {
@@ -1404,6 +1491,54 @@ gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
   return GST_FLOW_OK;
 }
 
+/* This function is called when there is data on all pads.
+ * 
+ * It finds a pad to pull on, this is done by looking at the buffers
+ * to decide which one to use, and using the 'oldest' one first. It then calls
+ * gst_ogg_mux_process_best_pad() to process as much data as possible.
+ * 
+ * If all the pads have received EOS, it flushes out all data by continually
+ * getting the best pad and calling gst_ogg_mux_process_best_pad() until they
+ * are all empty, and then sends EOS.
+ */
+static GstFlowReturn
+gst_ogg_mux_collected (GstCollectPads * pads, GstOggMux * ogg_mux)
+{
+  GstOggPad *best;
+  GstFlowReturn ret;
+
+  GST_LOG_OBJECT (ogg_mux, "collected");
+
+  /* queue buffers on all pads; find a buffer with the lowest timestamp */
+  best = gst_ogg_mux_queue_pads (ogg_mux);
+  if (best && !best->buffer) {
+    GST_DEBUG_OBJECT (ogg_mux, "No buffer available on best pad");
+    return GST_FLOW_OK;
+  }
+
+  if (!best) {
+    return GST_FLOW_WRONG_STATE;
+  }
+
+  ret = gst_ogg_mux_process_best_pad (ogg_mux, best);
+
+  /* if all the pads have been removed, flush all pending data */
+  if ((ret == GST_FLOW_OK) && (ogg_mux->active_pads == 0)) {
+    GST_LOG_OBJECT (ogg_mux, "no pads remaining, flushing data");
+
+    do {
+      best = gst_ogg_mux_queue_pads (ogg_mux);
+      if (best)
+        ret = gst_ogg_mux_process_best_pad (ogg_mux, best);
+    } while ((ret == GST_FLOW_OK) && (best != NULL));
+
+    GST_DEBUG_OBJECT (ogg_mux, "Pushing EOS");
+    gst_pad_push_event (ogg_mux->srcpad, gst_event_new_eos ());
+  }
+
+  return ret;
+}
+
 static void
 gst_ogg_mux_get_property (GObject * object,
     guint prop_id, GValue * value, GParamSpec * pspec)
@@ -1446,22 +1581,15 @@ gst_ogg_mux_set_property (GObject * object,
   }
 }
 
-/* Clear all buffers from the collectpads object */
+/* reset all variables in the ogg pads. */
 static void
-gst_ogg_mux_clear_collectpads (GstCollectPads * collect)
+gst_ogg_mux_init_collectpads (GstCollectPads * collect)
 {
   GSList *walk;
 
   walk = collect->data;
   while (walk) {
     GstOggPad *oggpad = (GstOggPad *) walk->data;
-    GstBuffer *buf;
-
-    ogg_stream_clear (&oggpad->stream);
-
-    while ((buf = g_queue_pop_head (oggpad->pagebuffers)) != NULL) {
-      gst_buffer_unref (buf);
-    }
 
     ogg_stream_init (&oggpad->stream, oggpad->serial);
     oggpad->packetno = 0;
@@ -1478,6 +1606,26 @@ gst_ogg_mux_clear_collectpads (GstCollectPads * collect)
   }
 }
 
+/* Clear all buffers from the collectpads object */
+static void
+gst_ogg_mux_clear_collectpads (GstCollectPads * collect)
+{
+  GSList *walk;
+
+  for (walk = collect->data; walk; walk = g_slist_next (walk)) {
+    GstOggPad *oggpad = (GstOggPad *) walk->data;
+    GstBuffer *buf;
+
+    ogg_stream_clear (&oggpad->stream);
+
+    while ((buf = g_queue_pop_head (oggpad->pagebuffers)) != NULL) {
+      gst_buffer_unref (buf);
+    }
+    g_queue_free (oggpad->pagebuffers);
+    oggpad->pagebuffers = NULL;
+  }
+}
+
 static GstStateChangeReturn
 gst_ogg_mux_change_state (GstElement * element, GstStateChange transition)
 {
@@ -1490,11 +1638,9 @@ gst_ogg_mux_change_state (GstElement * element, GstStateChange transition)
     case GST_STATE_CHANGE_NULL_TO_READY:
       break;
     case GST_STATE_CHANGE_READY_TO_PAUSED:
-      ogg_mux->next_ts = 0;
-      ogg_mux->offset = 0;
-      ogg_mux->pulling = NULL;
-      gst_collect_pads_start (ogg_mux->collect);
       gst_ogg_mux_clear (ogg_mux);
+      gst_ogg_mux_init_collectpads (ogg_mux->collect);
+      gst_collect_pads_start (ogg_mux->collect);
       break;
     case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
       break;
@@ -1510,9 +1656,11 @@ gst_ogg_mux_change_state (GstElement * element, GstStateChange transition)
   switch (transition) {
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       break;
-    case GST_STATE_CHANGE_READY_TO_NULL:
+    case GST_STATE_CHANGE_PAUSED_TO_READY:
       gst_ogg_mux_clear_collectpads (ogg_mux->collect);
       break;
+    case GST_STATE_CHANGE_READY_TO_NULL:
+      break;
     default:
       break;
   }
