@@ -30,15 +30,23 @@ G_BEGIN_DECLS
 #define GST_THEORA_ENC(obj) \
   (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_THEORA_ENC,GstTheoraEnc))
 #define GST_THEORA_ENC_CLASS(klass) \
-  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_THEORA_ENC,GstTheoraEnc))
+  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_THEORA_ENC,GstTheoraEncClass))
 #define GST_IS_THEORA_ENC(obj) \
   (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_THEORA_ENC))
-#define GST_IS_THEORA_ENC_CLASS(obj) \
+#define GST_IS_THEORA_ENC_CLASS(klass) \
   (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_THEORA_ENC))
 
 typedef struct _GstTheoraEnc GstTheoraEnc;
 typedef struct _GstTheoraEncClass GstTheoraEncClass;
 
+/**
+ * GstTheoraEncBorderMode:
+ * @BORDER_NONE: no border
+ * @BORDER_BLACK: black border
+ * @BORDER_MIRROR: Mirror image in border
+ *
+ * Border color to add when sizes not multiple of 16. 
+ */ 
 typedef enum
 {
   BORDER_NONE,
@@ -47,6 +55,11 @@ typedef enum
 }
 GstTheoraEncBorderMode;
 
+/**
+ * GstTheoraEnc:
+ *
+ * Opaque data structure.
+ */
 struct _GstTheoraEnc
 {
   GstElement element;
@@ -77,11 +90,17 @@ struct _GstTheoraEnc
   gint info_width, info_height;
   gint width, height;
   gint offset_x, offset_y;
-  gdouble fps;
+  gint fps_n, fps_d;
+  GstClockTime next_ts;
+
+  GstClockTime expected_ts;
+  gboolean next_discont;
 
   guint packetno;
   guint64 bytes_out;
-  guint64 initial_delay;
+  guint64 granulepos_offset;
+  guint64 timestamp_offset;
+  gint granule_shift;  
 };
 
 struct _GstTheoraEncClass
