@@ -60,18 +60,21 @@
 #endif
 
 #include "gsttheoraenc.h"
+
 #include <string.h>
+#include <stdlib.h>             /* free */
+
 #include <gst/tag/tag.h>
 
-GST_DEBUG_CATEGORY (theoraenc_debug);
 #define GST_CAT_DEFAULT theoraenc_debug
+GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
 
 #define GST_TYPE_BORDER_MODE (gst_border_mode_get_type())
 static GType
 gst_border_mode_get_type (void)
 {
   static GType border_mode_type = 0;
-  static GEnumValue border_mode[] = {
+  static const GEnumValue border_mode[] = {
     {BORDER_NONE, "No Border", "none"},
     {BORDER_BLACK, "Black Border", "black"},
     {BORDER_MIRROR, "Mirror image in borders", "mirror"},
@@ -85,10 +88,6 @@ gst_border_mode_get_type (void)
   return border_mode_type;
 }
 
-#define ROUND_UP_2(x) (((x) + 1) & ~1)
-#define ROUND_UP_4(x) (((x) + 3) & ~3)
-#define ROUND_UP_8(x) (((x) + 7) & ~7)
-
 /* taken from theora/lib/toplevel.c */
 static int
 _ilog (unsigned int v)
@@ -151,7 +150,7 @@ granulepos_to_timestamp (GstTheoraEnc * theoraenc, ogg_int64_t granulepos)
       theoraenc->info.fps_numerator);
 }
 
-static GstElementDetails theora_enc_details =
+static const GstElementDetails theora_enc_details =
 GST_ELEMENT_DETAILS ("Theora video encoder",
     "Codec/Encoder/Video",
     "encode raw YUV video to a theora stream",
@@ -293,6 +292,7 @@ gst_theora_enc_init (GstTheoraEnc * enc, GstTheoraEncClass * g_class)
   GST_DEBUG_OBJECT (enc,
       "keyframe_frequency_force is %d, granule shift is %d",
       enc->info.keyframe_frequency_force, enc->granule_shift);
+  enc->expected_ts = GST_CLOCK_TIME_NONE;
 }
 
 static void
@@ -308,6 +308,26 @@ theora_enc_finalize (GObject * object)
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static void
+theora_enc_reset (GstTheoraEnc * enc)
+{
+  theora_clear (&enc->state);
+  theora_encode_init (&enc->state, &enc->info);
+}
+
+static void
+theora_enc_clear (GstTheoraEnc * enc)
+{
+  enc->packetno = 0;
+  enc->bytes_out = 0;
+  enc->granulepos_offset = 0;
+  enc->timestamp_offset = 0;
+
+  enc->next_ts = GST_CLOCK_TIME_NONE;
+  enc->next_discont = FALSE;
+  enc->expected_ts = GST_CLOCK_TIME_NONE;
+}
+
 static gboolean
 theora_enc_sink_setcaps (GstPad * pad, GstCaps * caps)
 {
@@ -333,8 +353,8 @@ theora_enc_sink_setcaps (GstPad * pad, GstCaps * caps)
   /* center image if needed */
   if (enc->center) {
     /* make sure offset is even, for easier decoding */
-    enc->offset_x = ROUND_UP_2 ((enc->info_width - enc->width) / 2);
-    enc->offset_y = ROUND_UP_2 ((enc->info_height - enc->height) / 2);
+    enc->offset_x = GST_ROUND_UP_2 ((enc->info_width - enc->width) / 2);
+    enc->offset_y = GST_ROUND_UP_2 ((enc->info_height - enc->height) / 2);
   } else {
     enc->offset_x = 0;
     enc->offset_y = 0;
@@ -375,7 +395,7 @@ theora_enc_sink_setcaps (GstPad * pad, GstCaps * caps)
       "keyframe_frequency_force is %d, granule shift is %d",
       enc->info.keyframe_frequency_force, enc->granule_shift);
 
-  theora_encode_init (&enc->state, &enc->info);
+  theora_enc_reset (enc);
 
   gst_object_unref (enc);
 
@@ -419,6 +439,11 @@ theora_buffer_from_packet (GstTheoraEnc * enc, ogg_packet * packet,
   GST_BUFFER_TIMESTAMP (buf) = timestamp + enc->timestamp_offset;
   GST_BUFFER_DURATION (buf) = duration;
 
+  if (enc->next_discont) {
+    GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
+    enc->next_discont = FALSE;
+  }
+
   /* the second most significant bit of the first data byte is cleared
    * for keyframes */
   if ((packet->packet[0] & 0x40) == 0) {
@@ -544,6 +569,35 @@ theora_enc_sink_event (GstPad * pad, GstEvent * event)
   return res;
 }
 
+static gboolean
+theora_enc_is_discontinuous (GstTheoraEnc * enc, GstBuffer * buffer)
+{
+  GstClockTime ts = GST_BUFFER_TIMESTAMP (buffer);
+  GstClockTimeDiff max_diff;
+  gboolean ret = FALSE;
+
+  /* Allow 3/4 a frame off */
+  max_diff = (enc->info.fps_denominator * GST_SECOND * 3) /
+      (enc->info.fps_numerator * 4);
+
+  if (ts != GST_CLOCK_TIME_NONE && enc->expected_ts != GST_CLOCK_TIME_NONE) {
+    if ((GstClockTimeDiff) (ts - enc->expected_ts) > max_diff) {
+      GST_DEBUG_OBJECT (enc, "Incoming TS %" GST_TIME_FORMAT
+          " exceeds expected value %" GST_TIME_FORMAT
+          " by too much, marking discontinuity",
+          GST_TIME_ARGS (ts), GST_TIME_ARGS (enc->expected_ts));
+      ret = TRUE;
+    }
+  }
+
+  if (GST_CLOCK_TIME_IS_VALID (GST_BUFFER_DURATION (buffer)))
+    enc->expected_ts = ts + GST_BUFFER_DURATION (buffer);
+  else
+    enc->expected_ts = GST_CLOCK_TIME_NONE;
+
+  return ret;
+}
+
 static GstFlowReturn
 theora_enc_chain (GstPad * pad, GstBuffer * buffer)
 {
@@ -683,8 +737,8 @@ theora_enc_chain (GstPad * pad, GstBuffer * buffer)
       cheight = height / 2;
 
       /* source strides as defined in videotestsrc */
-      src_y_stride = ROUND_UP_4 (width);
-      src_uv_stride = ROUND_UP_8 (width) / 2;
+      src_y_stride = GST_ROUND_UP_4 (width);
+      src_uv_stride = GST_ROUND_UP_8 (width) / 2;
 
       /* destination strides from the real picture width */
       dst_y_stride = enc->info_width;
@@ -701,8 +755,8 @@ theora_enc_chain (GstPad * pad, GstBuffer * buffer)
       dest_v = yuv.v = yuv.u + y_size / 4;
 
       src_y = GST_BUFFER_DATA (buffer);
-      src_u = src_y + src_y_stride * ROUND_UP_2 (height);
-      src_v = src_u + src_uv_stride * ROUND_UP_2 (height) / 2;
+      src_u = src_y + src_y_stride * GST_ROUND_UP_2 (height);
+      src_v = src_u + src_uv_stride * GST_ROUND_UP_2 (height) / 2;
 
       if (enc->border != BORDER_NONE) {
         /* fill top border */
@@ -785,6 +839,16 @@ theora_enc_chain (GstPad * pad, GstBuffer * buffer)
       buffer = newbuf;
     }
 
+    if (theora_enc_is_discontinuous (enc, buffer)) {
+      theora_enc_reset (enc);
+      enc->granulepos_offset =
+          gst_util_uint64_scale (GST_BUFFER_TIMESTAMP (buffer), enc->fps_n,
+          GST_SECOND * enc->fps_d);
+      enc->timestamp_offset = GST_BUFFER_TIMESTAMP (buffer);
+      enc->next_ts = 0;
+      enc->next_discont = TRUE;
+    }
+
     res = theora_encode_YUVin (&enc->state, &yuv);
 
     ret = GST_FLOW_OK;
@@ -864,6 +928,8 @@ theora_enc_change_state (GstElement * element, GstStateChange transition)
       theora_clear (&enc->state);
       theora_comment_clear (&enc->comment);
       theora_info_clear (&enc->info);
+
+      theora_enc_clear (enc);
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
       break;
