@@ -65,8 +65,8 @@ GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("audio/x-raw-float, "
-        "rate = (int) [ 8000, 50000 ], "
-        "channels = (int) [ 1, 6 ], " "endianness = (int) BYTE_ORDER, "
+        "rate = (int) [ 1, MAX ], "
+        "channels = (int) [ 1, 256 ], " "endianness = (int) BYTE_ORDER, "
 /* no ifdef in macros, please
 #ifdef GST_VORBIS_DEC_SEQUENTIAL
       "layout = \"sequential\", "
@@ -87,6 +87,8 @@ GST_BOILERPLATE (GstVorbisDec, gst_vorbis_dec, GstElement, GST_TYPE_ELEMENT);
 static void vorbis_dec_finalize (GObject * object);
 static gboolean vorbis_dec_sink_event (GstPad * pad, GstEvent * event);
 static GstFlowReturn vorbis_dec_chain (GstPad * pad, GstBuffer * buffer);
+static GstFlowReturn vorbis_dec_chain_forward (GstVorbisDec * vd,
+    gboolean discont, GstBuffer * buffer);
 static GstStateChangeReturn vorbis_dec_change_state (GstElement * element,
     GstStateChange transition);
 
@@ -129,6 +131,8 @@ vorbis_get_query_types (GstPad * pad)
 {
   static const GstQueryType vorbis_dec_src_query_types[] = {
     GST_QUERY_POSITION,
+    GST_QUERY_DURATION,
+    GST_QUERY_CONVERT,
     0
   };
 
@@ -162,6 +166,8 @@ gst_vorbis_dec_init (GstVorbisDec * dec, GstVorbisDecClass * g_class)
   gst_element_add_pad (GST_ELEMENT (dec), dec->srcpad);
 
   dec->queued = NULL;
+  dec->pendingevents = NULL;
+  dec->taglist = NULL;
 }
 
 static void
@@ -180,6 +186,34 @@ vorbis_dec_finalize (GObject * object)
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+static void
+gst_vorbis_dec_reset (GstVorbisDec * dec)
+{
+  dec->cur_timestamp = GST_CLOCK_TIME_NONE;
+  dec->prev_timestamp = GST_CLOCK_TIME_NONE;
+  dec->granulepos = -1;
+  dec->discont = TRUE;
+  gst_segment_init (&dec->segment, GST_FORMAT_TIME);
+
+  g_list_foreach (dec->queued, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (dec->queued);
+  dec->queued = NULL;
+  g_list_foreach (dec->gather, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (dec->gather);
+  dec->gather = NULL;
+  g_list_foreach (dec->decode, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (dec->decode);
+  dec->decode = NULL;
+  g_list_foreach (dec->pendingevents, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (dec->pendingevents);
+  dec->pendingevents = NULL;
+
+  if (dec->taglist)
+    gst_tag_list_free (dec->taglist);
+  dec->taglist = NULL;
+}
+
+
 static gboolean
 vorbis_dec_convert (GstPad * pad,
     GstFormat src_format, gint64 src_value,
@@ -189,19 +223,19 @@ vorbis_dec_convert (GstPad * pad,
   GstVorbisDec *dec;
   guint64 scale = 1;
 
-  dec = GST_VORBIS_DEC (GST_PAD_PARENT (pad));
-
-  if (dec->packetno < 1)
-    return FALSE;
-
   if (src_format == *dest_format) {
     *dest_value = src_value;
     return TRUE;
   }
 
+  dec = GST_VORBIS_DEC (gst_pad_get_parent (pad));
+
+  if (!dec->initialized)
+    goto no_header;
+
   if (dec->sinkpad == pad &&
       (src_format == GST_FORMAT_BYTES || *dest_format == GST_FORMAT_BYTES))
-    return FALSE;
+    goto no_format;
 
   switch (src_format) {
     case GST_FORMAT_TIME:
@@ -246,55 +280,86 @@ vorbis_dec_convert (GstPad * pad,
     default:
       res = FALSE;
   }
+done:
+  gst_object_unref (dec);
 
   return res;
+
+  /* ERRORS */
+no_header:
+  {
+    GST_DEBUG_OBJECT (dec, "no header packets received");
+    res = FALSE;
+    goto done;
+  }
+no_format:
+  {
+    GST_DEBUG_OBJECT (dec, "formats unsupported");
+    res = FALSE;
+    goto done;
+  }
 }
 
 static gboolean
 vorbis_dec_src_query (GstPad * pad, GstQuery * query)
 {
-  gint64 granulepos;
   GstVorbisDec *dec;
   gboolean res = FALSE;
 
-  dec = GST_VORBIS_DEC (GST_PAD_PARENT (pad));
+  dec = GST_VORBIS_DEC (gst_pad_get_parent (pad));
 
   switch (GST_QUERY_TYPE (query)) {
     case GST_QUERY_POSITION:
     {
-      GstFormat format;
-      gint64 value;
+      gint64 granulepos, value;
+      GstFormat my_format, format;
+      gint64 time;
 
+      /* we start from the last seen granulepos */
       granulepos = dec->granulepos;
 
       gst_query_parse_position (query, &format, NULL);
 
-      /* and convert to the final format */
+      /* and convert to the final format in two steps with time as the
+       * intermediate step */
+      my_format = GST_FORMAT_TIME;
       if (!(res =
-              vorbis_dec_convert (pad, GST_FORMAT_DEFAULT, granulepos, &format,
-                  &value)))
+              vorbis_dec_convert (pad, GST_FORMAT_DEFAULT, granulepos,
+                  &my_format, &time)))
         goto error;
 
-      value = (value - dec->segment.start) + dec->segment.time;
+      /* correct for the segment values */
+      time = gst_segment_to_stream_time (&dec->segment, GST_FORMAT_TIME, time);
+
+      GST_LOG_OBJECT (dec,
+          "query %p: our time: %" GST_TIME_FORMAT, query, GST_TIME_ARGS (time));
+
+      /* and convert to the final format */
+      if (!(res = vorbis_dec_convert (pad, my_format, time, &format, &value)))
+        goto error;
 
       gst_query_set_position (query, format, value);
 
       GST_LOG_OBJECT (dec,
-          "query %u: peer returned granulepos: %llu - we return %llu (format %u)",
-          query, granulepos, value, format);
+          "query %p: we return %lld (format %u)", query, value, format);
 
       break;
     }
     case GST_QUERY_DURATION:
     {
-      /* query peer for total length */
-      if (!gst_pad_is_linked (dec->sinkpad)) {
+      GstPad *peer;
+
+      if (!(peer = gst_pad_get_peer (dec->sinkpad))) {
         GST_WARNING_OBJECT (dec, "sink pad %" GST_PTR_FORMAT " is not linked",
             dec->sinkpad);
         goto error;
       }
-      if (!(res = gst_pad_query (GST_PAD_PEER (dec->sinkpad), query)))
+
+      res = gst_pad_query (peer, query);
+      gst_object_unref (peer);
+      if (!res)
         goto error;
+
       break;
     }
     case GST_QUERY_CONVERT:
@@ -313,12 +378,16 @@ vorbis_dec_src_query (GstPad * pad, GstQuery * query)
       res = gst_pad_query_default (pad, query);
       break;
   }
+done:
+  gst_object_unref (dec);
+
   return res;
 
+  /* ERRORS */
 error:
   {
     GST_WARNING_OBJECT (dec, "error handling query");
-    return res;
+    goto done;
   }
 }
 
@@ -328,7 +397,7 @@ vorbis_dec_sink_query (GstPad * pad, GstQuery * query)
   GstVorbisDec *dec;
   gboolean res;
 
-  dec = GST_VORBIS_DEC (GST_PAD_PARENT (pad));
+  dec = GST_VORBIS_DEC (gst_pad_get_parent (pad));
 
   switch (GST_QUERY_TYPE (query)) {
     case GST_QUERY_CONVERT:
@@ -347,18 +416,31 @@ vorbis_dec_sink_query (GstPad * pad, GstQuery * query)
       res = gst_pad_query_default (pad, query);
       break;
   }
-error:
+
+done:
+  gst_object_unref (dec);
+
   return res;
+
+  /* ERRORS */
+error:
+  {
+    GST_DEBUG_OBJECT (dec, "error converting value");
+    goto done;
+  }
 }
 
 static gboolean
 vorbis_dec_src_event (GstPad * pad, GstEvent * event)
 {
   gboolean res = TRUE;
-  GstVorbisDec *dec = GST_VORBIS_DEC (GST_PAD_PARENT (pad));
+  GstVorbisDec *dec;
+
+  dec = GST_VORBIS_DEC (gst_pad_get_parent (pad));
 
   switch (GST_EVENT_TYPE (event)) {
-    case GST_EVENT_SEEK:{
+    case GST_EVENT_SEEK:
+    {
       GstFormat format, tformat;
       gdouble rate;
       GstEvent *real_seek;
@@ -369,6 +451,7 @@ vorbis_dec_src_event (GstPad * pad, GstEvent * event)
 
       gst_event_parse_seek (event, &rate, &format, &flags, &cur_type, &cur,
           &stop_type, &stop);
+      gst_event_unref (event);
 
       /* we have to ask our peer to seek to time here as we know
        * nothing about how to generate a granulepos from the src
@@ -378,9 +461,9 @@ vorbis_dec_src_event (GstPad * pad, GstEvent * event)
        */
       tformat = GST_FORMAT_TIME;
       if (!(res = vorbis_dec_convert (pad, format, cur, &tformat, &tcur)))
-        goto error;
+        goto convert_error;
       if (!(res = vorbis_dec_convert (pad, format, stop, &tformat, &tstop)))
-        goto error;
+        goto convert_error;
 
       /* then seek with time on the peer */
       real_seek = gst_event_new_seek (rate, GST_FORMAT_TIME,
@@ -388,19 +471,23 @@ vorbis_dec_src_event (GstPad * pad, GstEvent * event)
 
       res = gst_pad_push_event (dec->sinkpad, real_seek);
 
-      gst_event_unref (event);
       break;
     }
     default:
-      res = gst_pad_event_default (pad, event);
+      res = gst_pad_push_event (dec->sinkpad, event);
       break;
   }
+done:
+  gst_object_unref (dec);
 
   return res;
 
-error:
-  gst_event_unref (event);
-  return res;
+  /* ERRORS */
+convert_error:
+  {
+    GST_DEBUG_OBJECT (dec, "cannot convert start/stop for seek");
+    goto done;
+  }
 }
 
 static gboolean
@@ -416,29 +503,49 @@ vorbis_dec_sink_event (GstPad * pad, GstEvent * event)
     case GST_EVENT_EOS:
       ret = gst_pad_push_event (dec->srcpad, event);
       break;
+    case GST_EVENT_FLUSH_START:
+      ret = gst_pad_push_event (dec->srcpad, event);
+      break;
+    case GST_EVENT_FLUSH_STOP:
+      /* here we must clean any state in the decoder */
+#ifdef HAVE_VORBIS_SYNTHESIS_RESTART
+      vorbis_synthesis_restart (&dec->vd);
+#endif
+      gst_vorbis_dec_reset (dec);
+      ret = gst_pad_push_event (dec->srcpad, event);
+      break;
     case GST_EVENT_NEWSEGMENT:
     {
       GstFormat format;
-      gdouble rate;
+      gdouble rate, arate;
       gint64 start, stop, time;
       gboolean update;
 
-      gst_event_parse_new_segment (event, &update, &rate, &format, &start,
-          &stop, &time);
+      gst_event_parse_new_segment_full (event, &update, &rate, &arate, &format,
+          &start, &stop, &time);
 
-      /* we need time and a positive rate for now */
+      /* we need time for now */
       if (format != GST_FORMAT_TIME)
         goto newseg_wrong_format;
 
-      if (rate <= 0.0)
-        goto newseg_wrong_rate;
+      GST_DEBUG_OBJECT (dec,
+          "newsegment: update %d, rate %g, arate %g, start %" GST_TIME_FORMAT
+          ", stop %" GST_TIME_FORMAT ", time %" GST_TIME_FORMAT,
+          update, rate, arate, GST_TIME_ARGS (start), GST_TIME_ARGS (stop),
+          GST_TIME_ARGS (time));
 
       /* now configure the values */
-      gst_segment_set_newsegment (&dec->segment, update,
-          rate, format, start, stop, time);
+      gst_segment_set_newsegment_full (&dec->segment, update,
+          rate, arate, format, start, stop, time);
 
-      /* and forward */
-      ret = gst_pad_push_event (dec->srcpad, event);
+      if (dec->initialized)
+        /* and forward */
+        ret = gst_pad_push_event (dec->srcpad, event);
+      else {
+        /* store it to send once we're initialized */
+        dec->pendingevents = g_list_append (dec->pendingevents, event);
+        ret = TRUE;
+      }
       break;
     }
     default:
@@ -449,18 +556,13 @@ done:
   gst_object_unref (dec);
 
   return ret;
+
   /* ERRORS */
 newseg_wrong_format:
   {
-    GST_DEBUG ("received non TIME newsegment");
+    GST_DEBUG_OBJECT (dec, "received non TIME newsegment");
     goto done;
   }
-newseg_wrong_rate:
-  {
-    GST_DEBUG ("negative rates not supported yet");
-    goto done;
-  }
-
 }
 
 static GstFlowReturn
@@ -469,18 +571,13 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
   GstCaps *caps;
   const GstAudioChannelPosition *pos = NULL;
 
-  caps = gst_caps_new_simple ("audio/x-raw-float",
-      "rate", G_TYPE_INT, vd->vi.rate,
-      "channels", G_TYPE_INT, vd->vi.channels,
-      "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32, NULL);
-
   switch (vd->vi.channels) {
     case 1:
     case 2:
       /* nothing */
       break;
     case 3:{
-      static GstAudioChannelPosition pos3[] = {
+      static const GstAudioChannelPosition pos3[] = {
         GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
         GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
         GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT
@@ -489,7 +586,7 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
       break;
     }
     case 4:{
-      static GstAudioChannelPosition pos4[] = {
+      static const GstAudioChannelPosition pos4[] = {
         GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
         GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
         GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
@@ -499,7 +596,7 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
       break;
     }
     case 5:{
-      static GstAudioChannelPosition pos5[] = {
+      static const GstAudioChannelPosition pos5[] = {
         GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
         GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
         GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
@@ -510,7 +607,7 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
       break;
     }
     case 6:{
-      static GstAudioChannelPosition pos6[] = {
+      static const GstAudioChannelPosition pos6[] = {
         GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
         GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
         GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
@@ -524,6 +621,12 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
     default:
       goto channel_count_error;
   }
+
+  caps = gst_caps_new_simple ("audio/x-raw-float",
+      "rate", G_TYPE_INT, vd->vi.rate,
+      "channels", G_TYPE_INT, vd->vi.channels,
+      "endianness", G_TYPE_INT, G_BYTE_ORDER, "width", G_TYPE_INT, 32, NULL);
+
   if (pos) {
     gst_audio_set_channel_positions (gst_caps_get_structure (caps, 0), pos);
   }
@@ -535,7 +638,6 @@ vorbis_handle_identification_packet (GstVorbisDec * vd)
   /* ERROR */
 channel_count_error:
   {
-    gst_caps_unref (caps);
     GST_ELEMENT_ERROR (vd, STREAM, NOT_IMPLEMENTED, (NULL),
         ("Unsupported channel count %d", vd->vi.channels));
     return GST_FLOW_ERROR;
@@ -559,43 +661,55 @@ vorbis_handle_comment_packet (GstVorbisDec * vd, ogg_packet * packet)
       gst_tag_list_from_vorbiscomment_buffer (buf, (guint8 *) "\003vorbis", 7,
       &encoder);
 
+  vd->taglist = gst_tag_list_merge (vd->taglist, list, GST_TAG_MERGE_REPLACE);
+
+  gst_tag_list_free (list);
   gst_buffer_unref (buf);
 
-  if (!list) {
+  if (!vd->taglist) {
     GST_ERROR_OBJECT (vd, "couldn't decode comments");
-    list = gst_tag_list_new ();
+    vd->taglist = gst_tag_list_new ();
   }
   if (encoder) {
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
         GST_TAG_ENCODER, encoder, NULL);
     g_free (encoder);
   }
-  gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+  gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
       GST_TAG_ENCODER_VERSION, vd->vi.version,
       GST_TAG_AUDIO_CODEC, "Vorbis", NULL);
   if (vd->vi.bitrate_nominal > 0) {
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
         GST_TAG_NOMINAL_BITRATE, (guint) vd->vi.bitrate_nominal, NULL);
     bitrate = vd->vi.bitrate_nominal;
   }
   if (vd->vi.bitrate_upper > 0) {
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
         GST_TAG_MAXIMUM_BITRATE, (guint) vd->vi.bitrate_upper, NULL);
     if (!bitrate)
       bitrate = vd->vi.bitrate_upper;
   }
   if (vd->vi.bitrate_lower > 0) {
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
         GST_TAG_MINIMUM_BITRATE, (guint) vd->vi.bitrate_lower, NULL);
     if (!bitrate)
       bitrate = vd->vi.bitrate_lower;
   }
   if (bitrate) {
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (vd->taglist, GST_TAG_MERGE_REPLACE,
         GST_TAG_BITRATE, (guint) bitrate, NULL);
   }
 
-  gst_element_found_tags_for_pad (GST_ELEMENT (vd), vd->srcpad, list);
+  if (vd->initialized) {
+    gst_element_found_tags_for_pad (GST_ELEMENT_CAST (vd), vd->srcpad,
+        vd->taglist);
+    vd->taglist = NULL;
+  } else {
+    /* Only post them as messages for the time being. *
+     * They will be pushed on the pad once the decoder is initialized */
+    gst_element_post_message (GST_ELEMENT_CAST (vd),
+        gst_message_new_tag (GST_OBJECT (vd), gst_tag_list_copy (vd->taglist)));
+  }
 
   return GST_FLOW_OK;
 }
@@ -603,12 +717,27 @@ vorbis_handle_comment_packet (GstVorbisDec * vd, ogg_packet * packet)
 static GstFlowReturn
 vorbis_handle_type_packet (GstVorbisDec * vd)
 {
+  GList *walk;
+
   g_assert (vd->initialized == FALSE);
 
   vorbis_synthesis_init (&vd->vd, &vd->vi);
   vorbis_block_init (&vd->vd, &vd->vb);
   vd->initialized = TRUE;
 
+  if (vd->pendingevents) {
+    for (walk = vd->pendingevents; walk; walk = g_list_next (walk))
+      gst_pad_push_event (vd->srcpad, GST_EVENT_CAST (walk->data));
+    g_list_free (vd->pendingevents);
+    vd->pendingevents = NULL;
+  }
+
+  if (vd->taglist) {
+    /* The tags have already been sent on the bus as messages. */
+    gst_pad_push_event (vd->srcpad, gst_event_new_tag (vd->taglist));
+    vd->taglist = NULL;
+  }
+
   return GST_FLOW_OK;
 }
 
@@ -673,8 +802,60 @@ copy_samples (float *out, float **in, guint samples, gint channels)
 #endif
 }
 
+/* clip output samples to the segment boundaries
+ */
+static gboolean
+vorbis_do_clip (GstVorbisDec * dec, GstBuffer * buf)
+{
+  gint64 start, stop, cstart, cstop, diff;
+
+  start = GST_BUFFER_TIMESTAMP (buf);
+  stop = start + GST_BUFFER_DURATION (buf);
+
+  if (!gst_segment_clip (&dec->segment, GST_FORMAT_TIME,
+          start, stop, &cstart, &cstop))
+    goto clipped;
+
+  /* see if some clipping happened */
+  diff = cstart - start;
+  if (diff > 0) {
+    GST_BUFFER_TIMESTAMP (buf) = cstart;
+    GST_BUFFER_DURATION (buf) -= diff;
+
+    /* bring clipped time to samples */
+    diff = gst_util_uint64_scale_int (diff, dec->vi.rate, GST_SECOND);
+    /* samples to bytes */
+    diff *= (sizeof (float) * dec->vi.channels);
+    GST_DEBUG_OBJECT (dec, "clipping start to %" GST_TIME_FORMAT " %"
+        G_GUINT64_FORMAT " bytes", GST_TIME_ARGS (cstart), diff);
+    GST_BUFFER_DATA (buf) += diff;
+    GST_BUFFER_SIZE (buf) -= diff;
+  }
+  diff = stop - cstop;
+  if (diff > 0) {
+    GST_BUFFER_DURATION (buf) -= diff;
+
+    /* bring clipped time to samples and then to bytes */
+    diff = gst_util_uint64_scale_int (diff, dec->vi.rate, GST_SECOND);
+    diff *= (sizeof (float) * dec->vi.channels);
+    GST_DEBUG_OBJECT (dec, "clipping stop to %" GST_TIME_FORMAT " %"
+        G_GUINT64_FORMAT " bytes", GST_TIME_ARGS (cstop), diff);
+    GST_BUFFER_SIZE (buf) -= diff;
+  }
+
+  return FALSE;
+
+  /* dropped buffer */
+clipped:
+  {
+    GST_DEBUG_OBJECT (dec, "clipped buffer");
+    gst_buffer_unref (buf);
+    return TRUE;
+  }
+}
+
 static GstFlowReturn
-vorbis_dec_push (GstVorbisDec * dec, GstBuffer * buf)
+vorbis_dec_push_forward (GstVorbisDec * dec, GstBuffer * buf)
 {
   GstFlowReturn result;
   gint64 outoffset = GST_BUFFER_OFFSET (buf);
@@ -684,7 +865,7 @@ vorbis_dec_push (GstVorbisDec * dec, GstBuffer * buf)
     GST_DEBUG_OBJECT (dec, "queued buffer");
     result = GST_FLOW_OK;
   } else {
-    if (dec->queued) {
+    if (G_UNLIKELY (dec->queued)) {
       gint64 size;
       GList *walk;
 
@@ -708,26 +889,47 @@ vorbis_dec_push (GstVorbisDec * dec, GstBuffer * buf)
       for (walk = dec->queued; walk; walk = g_list_next (walk)) {
         GstBuffer *buffer = GST_BUFFER (walk->data);
 
-        /* ignore the result */
+        /* clips or returns FALSE with buffer unreffed when completely
+         * clipped */
+        if (vorbis_do_clip (dec, buffer))
+          continue;
+
         if (dec->discont) {
           GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DISCONT);
           dec->discont = FALSE;
         }
+        /* ignore the result */
         gst_pad_push (dec->srcpad, buffer);
       }
       g_list_free (dec->queued);
       dec->queued = NULL;
     }
+
+    /* clip */
+    if (vorbis_do_clip (dec, buf))
+      return GST_FLOW_OK;
+
     if (dec->discont) {
       GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
       dec->discont = FALSE;
     }
+
     result = gst_pad_push (dec->srcpad, buf);
   }
 
   return result;
 }
 
+static GstFlowReturn
+vorbis_dec_push_reverse (GstVorbisDec * dec, GstBuffer * buf)
+{
+  GstFlowReturn result = GST_FLOW_OK;
+
+  dec->queued = g_list_prepend (dec->queued, buf);
+
+  return result;
+}
+
 static GstFlowReturn
 vorbis_handle_data_packet (GstVorbisDec * vd, ogg_packet * packet)
 {
@@ -735,16 +937,30 @@ vorbis_handle_data_packet (GstVorbisDec * vd, ogg_packet * packet)
   guint sample_count;
   GstBuffer *out;
   GstFlowReturn result;
+  GstClockTime timestamp = -1;
   gint size;
 
   if (!vd->initialized)
     goto not_initialized;
 
+  /* FIXME, we should queue undecoded packets here until we get
+   * a timestamp, then we reverse timestamp the queued packets and
+   * clip them, then we decode only the ones we want and don't
+   * keep decoded data in memory.
+   * Ideally, of course, the demuxer gives us a valid timestamp on
+   * the first packet.
+   */
+
   /* normal data packet */
-  if (vorbis_synthesis (&vd->vb, packet))
+  /* FIXME, we can skip decoding if the packet is outside of the
+   * segment, this is however not very trivial as we need a previous
+   * packet to decode the current one so we must be carefull not to
+   * throw away too much. For now we decode everything and clip right
+   * before pushing data. */
+  if (G_UNLIKELY (vorbis_synthesis (&vd->vb, packet)))
     goto could_not_read;
 
-  if (vorbis_synthesis_blockin (&vd->vd, &vd->vb) < 0)
+  if (G_UNLIKELY (vorbis_synthesis_blockin (&vd->vd, &vd->vb) < 0))
     goto not_accepted;
 
   /* assume all goes well here */
@@ -760,11 +976,11 @@ vorbis_handle_data_packet (GstVorbisDec * vd, ogg_packet * packet)
   result =
       gst_pad_alloc_buffer_and_set_caps (vd->srcpad, GST_BUFFER_OFFSET_NONE,
       size, GST_PAD_CAPS (vd->srcpad), &out);
-  if (result != GST_FLOW_OK)
+  if (G_UNLIKELY (result != GST_FLOW_OK))
     goto done;
 
   /* get samples ready for reading now, should be sample_count */
-  if ((vorbis_synthesis_pcmout (&vd->vd, &pcm)) != sample_count)
+  if (G_UNLIKELY ((vorbis_synthesis_pcmout (&vd->vd, &pcm)) != sample_count))
     goto wrong_samples;
 
   /* copy samples in buffer */
@@ -772,20 +988,18 @@ vorbis_handle_data_packet (GstVorbisDec * vd, ogg_packet * packet)
       vd->vi.channels);
 
   GST_BUFFER_SIZE (out) = size;
-  GST_BUFFER_OFFSET (out) = vd->granulepos;
-  if (vd->granulepos != -1) {
-    GST_BUFFER_OFFSET_END (out) = vd->granulepos + sample_count;
-    GST_BUFFER_TIMESTAMP (out) =
-        gst_util_uint64_scale_int (vd->granulepos, GST_SECOND, vd->vi.rate);
-  } else {
-    GST_BUFFER_TIMESTAMP (out) = -1;
-  }
+
   /* this should not overflow */
   GST_BUFFER_DURATION (out) = sample_count * GST_SECOND / vd->vi.rate;
 
+  if (packet->granulepos != -1)
+    vd->granulepos = packet->granulepos - sample_count;
+
   if (vd->cur_timestamp != GST_CLOCK_TIME_NONE) {
-    GST_BUFFER_TIMESTAMP (out) = vd->cur_timestamp;
-    GST_DEBUG ("cur_timestamp: %" GST_TIME_FORMAT " + %" GST_TIME_FORMAT " = % "
+    /* we have incomming timestamps */
+    timestamp = vd->cur_timestamp;
+    GST_DEBUG_OBJECT (vd,
+        "cur_timestamp: %" GST_TIME_FORMAT " + %" GST_TIME_FORMAT " = %"
         GST_TIME_FORMAT, GST_TIME_ARGS (vd->cur_timestamp),
         GST_TIME_ARGS (GST_BUFFER_DURATION (out)),
         GST_TIME_ARGS (vd->cur_timestamp + GST_BUFFER_DURATION (out)));
@@ -793,16 +1007,34 @@ vorbis_handle_data_packet (GstVorbisDec * vd, ogg_packet * packet)
     GST_BUFFER_OFFSET (out) = GST_CLOCK_TIME_TO_FRAMES (vd->cur_timestamp,
         vd->vi.rate);
     GST_BUFFER_OFFSET_END (out) = GST_BUFFER_OFFSET (out) + sample_count;
+  } else {
+    /* we have incomming granulepos */
+    GST_BUFFER_OFFSET (out) = vd->granulepos;
+    if (vd->granulepos != -1) {
+      GST_BUFFER_OFFSET_END (out) = vd->granulepos + sample_count;
+      timestamp =
+          gst_util_uint64_scale_int (vd->granulepos, GST_SECOND, vd->vi.rate);
+    } else {
+      timestamp = -1;
+    }
   }
+  GST_BUFFER_TIMESTAMP (out) = timestamp;
 
   if (vd->granulepos != -1)
     vd->granulepos += sample_count;
 
-  result = vorbis_dec_push (vd, out);
+  if (vd->segment.rate >= 0.0)
+    result = vorbis_dec_push_forward (vd, out);
+  else
+    result = vorbis_dec_push_reverse (vd, out);
 
 done:
   vorbis_synthesis_read (&vd->vd, sample_count);
 
+  GST_DEBUG_OBJECT (vd,
+      "decoded %d bytes into %d samples, ts %" GST_TIME_FORMAT, packet->bytes,
+      sample_count, GST_TIME_ARGS (timestamp));
+
   /* granulepos is the last sample in the packet */
   if (packet->granulepos != -1)
     vd->granulepos = packet->granulepos;
@@ -813,7 +1045,7 @@ done:
 not_initialized:
   {
     GST_ELEMENT_ERROR (GST_ELEMENT (vd), STREAM, DECODE,
-        (NULL), ("no header sent yet (packet no is %d)", packet->packetno));
+        (NULL), ("no header sent yet"));
     return GST_FLOW_ERROR;
   }
 could_not_read:
@@ -838,34 +1070,23 @@ wrong_samples:
 }
 
 static GstFlowReturn
-vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
+vorbis_dec_decode_buffer (GstVorbisDec * vd, GstBuffer * buffer)
 {
-  GstVorbisDec *vd;
   ogg_packet packet;
   GstFlowReturn result = GST_FLOW_OK;
+  GstClockTime timestamp;
+  guint64 offset_end;
 
-  vd = GST_VORBIS_DEC (gst_pad_get_parent (pad));
-
-  /* resync on DISCONT */
-  if (G_UNLIKELY (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DISCONT))) {
-    GST_DEBUG_OBJECT (vd, "received DISCONT buffer");
-    vd->granulepos = -1;
-    vd->cur_timestamp = GST_CLOCK_TIME_NONE;
-    vd->prev_timestamp = GST_CLOCK_TIME_NONE;
-#ifdef HAVE_VORBIS_SYNTHESIS_RESTART
-    vorbis_synthesis_restart (&vd->vd);
-#endif
-    vd->discont = TRUE;
-  }
+  timestamp = GST_BUFFER_TIMESTAMP (buffer);
+  offset_end = GST_BUFFER_OFFSET_END (buffer);
 
   /* only ogg has granulepos, demuxers of other container formats 
    * might provide us with timestamps instead (e.g. matroskademux) */
-  if (GST_BUFFER_OFFSET_END (buffer) == GST_BUFFER_OFFSET_NONE &&
-      GST_BUFFER_TIMESTAMP (buffer) != GST_CLOCK_TIME_NONE) {
+  if (offset_end == GST_BUFFER_OFFSET_NONE && timestamp != GST_CLOCK_TIME_NONE) {
     /* we might get multiple consecutive buffers with the same timestamp */
-    if (GST_BUFFER_TIMESTAMP (buffer) != vd->prev_timestamp) {
-      vd->cur_timestamp = GST_BUFFER_TIMESTAMP (buffer);
-      vd->prev_timestamp = GST_BUFFER_TIMESTAMP (buffer);
+    if (timestamp != vd->prev_timestamp) {
+      vd->cur_timestamp = timestamp;
+      vd->prev_timestamp = timestamp;
     }
   } else {
     vd->cur_timestamp = GST_CLOCK_TIME_NONE;
@@ -875,8 +1096,8 @@ vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
   /* make ogg_packet out of the buffer */
   packet.packet = GST_BUFFER_DATA (buffer);
   packet.bytes = GST_BUFFER_SIZE (buffer);
-  packet.granulepos = GST_BUFFER_OFFSET_END (buffer);
-  packet.packetno = vd->packetno++;
+  packet.granulepos = offset_end;
+  packet.packetno = 0;          /* we don't care */
   /*
    * FIXME. Is there anyway to know that this is the last packet and
    * set e_o_s??
@@ -885,7 +1106,7 @@ vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
    */
   packet.e_o_s = 0;
 
-  if (packet.bytes < 1)
+  if (G_UNLIKELY (packet.bytes < 1))
     goto wrong_size;
 
   GST_DEBUG_OBJECT (vd, "vorbis granule: %" G_GINT64_FORMAT,
@@ -894,7 +1115,7 @@ vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
   /* switch depending on packet type */
   if (packet.packet[0] & 1) {
     if (vd->initialized) {
-      GST_WARNING_OBJECT (vd, "Ignoring header");
+      GST_WARNING_OBJECT (vd, "Already initialized, so ignoring header packet");
       goto done;
     }
     result = vorbis_handle_header_packet (vd, &packet);
@@ -902,13 +1123,7 @@ vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
     result = vorbis_handle_data_packet (vd, &packet);
   }
 
-  GST_DEBUG_OBJECT (vd, "offset end: %" G_GINT64_FORMAT,
-      (gint64) GST_BUFFER_OFFSET_END (buffer));
-
 done:
-  gst_buffer_unref (buffer);
-  gst_object_unref (vd);
-
   return result;
 
   /* ERRORS */
@@ -921,13 +1136,236 @@ wrong_size:
   }
 }
 
+/* 
+ * Input:
+ *  Buffer decoding order:  7  8  9  4  5  6  3  1  2  EOS
+ *  Discont flag:           D        D        D  D
+ *
+ * - Each Discont marks a discont in the decoding order.
+ *
+ * for vorbis, each buffer is a keyframe when we have the previous
+ * buffer. This means that to decode buffer 7, we need buffer 6, which
+ * arrives out of order.
+ *
+ * we first gather buffers in the gather queue until we get a DISCONT. We 
+ * prepend each incomming buffer so that they are in reversed order.
+ *   
+ *    gather queue:    9  8  7
+ *    decode queue:    
+ *    output queue:    
+ *
+ * When a DISCONT is received (buffer 4), we move the gather queue to the 
+ * decode queue. This simply done be taking the head of the gather queue
+ * and prepending it to the decode queue. This yields:
+ * 
+ *    gather queue:    
+ *    decode queue:    7  8  9
+ *    output queue:    
+ *
+ * Then we decode each buffer in the decode queue in order and put the output
+ * buffer in the output queue. The first buffer (7) will not produce and output
+ * because it needs the previous buffer (6) which did not arrive yet. This
+ * yields:
+ *
+ *    gather queue:    
+ *    decode queue:    7  8  9
+ *    output queue:    8  9
+ *
+ * Then we remove the consumed buffers from the decode queue. Buffer 7 is not
+ * completely consumed, we need to keep it around for when we receive buffer 
+ * 6. This yields:
+ *
+ *    gather queue:    
+ *    decode queue:    7 
+ *    output queue:    9  8
+ *
+ * Then we accumulate more buffers:
+ *
+ *    gather queue:    6  5  4
+ *    decode queue:    7
+ *    output queue:    
+ *
+ * prepending to the decode queue on DISCONT yields:
+ *
+ *    gather queue:   
+ *    decode queue:    4  5  6  7
+ *    output queue:    
+ *
+ * after decoding and keeping buffer 4:
+ *
+ *    gather queue:   
+ *    decode queue:    4 
+ *    output queue:    7  6  5 
+ *
+ * Etc..
+ */
+static GstFlowReturn
+vorbis_dec_flush_decode (GstVorbisDec * dec)
+{
+  GstFlowReturn res = GST_FLOW_OK;
+  GList *walk;
+
+  walk = dec->decode;
+
+  GST_DEBUG_OBJECT (dec, "flushing buffers to decoder");
+
+  while (walk) {
+    GList *next;
+    GstBuffer *buf = GST_BUFFER_CAST (walk->data);
+
+    GST_DEBUG_OBJECT (dec, "decoding buffer %p, ts %" GST_TIME_FORMAT,
+        buf, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
+
+    next = g_list_next (walk);
+
+    /* decode buffer, prepend to output queue */
+    res = vorbis_dec_decode_buffer (dec, buf);
+
+    /* if we generated output, we can discard the buffer, else we
+     * keep it in the queue */
+    if (dec->queued) {
+      GST_DEBUG_OBJECT (dec, "decoded buffer to %p", dec->queued->data);
+      dec->decode = g_list_delete_link (dec->decode, walk);
+      gst_buffer_unref (buf);
+    } else {
+      GST_DEBUG_OBJECT (dec, "buffer did not decode, keeping");
+    }
+    walk = next;
+  }
+  if (dec->granulepos != -1) {
+    GstClockTime endts;
+
+    endts =
+        gst_util_uint64_scale_int (dec->granulepos, GST_SECOND, dec->vi.rate);
+
+    GST_DEBUG_OBJECT (dec, "we have granulepos %" G_GUINT64_FORMAT ", ts %"
+        GST_TIME_FORMAT, dec->granulepos, GST_TIME_ARGS (endts));
+
+    while (dec->queued) {
+      GstBuffer *buf;
+      guint sample_count;
+
+      buf = GST_BUFFER_CAST (dec->queued->data);
+
+      sample_count =
+          GST_BUFFER_SIZE (buf) / (dec->vi.channels * sizeof (float));
+
+      GST_BUFFER_OFFSET_END (buf) = dec->granulepos;
+      endts =
+          gst_util_uint64_scale_int (dec->granulepos, GST_SECOND, dec->vi.rate);
+      dec->granulepos -= sample_count;
+      GST_BUFFER_OFFSET (buf) = dec->granulepos;
+      GST_BUFFER_TIMESTAMP (buf) =
+          gst_util_uint64_scale_int (dec->granulepos, GST_SECOND, dec->vi.rate);
+      GST_BUFFER_DURATION (buf) = endts - GST_BUFFER_TIMESTAMP (buf);
+
+      /* clip, this will unref the buffer in case of clipping */
+      if (vorbis_do_clip (dec, buf)) {
+        GST_DEBUG_OBJECT (dec, "clipped buffer %p", buf);
+        goto next;
+      }
+
+      if (dec->discont) {
+        GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_DISCONT);
+        dec->discont = FALSE;
+      }
+      GST_DEBUG_OBJECT (dec, "pushing buffer %p, samples %u, "
+          "ts %" GST_TIME_FORMAT ", dur %" GST_TIME_FORMAT,
+          buf, sample_count, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)),
+          GST_TIME_ARGS (GST_BUFFER_DURATION (buf)));
+
+      res = gst_pad_push (dec->srcpad, buf);
+    next:
+      dec->queued = g_list_delete_link (dec->queued, dec->queued);
+    }
+  } else {
+    GST_DEBUG_OBJECT (dec, "we don't have a granulepos yet, delayed push");
+  }
+  return res;
+}
+
+static GstFlowReturn
+vorbis_dec_chain_reverse (GstVorbisDec * vd, gboolean discont, GstBuffer * buf)
+{
+  GstFlowReturn result = GST_FLOW_OK;
+
+  /* if we have a discont, move buffers to the decode list */
+  if (G_UNLIKELY (discont)) {
+    GST_DEBUG_OBJECT (vd, "received discont");
+    while (vd->gather) {
+      GstBuffer *gbuf;
+
+      gbuf = GST_BUFFER_CAST (vd->gather->data);
+      /* remove from the gather list */
+      vd->gather = g_list_delete_link (vd->gather, vd->gather);
+      /* copy to decode queue */
+      vd->decode = g_list_prepend (vd->decode, gbuf);
+    }
+    /* flush and decode the decode queue */
+    result = vorbis_dec_flush_decode (vd);
+  }
+
+  GST_DEBUG_OBJECT (vd, "gathering buffer %p, size %u", buf,
+      GST_BUFFER_SIZE (buf));
+  /* add buffer to gather queue */
+  vd->gather = g_list_prepend (vd->gather, buf);
+
+  return result;
+}
+
+static GstFlowReturn
+vorbis_dec_chain_forward (GstVorbisDec * vd, gboolean discont,
+    GstBuffer * buffer)
+{
+  GstFlowReturn result;
+
+  result = vorbis_dec_decode_buffer (vd, buffer);
+
+  gst_buffer_unref (buffer);
+
+  return result;
+
+}
+
+static GstFlowReturn
+vorbis_dec_chain (GstPad * pad, GstBuffer * buffer)
+{
+  GstVorbisDec *vd;
+  GstFlowReturn result = GST_FLOW_OK;
+  gboolean discont;
+
+  vd = GST_VORBIS_DEC (gst_pad_get_parent (pad));
+
+  discont = GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DISCONT);
+
+  /* resync on DISCONT */
+  if (G_UNLIKELY (discont)) {
+    GST_DEBUG_OBJECT (vd, "received DISCONT buffer");
+    vd->granulepos = -1;
+    vd->cur_timestamp = GST_CLOCK_TIME_NONE;
+    vd->prev_timestamp = GST_CLOCK_TIME_NONE;
+#ifdef HAVE_VORBIS_SYNTHESIS_RESTART
+    vorbis_synthesis_restart (&vd->vd);
+#endif
+    vd->discont = TRUE;
+  }
+
+  if (vd->segment.rate >= 0.0)
+    result = vorbis_dec_chain_forward (vd, discont, buffer);
+  else
+    result = vorbis_dec_chain_reverse (vd, discont, buffer);
+
+  gst_object_unref (vd);
+
+  return result;
+}
+
 static GstStateChangeReturn
 vorbis_dec_change_state (GstElement * element, GstStateChange transition)
 {
   GstVorbisDec *vd = GST_VORBIS_DEC (element);
   GstStateChangeReturn res;
 
-
   switch (transition) {
     case GST_STATE_CHANGE_NULL_TO_READY:
       break;
@@ -935,11 +1373,7 @@ vorbis_dec_change_state (GstElement * element, GstStateChange transition)
       vorbis_info_init (&vd->vi);
       vorbis_comment_init (&vd->vc);
       vd->initialized = FALSE;
-      vd->cur_timestamp = GST_CLOCK_TIME_NONE;
-      vd->prev_timestamp = GST_CLOCK_TIME_NONE;
-      vd->granulepos = -1;
-      vd->packetno = 0;
-      vd->discont = TRUE;
+      gst_vorbis_dec_reset (vd);
       break;
     case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
       break;
@@ -958,6 +1392,7 @@ vorbis_dec_change_state (GstElement * element, GstStateChange transition)
       vorbis_dsp_clear (&vd->vd);
       vorbis_comment_clear (&vd->vc);
       vorbis_info_clear (&vd->vi);
+      gst_vorbis_dec_reset (vd);
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
       break;
