@@ -27,6 +27,11 @@
 
 #include <string.h>
 
+static const GstElementDetails audio_filter_details =
+GST_ELEMENT_DETAILS ("Audio filter base class",
+    "Filter/Effect/Audio",
+    "Filters audio",
+    "David Schleef <ds@schleef.org>");
 
 /* GstAudioFilter signals and args */
 enum
@@ -83,12 +88,6 @@ gst_audio_filter_get_type (void)
 static void
 gst_audio_filter_base_init (gpointer g_class)
 {
-  static GstElementDetails audio_filter_details = {
-    "Audio filter base class",
-    "Filter/Effect/Audio",
-    "Filters audio",
-    "David Schleef <ds@schleef.org>"
-  };
   GstAudioFilterClass *klass = (GstAudioFilterClass *) g_class;
   GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
 
@@ -106,7 +105,7 @@ gst_audio_filter_class_init (gpointer g_class, gpointer class_data)
   gobject_class = (GObjectClass *) klass;
   gstelement_class = (GstElementClass *) klass;
 
-  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);
+  parent_class = g_type_class_peek_parent (klass);
 
   gobject_class->set_property = gst_audio_filter_set_property;
   gobject_class->get_property = gst_audio_filter_get_property;
@@ -136,6 +135,7 @@ gst_audio_filter_link (GstPad * pad, GstPad * peer)
   }
 
   if (GST_PAD_LINK_FAILED (link_ret)) {
+    gst_object_unref (audiofilter);
     return link_ret;
   }
 
@@ -153,8 +153,10 @@ gst_audio_filter_link (GstPad * pad, GstPad * peer)
   ret &= gst_structure_get_int (structure, "rate", &audiofilter->rate);
   ret &= gst_structure_get_int (structure, "channels", &audiofilter->channels);
 
-  if (!ret)
+  if (!ret) {
+    gst_object_unref (audiofilter);
     return GST_PAD_LINK_REFUSED;
+  }
 
   audiofilter->bytes_per_sample = (audiofilter->width / 8) *
       audiofilter->channels;
@@ -163,6 +165,7 @@ gst_audio_filter_link (GstPad * pad, GstPad * peer)
     (audio_filter_class->setup) (audiofilter);
 #endif
 
+  gst_object_unref (audiofilter);
   return GST_PAD_LINK_OK;
 }
 
@@ -197,6 +200,7 @@ gst_audio_filter_init (GTypeInstance * instance, gpointer g_class)
 static GstFlowReturn
 gst_audio_filter_chain (GstPad * pad, GstBuffer * buffer)
 {
+  GstFlowReturn ret = GST_FLOW_OK;
   GstBuffer *inbuf = GST_BUFFER (buffer);
   GstAudioFilter *audiofilter;
   GstBuffer *outbuf;
@@ -209,7 +213,7 @@ gst_audio_filter_chain (GstPad * pad, GstBuffer * buffer)
   g_return_val_if_fail (inbuf != NULL, GST_FLOW_ERROR);
 
   audiofilter = GST_AUDIO_FILTER (gst_pad_get_parent (pad));
-  //g_return_if_fail (audiofilter->inited);
+  /* g_return_if_fail (audiofilter->inited); */
   audio_filter_class =
       GST_AUDIO_FILTER_CLASS (G_OBJECT_GET_CLASS (audiofilter));
 
@@ -217,8 +221,9 @@ gst_audio_filter_chain (GstPad * pad, GstBuffer * buffer)
       GST_BUFFER_SIZE (inbuf), GST_OBJECT_NAME (audiofilter));
 
   if (audiofilter->passthru) {
-    gst_pad_push (audiofilter->srcpad, buffer);
-    return GST_FLOW_OK;
+    ret = gst_pad_push (audiofilter->srcpad, buffer);
+    gst_object_unref (audiofilter);
+    return ret;
   }
 
   audiofilter->size = GST_BUFFER_SIZE (inbuf);
@@ -251,9 +256,10 @@ gst_audio_filter_chain (GstPad * pad, GstBuffer * buffer)
     gst_buffer_unref (inbuf);
   }
 
-  gst_pad_push (audiofilter->srcpad, outbuf);
+  ret = gst_pad_push (audiofilter->srcpad, outbuf);
 
-  return GST_FLOW_OK;
+  gst_object_unref (audiofilter);
+  return ret;
 }
 
 static void
