@@ -20,6 +20,18 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/**
+ * SECTION:gstbaseaudiosink
+ * @short_description: Base class for audio sinks
+ * @see_also: #GstAudioSink, #GstRingBuffer.
+ *
+ * This is the base class for audio sinks. Subclasses need to implement the
+ * ::create_ringbuffer vmethod. This base class will then take care of
+ * writing samples to the ringbuffer, synchronisation, clipping and flushing.
+ *
+ * Last reviewed on 2006-09-27 (0.10.12)
+ */
+
 #include <string.h>
 
 #include "gstbaseaudiosink.h"
@@ -39,12 +51,13 @@ enum
  * and sample offset position. 
  * This is an emergency resync fallback since buffers marked as DISCONT will
  * always lock to the correct timestamp immediatly and buffers not marked as
- * DISCONT are contiguous bu definition.
+ * DISCONT are contiguous by definition.
  */
 #define DIFF_TOLERANCE  2
 
-#define DEFAULT_BUFFER_TIME     200 * GST_USECOND
-#define DEFAULT_LATENCY_TIME    10 * GST_USECOND
+/* FIXME: 0.11, store the buffer_time and latency_time in nanoseconds */
+#define DEFAULT_BUFFER_TIME     ((200 * GST_MSECOND) / GST_USECOND)
+#define DEFAULT_LATENCY_TIME    ((10 * GST_MSECOND) / GST_USECOND)
 #define DEFAULT_PROVIDE_CLOCK   TRUE
 
 enum
@@ -90,7 +103,7 @@ static void gst_base_audio_sink_get_times (GstBaseSink * bsink,
 static gboolean gst_base_audio_sink_setcaps (GstBaseSink * bsink,
     GstCaps * caps);
 
-//static guint gst_base_audio_sink_signals[LAST_SIGNAL] = { 0 };
+/* static guint gst_base_audio_sink_signals[LAST_SIGNAL] = { 0 }; */
 
 static void
 gst_base_audio_sink_base_init (gpointer g_class)
@@ -100,8 +113,6 @@ gst_base_audio_sink_base_init (gpointer g_class)
 static void
 gst_base_audio_sink_class_init (GstBaseAudioSinkClass * klass)
 {
-  gchar *longdesc;
-
   GObjectClass *gobject_class;
   GstElementClass *gstelement_class;
   GstBaseSinkClass *gstbasesink_class;
@@ -116,22 +127,17 @@ gst_base_audio_sink_class_init (GstBaseAudioSinkClass * klass)
       GST_DEBUG_FUNCPTR (gst_base_audio_sink_get_property);
   gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_base_audio_sink_dispose);
 
-  longdesc =
-      g_strdup_printf
-      ("Size of audio buffer in microseconds (use -1 for default of %"
-      G_GUINT64_FORMAT " us)", DEFAULT_BUFFER_TIME / GST_USECOND);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_BUFFER_TIME,
-      g_param_spec_int64 ("buffer-time", "Buffer Time", longdesc, -1,
+  g_object_class_install_property (gobject_class, PROP_BUFFER_TIME,
+      g_param_spec_int64 ("buffer-time", "Buffer Time",
+          "Size of audio buffer in microseconds", 1,
           G_MAXINT64, DEFAULT_BUFFER_TIME, G_PARAM_READWRITE));
-  g_free (longdesc);
-  longdesc =
-      g_strdup_printf ("Audio latency in microseconds (use -1 for default of %"
-      G_GUINT64_FORMAT " us)", DEFAULT_LATENCY_TIME / GST_USECOND);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_LATENCY_TIME,
-      g_param_spec_int64 ("latency-time", "Latency Time", longdesc, -1,
+
+  g_object_class_install_property (gobject_class, PROP_LATENCY_TIME,
+      g_param_spec_int64 ("latency-time", "Latency Time",
+          "Audio latency in microseconds", 1,
           G_MAXINT64, DEFAULT_LATENCY_TIME, G_PARAM_READWRITE));
-  g_free (longdesc);
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_PROVIDE_CLOCK,
+
+  g_object_class_install_property (gobject_class, PROP_PROVIDE_CLOCK,
       g_param_spec_boolean ("provide-clock", "Provide Clock",
           "Provide a clock to be used as the global pipeline clock",
           DEFAULT_PROVIDE_CLOCK, G_PARAM_READWRITE));
@@ -190,7 +196,7 @@ gst_base_audio_sink_provide_clock (GstElement * elem)
 
   sink = GST_BASE_AUDIO_SINK (elem);
 
-  /* we have no ringbuffer (must be NULL state */
+  /* we have no ringbuffer (must be NULL state) */
   if (sink->ringbuffer == NULL)
     goto wrong_state;
 
@@ -206,6 +212,7 @@ gst_base_audio_sink_provide_clock (GstElement * elem)
 
   return clock;
 
+  /* ERRORS */
 wrong_state:
   {
     GST_DEBUG_OBJECT (sink, "ringbuffer not acquired");
@@ -222,18 +229,32 @@ clock_disabled:
 static GstClockTime
 gst_base_audio_sink_get_time (GstClock * clock, GstBaseAudioSink * sink)
 {
-  guint64 samples;
+  guint64 raw, samples;
+  guint delay;
   GstClockTime result;
 
   if (sink->ringbuffer == NULL || sink->ringbuffer->spec.rate == 0)
     return GST_CLOCK_TIME_NONE;
 
   /* our processed samples are always increasing */
-  samples = gst_ring_buffer_samples_done (sink->ringbuffer);
+  raw = samples = gst_ring_buffer_samples_done (sink->ringbuffer);
+
+  /* the number of samples not yet processed, this is still queued in the
+   * device (not played for playback). */
+  delay = gst_ring_buffer_delay (sink->ringbuffer);
+
+  if (G_LIKELY (samples >= delay))
+    samples -= delay;
+  else
+    samples = 0;
 
   result = gst_util_uint64_scale_int (samples, GST_SECOND,
       sink->ringbuffer->spec.rate);
 
+  GST_DEBUG_OBJECT (sink,
+      "processed samples: raw %llu, delay %u, real %llu, time %"
+      GST_TIME_FORMAT, raw, delay, samples, GST_TIME_ARGS (result));
+
   return result;
 }
 
@@ -295,6 +316,9 @@ gst_base_audio_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   GstBaseAudioSink *sink = GST_BASE_AUDIO_SINK (bsink);
   GstRingBufferSpec *spec;
 
+  if (!sink->ringbuffer)
+    return FALSE;
+
   spec = &sink->ringbuffer->spec;
 
   GST_DEBUG_OBJECT (sink, "release old ringbuffer");
@@ -318,12 +342,12 @@ gst_base_audio_sink_setcaps (GstBaseSink * bsink, GstCaps * caps)
   if (!gst_ring_buffer_acquire (sink->ringbuffer, spec))
     goto acquire_error;
 
-  /* calculate actual latency and buffer times */
-  spec->latency_time =
-      spec->segsize * GST_MSECOND / (spec->rate * spec->bytes_per_sample);
-  spec->buffer_time =
-      spec->segtotal * spec->segsize * GST_MSECOND / (spec->rate *
-      spec->bytes_per_sample);
+  /* calculate actual latency and buffer times. 
+   * FIXME: In 0.11, store the latency_time internally in ns */
+  spec->latency_time = gst_util_uint64_scale (spec->segsize,
+      (GST_SECOND / GST_USECOND), spec->rate * spec->bytes_per_sample);
+
+  spec->buffer_time = spec->segtotal * spec->latency_time;
 
   gst_ring_buffer_debug_spec_buff (spec);
 
@@ -417,6 +441,17 @@ gst_base_audio_sink_event (GstBaseSink * bsink, GstEvent * event)
       /* now wait till we played everything */
       gst_base_audio_sink_drain (sink);
       break;
+    case GST_EVENT_NEWSEGMENT:
+    {
+      gdouble rate;
+
+      /* we only need the rate */
+      gst_event_parse_new_segment_full (event, NULL, &rate, NULL, NULL,
+          NULL, NULL, NULL);
+
+      GST_DEBUG_OBJECT (sink, "new rate of %f", rate);
+      break;
+    }
     default:
       break;
   }
@@ -480,27 +515,25 @@ gst_base_audio_sink_get_offset (GstBaseAudioSink * sink)
 static GstFlowReturn
 gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
 {
-  guint64 render_offset, in_offset;
-  GstClockTime time, stop, render_time, duration;
+  guint64 in_offset, clock_offset;
+  GstClockTime time, stop, render_start, render_stop, sample_offset;
   GstBaseAudioSink *sink;
   GstRingBuffer *ringbuf;
-  gint64 diff, ctime, cstop;
+  gint64 diff, align, ctime, cstop;
   guint8 *data;
   guint size;
   guint samples, written;
   gint bps;
-  gdouble crate = 1.0;
+  gint accum;
   GstClockTime crate_num;
   GstClockTime crate_denom;
+  gint out_samples;
   GstClockTime cinternal, cexternal;
+  GstClock *clock;
+  gboolean sync;
 
   sink = GST_BASE_AUDIO_SINK (bsink);
 
-  if (G_UNLIKELY (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DISCONT))) {
-    /* always resync after a discont */
-    sink->next_sample = -1;
-  }
-
   ringbuf = sink->ringbuffer;
 
   /* can't do anything when we don't have the device */
@@ -514,24 +547,28 @@ gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
     goto wrong_size;
 
   samples = size / bps;
+  out_samples = samples;
 
   in_offset = GST_BUFFER_OFFSET (buf);
   time = GST_BUFFER_TIMESTAMP (buf);
-  duration = GST_BUFFER_DURATION (buf);
-  data = GST_BUFFER_DATA (buf);
+  stop = time + gst_util_uint64_scale_int (samples, GST_SECOND,
+      ringbuf->spec.rate);
 
   GST_DEBUG_OBJECT (sink,
-      "time %" GST_TIME_FORMAT ", offset %llu, start %" GST_TIME_FORMAT,
-      GST_TIME_ARGS (time), in_offset, GST_TIME_ARGS (bsink->segment.start));
+      "time %" GST_TIME_FORMAT ", offset %llu, start %" GST_TIME_FORMAT
+      ", samples %u", GST_TIME_ARGS (time), in_offset,
+      GST_TIME_ARGS (bsink->segment.start), samples);
 
-  /* if not valid timestamp or we don't need to sync, try to play
+  data = GST_BUFFER_DATA (buf);
+
+  /* if not valid timestamp or we can't clip or sync, try to play
    * sample ASAP */
-  if (!GST_CLOCK_TIME_IS_VALID (time) || !bsink->sync) {
-    render_offset = gst_base_audio_sink_get_offset (sink);
-    stop = -1;
+  if (!GST_CLOCK_TIME_IS_VALID (time)) {
+    render_start = gst_base_audio_sink_get_offset (sink);
+    render_stop = render_start + samples;
     GST_DEBUG_OBJECT (sink,
-        "Buffer of size %u has no time. Using render_offset=%" G_GUINT64_FORMAT,
-        GST_BUFFER_SIZE (buf), render_offset);
+        "Buffer of size %u has no time. Using render_start=%" G_GUINT64_FORMAT,
+        GST_BUFFER_SIZE (buf), render_start);
     goto no_sync;
   }
 
@@ -541,9 +578,6 @@ gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
    * boundaries */
   /* let's calc stop based on the number of samples in the buffer instead
    * of trusting the DURATION */
-  stop =
-      time + gst_util_uint64_scale_int (samples, GST_SECOND,
-      ringbuf->spec.rate);
   if (!gst_segment_clip (&bsink->segment, GST_FORMAT_TIME, time, stop, &ctime,
           &cstop))
     goto out_of_segment;
@@ -569,83 +603,149 @@ gst_base_audio_sink_render (GstBaseSink * bsink, GstBuffer * buf)
     stop = cstop;
   }
 
+  /* figure out how to sync */
+  if ((clock = GST_ELEMENT_CLOCK (bsink)))
+    sync = bsink->sync;
+  else
+    sync = FALSE;
+
+  if (!sync) {
+    /* no sync needed, play sample ASAP */
+    render_start = gst_base_audio_sink_get_offset (sink);
+    render_stop = render_start + samples;
+    GST_DEBUG_OBJECT (sink,
+        "no sync needed. Using render_start=%" G_GUINT64_FORMAT, render_start);
+    goto no_sync;
+  }
+
+  /* bring buffer start and stop times to running time */
+  render_start =
+      gst_segment_to_running_time (&bsink->segment, GST_FORMAT_TIME, time);
+  render_stop =
+      gst_segment_to_running_time (&bsink->segment, GST_FORMAT_TIME, stop);
+
+  GST_DEBUG_OBJECT (sink,
+      "running: start %" GST_TIME_FORMAT " - stop %" GST_TIME_FORMAT,
+      GST_TIME_ARGS (render_start), GST_TIME_ARGS (render_stop));
+
+  /* get calibration parameters to compensate for speed and offset differences
+   * when we are slaved */
   gst_clock_get_calibration (sink->provided_clock, &cinternal, &cexternal,
       &crate_num, &crate_denom);
 
-  /* bring buffer timestamp to running time */
-  render_time =
-      gst_segment_to_running_time (&bsink->segment, GST_FORMAT_TIME, time);
-  /* add base time to get absolute clock time */
-  render_time +=
+  clock_offset =
       (gst_element_get_base_time (GST_ELEMENT_CAST (bsink)) - cexternal) +
       cinternal;
-  /* and bring the time to the offset in the buffer */
-  render_offset =
-      gst_util_uint64_scale_int (render_time, ringbuf->spec.rate, GST_SECOND);
 
-  GST_DEBUG_OBJECT (sink, "render time %" GST_TIME_FORMAT
-      ", render offset %llu, samples %lu",
-      GST_TIME_ARGS (render_time), render_offset, samples);
+  GST_DEBUG_OBJECT (sink, "clock offset %" GST_TIME_FORMAT " %" G_GUINT64_FORMAT
+      "/%" G_GUINT64_FORMAT, GST_TIME_ARGS (clock_offset), crate_num,
+      crate_denom);
 
-  /* roundoff errors in timestamp conversion */
-  if (G_LIKELY (sink->next_sample != -1)) {
-    diff = ABS ((gint64) render_offset - (gint64) sink->next_sample);
+  /* and bring the time to the rate corrected offset in the buffer */
+  render_start = gst_util_uint64_scale_int (render_start + clock_offset,
+      ringbuf->spec.rate, GST_SECOND);
+  render_stop = gst_util_uint64_scale_int (render_stop + clock_offset,
+      ringbuf->spec.rate, GST_SECOND);
 
-    /* we tollerate half a second diff before we start resyncing. This
-     * should be enough to compensate for various rounding errors in the timestamp
-     * and sample offset position. We always resync if we got a discont anyway and
-     * non-discont should be aligned by definition. */
-    if (diff < ringbuf->spec.rate / DIFF_TOLERANCE) {
-      GST_DEBUG_OBJECT (sink,
-          "align with prev sample, %" G_GINT64_FORMAT " < %lu", diff,
-          ringbuf->spec.rate / DIFF_TOLERANCE);
-      /* just align with previous sample then */
-      render_offset = sink->next_sample;
-    } else {
-      /* timestamps drifted apart from previous samples too much, we need to
-       * resync. */
-      GST_WARNING_OBJECT (sink,
-          "resync after discont with previous sample of diff: %lu", diff);
-    }
-  } else {
+  GST_DEBUG_OBJECT (sink,
+      "render: start %" GST_TIME_FORMAT " - stop %" GST_TIME_FORMAT,
+      GST_TIME_ARGS (render_start), GST_TIME_ARGS (render_stop));
+
+  /* always resync after a discont */
+  if (G_UNLIKELY (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLAG_DISCONT))) {
     GST_DEBUG_OBJECT (sink, "resync after discont");
+    goto no_align;
   }
 
-  crate =
-      gst_guint64_to_gdouble (crate_num) / gst_guint64_to_gdouble (crate_denom);
-  GST_DEBUG_OBJECT (sink,
-      "internal %" G_GUINT64_FORMAT ", %" G_GUINT64_FORMAT ", rate %g",
-      cinternal, cexternal, crate);
+  if (G_UNLIKELY (sink->next_sample == -1)) {
+    GST_DEBUG_OBJECT (sink,
+        "no align possible: no previous sample position known");
+    goto no_align;
+  }
+
+  if (bsink->segment.rate >= 1.0)
+    sample_offset = render_start;
+  else
+    sample_offset = render_stop;
+
+  /* now try to align the sample to the previous one */
+  if (sample_offset >= sink->next_sample)
+    diff = sample_offset - sink->next_sample;
+  else
+    diff = sink->next_sample - sample_offset;
+
+  /* we tollerate half a second diff before we start resyncing. This
+   * should be enough to compensate for various rounding errors in the timestamp
+   * and sample offset position. We always resync if we got a discont anyway and
+   * non-discont should be aligned by definition. */
+  if (G_LIKELY (diff < ringbuf->spec.rate / DIFF_TOLERANCE)) {
+    GST_DEBUG_OBJECT (sink,
+        "align with prev sample, %" G_GINT64_FORMAT " < %d", diff,
+        ringbuf->spec.rate / DIFF_TOLERANCE);
+    /* calc align with previous sample */
+    align = sink->next_sample - sample_offset;
+  } else {
+    /* bring sample diff to seconds for error message */
+    diff = gst_util_uint64_scale_int (diff, GST_SECOND, ringbuf->spec.rate);
+    /* timestamps drifted apart from previous samples too much, we need to
+     * resync. We log this as an element warning. */
+    GST_ELEMENT_WARNING (sink, CORE, CLOCK,
+        ("Compensating for audio synchronisation problems"),
+        ("Unexpected discontinuity in audio timestamps of more "
+            "than half a second (%" GST_TIME_FORMAT "), resyncing",
+            GST_TIME_ARGS (diff)));
+    align = 0;
+  }
+
+  /* apply alignment */
+  render_start += align;
+
+  /* only align stop if we are not slaved */
+  if (clock != sink->provided_clock) {
+    GST_DEBUG_OBJECT (sink, "no stop time align needed: we are slaved");
+    goto no_align;
+  }
+  render_stop += align;
+
+no_align:
+  /* number of target samples is difference between start and stop */
+  out_samples = render_stop - render_start;
 
 no_sync:
-  /* clip length based on rate */
-  samples = MIN (samples, samples / (crate * bsink->segment.abs_rate));
+  GST_DEBUG_OBJECT (sink, "rendering at %" G_GUINT64_FORMAT " %d/%d",
+      sink->next_sample, samples, out_samples);
 
-  /* the next sample should be current sample and its length */
-  sink->next_sample = render_offset + samples;
+  /* we render the first or last sample first, depending on the rate */
+  if (bsink->segment.rate >= 1.0)
+    sample_offset = render_start;
+  else
+    sample_offset = render_stop;
 
+  /* we need to accumulate over different runs for when we get interrupted */
+  accum = 0;
   do {
-    written = gst_ring_buffer_commit (ringbuf, render_offset, data, samples);
+    written =
+        gst_ring_buffer_commit_full (ringbuf, &sample_offset, data, samples,
+        out_samples, &accum);
+
     GST_DEBUG_OBJECT (sink, "wrote %u of %u", written, samples);
     /* if we wrote all, we're done */
     if (written == samples)
       break;
 
-    /* else something interrupted us */
-    GST_DEBUG_OBJECT (sink, "wait for preroll...");
-    bsink->have_preroll = TRUE;
-    GST_PAD_PREROLL_WAIT (bsink->sinkpad);
-    bsink->have_preroll = FALSE;
-    GST_DEBUG_OBJECT (sink, "preroll done");
-    if (G_UNLIKELY (bsink->flushing))
+    /* else something interrupted us and we wait for preroll. */
+    if (gst_base_sink_wait_preroll (bsink) != GST_FLOW_OK)
       goto stopping;
-    GST_DEBUG_OBJECT (sink, "continue after preroll");
 
-    render_offset += written;
     samples -= written;
     data += written * bps;
   } while (TRUE);
 
+  sink->next_sample = sample_offset;
+
+  GST_DEBUG_OBJECT (sink, "next sample expected at %" G_GUINT64_FORMAT,
+      sink->next_sample);
+
   if (GST_CLOCK_TIME_IS_VALID (stop) && stop >= bsink->segment.stop) {
     GST_DEBUG_OBJECT (sink,
         "start playback because we are at the end of segment");
@@ -684,6 +784,16 @@ stopping:
   }
 }
 
+/**
+ * gst_base_audio_sink_create_ringbuffer:
+ * @sink: a #GstBaseAudioSink.
+ *
+ * Create and return the #GstRingBuffer for @sink. This function will call the
+ * ::create_ringbuffer vmethod and will set @sink as the parent of the returned
+ * buffer (see gst_object_set_parent()).
+ *
+ * Returns: The new ringbuffer of @sink.
+ */
 GstRingBuffer *
 gst_base_audio_sink_create_ringbuffer (GstBaseAudioSink * sink)
 {
@@ -704,7 +814,7 @@ static void
 gst_base_audio_sink_callback (GstRingBuffer * rbuf, guint8 * data, guint len,
     gpointer user_data)
 {
-  //GstBaseAudioSink *sink = GST_BASE_AUDIO_SINK (data);
+  /* GstBaseAudioSink *sink = GST_BASE_AUDIO_SINK (data); */
 }
 
 /* should be called with the LOCK */
