@@ -1,8 +1,6 @@
 /* GStreamer
  * Copyright (C) 2005 Wim Taymans <wim@fluendo.com>
  *
- * gstringbuffer.c: 
- *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
  * License as published by the Free Software Foundation; either
@@ -18,6 +16,7 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
+
 /**
  * SECTION:gstringbuffer
  * @short_description: Base class for audio ringbuffer implementations
@@ -30,9 +29,9 @@
  * </para>
  * <para>
  * The ringbuffer abstracts a circular buffer of data. One reader and
- * one writer can operate on the data from different threads in a lockfree 
+ * one writer can operate on the data from different threads in a lockfree
  * manner. The base class is sufficiently flexible to be used as an
- * abstraction for DMA based ringbuffers as well as a pure software 
+ * abstraction for DMA based ringbuffers as well as a pure software
  * implementations.
  * </para>
  * </refsect2>
@@ -117,6 +116,8 @@ gst_ring_buffer_dispose (GObject * object)
 {
   GstRingBuffer *ringbuffer = GST_RING_BUFFER (object);
 
+  gst_caps_replace (&ringbuffer->spec.caps, NULL);
+
   G_OBJECT_CLASS (parent_class)->dispose (G_OBJECT (ringbuffer));
 }
 
@@ -131,55 +132,61 @@ gst_ring_buffer_finalize (GObject * object)
   G_OBJECT_CLASS (parent_class)->finalize (G_OBJECT (ringbuffer));
 }
 
-static const int linear_formats[4 * 2 * 2] = {
-  GST_S8,
-  GST_S8,
-  GST_U8,
-  GST_U8,
-  GST_S16_LE,
-  GST_S16_BE,
-  GST_U16_LE,
-  GST_U16_BE,
-  GST_S24_LE,
-  GST_S24_BE,
-  GST_U24_LE,
-  GST_U24_BE,
-  GST_S32_LE,
-  GST_S32_BE,
-  GST_U32_LE,
-  GST_U32_BE
+typedef struct
+{
+  const GstBufferFormat format;
+  const guint8 silence[4];
+} FormatDef;
+
+static const FormatDef linear_defs[4 * 2 * 2] = {
+  {GST_S8, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S8, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U8, {0x80, 0x80, 0x80, 0x80}},
+  {GST_U8, {0x80, 0x80, 0x80, 0x80}},
+  {GST_S16_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S16_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U16_LE, {0x00, 0x80, 0x00, 0x80}},
+  {GST_U16_BE, {0x80, 0x00, 0x80, 0x00}},
+  {GST_S24_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S24_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U24_LE, {0x00, 0x00, 0x80, 0x00}},
+  {GST_U24_BE, {0x80, 0x00, 0x00, 0x00}},
+  {GST_S32_LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S32_BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U32_LE, {0x00, 0x00, 0x00, 0x80}},
+  {GST_U32_BE, {0x80, 0x00, 0x00, 0x00}}
 };
 
-static const int linear24_formats[3 * 2 * 2] = {
-  GST_S24_3LE,
-  GST_S24_3BE,
-  GST_U24_3LE,
-  GST_U24_3BE,
-  GST_S20_3LE,
-  GST_S20_3BE,
-  GST_U20_3LE,
-  GST_U20_3BE,
-  GST_S18_3LE,
-  GST_S18_3BE,
-  GST_U18_3LE,
-  GST_U18_3BE,
+static const FormatDef linear24_defs[3 * 2 * 2] = {
+  {GST_S24_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S24_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U24_3LE, {0x00, 0x00, 0x80, 0x00}},
+  {GST_U24_3BE, {0x80, 0x00, 0x00, 0x00}},
+  {GST_S20_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S20_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U20_3LE, {0x00, 0x00, 0x08, 0x00}},
+  {GST_U20_3BE, {0x08, 0x00, 0x00, 0x00}},
+  {GST_S18_3LE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_S18_3BE, {0x00, 0x00, 0x00, 0x00}},
+  {GST_U18_3LE, {0x00, 0x00, 0x02, 0x00}},
+  {GST_U18_3BE, {0x02, 0x00, 0x00, 0x00}}
 };
 
-static GstBufferFormat
+static const FormatDef *
 build_linear_format (int depth, int width, int unsignd, int big_endian)
 {
-  const gint *formats;
+  const FormatDef *formats;
 
   if (width == 24) {
     switch (depth) {
       case 24:
-        formats = &linear24_formats[0];
+        formats = &linear24_defs[0];
         break;
       case 20:
-        formats = &linear24_formats[4];
+        formats = &linear24_defs[4];
         break;
       case 18:
-        formats = &linear24_formats[8];
+        formats = &linear24_defs[8];
         break;
       default:
         return GST_UNKNOWN;
@@ -187,16 +194,16 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
   } else {
     switch (depth) {
       case 8:
-        formats = &linear_formats[0];
+        formats = &linear_defs[0];
         break;
       case 16:
-        formats = &linear_formats[4];
+        formats = &linear_defs[4];
         break;
       case 24:
-        formats = &linear_formats[8];
+        formats = &linear_defs[8];
         break;
       case 32:
-        formats = &linear_formats[12];
+        formats = &linear_defs[12];
         break;
       default:
         return GST_UNKNOWN;
@@ -206,7 +213,8 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
     formats += 2;
   if (big_endian)
     formats += 1;
-  return (GstBufferFormat) * formats;
+
+  return formats;
 }
 
 /**
@@ -218,6 +226,8 @@ build_linear_format (int depth, int width, int unsignd, int big_endian)
 void
 gst_ring_buffer_debug_spec_caps (GstRingBufferSpec * spec)
 {
+  gint i, bytes;
+
   GST_DEBUG ("spec caps: %p %" GST_PTR_FORMAT, spec->caps, spec->caps);
   GST_DEBUG ("parsed caps: type:         %d", spec->type);
   GST_DEBUG ("parsed caps: format:       %d", spec->format);
@@ -228,6 +238,10 @@ gst_ring_buffer_debug_spec_caps (GstRingBufferSpec * spec)
   GST_DEBUG ("parsed caps: rate:         %d", spec->rate);
   GST_DEBUG ("parsed caps: channels:     %d", spec->channels);
   GST_DEBUG ("parsed caps: sample bytes: %d", spec->bytes_per_sample);
+  bytes = (spec->width >> 3) * spec->channels;
+  for (i = 0; i < bytes; i++) {
+    GST_DEBUG ("silence byte %d: %02x", i, spec->silence_sample[i]);
+  }
 }
 
 /**
@@ -265,14 +279,22 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
 {
   const gchar *mimetype;
   GstStructure *structure;
+  gint i;
 
   structure = gst_caps_get_structure (caps, 0);
 
   /* we have to differentiate between int and float formats */
   mimetype = gst_structure_get_name (structure);
 
+  /* get rate and channels */
+  if (!(gst_structure_get_int (structure, "rate", &spec->rate) &&
+          gst_structure_get_int (structure, "channels", &spec->channels)))
+    goto parse_error;
+
   if (!strncmp (mimetype, "audio/x-raw-int", 15)) {
     gint endianness;
+    const FormatDef *def;
+    gint j, bytes;
 
     spec->type = GST_BUFTYPE_LINEAR;
 
@@ -292,9 +314,18 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
 
     spec->bigend = endianness == G_LITTLE_ENDIAN ? FALSE : TRUE;
 
-    spec->format =
-        build_linear_format (spec->depth, spec->width, spec->sign ? 0 : 1,
+    def = build_linear_format (spec->depth, spec->width, spec->sign ? 0 : 1,
         spec->bigend ? 1 : 0);
+
+    spec->format = def->format;
+
+    bytes = spec->width >> 3;
+
+    for (i = 0; i < spec->channels; i++) {
+      for (j = 0; j < bytes; j++) {
+        spec->silence_sample[i * bytes + j] = def->silence[j];
+      }
+    }
   } else if (!strncmp (mimetype, "audio/x-raw-float", 17)) {
 
     spec->type = GST_BUFTYPE_FLOAT;
@@ -316,34 +347,40 @@ gst_ring_buffer_parse_caps (GstRingBufferSpec * spec, GstCaps * caps)
       default:
         goto parse_error;
     }
+    /* float silence is all zeros.. */
+    memset (spec->silence_sample, 0, 32);
   } else if (!strncmp (mimetype, "audio/x-alaw", 12)) {
     spec->type = GST_BUFTYPE_A_LAW;
     spec->format = GST_A_LAW;
     spec->width = 8;
     spec->depth = 8;
+    for (i = 0; i < spec->channels; i++)
+      spec->silence_sample[i] = 0xd5;
   } else if (!strncmp (mimetype, "audio/x-mulaw", 13)) {
     spec->type = GST_BUFTYPE_MU_LAW;
     spec->format = GST_MU_LAW;
     spec->width = 8;
     spec->depth = 8;
+    for (i = 0; i < spec->channels; i++)
+      spec->silence_sample[i] = 0xff;
   } else {
     goto parse_error;
   }
 
-  /* get rate and channels */
-  if (!(gst_structure_get_int (structure, "rate", &spec->rate) &&
-          gst_structure_get_int (structure, "channels", &spec->channels)))
-    goto parse_error;
-
   spec->bytes_per_sample = (spec->width >> 3) * spec->channels;
 
   gst_caps_replace (&spec->caps, caps);
 
   g_return_val_if_fail (spec->latency_time != 0, FALSE);
 
-  /* calculate suggested segsize and segtotal */
-  spec->segsize =
-      spec->rate * spec->bytes_per_sample * spec->latency_time / GST_MSECOND;
+  /* calculate suggested segsize and segtotal. segsize should be one unit
+   * of 'latency_time' samples, scaling for the fact that latency_time is
+   * currently stored in microseconds (FIXME: in 0.11) */
+  spec->segsize = gst_util_uint64_scale (spec->rate * spec->bytes_per_sample,
+      spec->latency_time, GST_SECOND / GST_USECOND);
+  /* Round to an integer number of samples */
+  spec->segsize -= spec->segsize % spec->bytes_per_sample;
+
   spec->segtotal = spec->buffer_time / spec->latency_time;
 
   gst_ring_buffer_debug_spec_caps (spec);
@@ -587,6 +624,10 @@ gst_ring_buffer_acquire (GstRingBuffer * buf, GstRingBufferSpec * spec)
   /* create an empty segment */
   g_free (buf->empty_seg);
   buf->empty_seg = g_malloc (segsize);
+
+  /* FIXME, we only have 32 silence samples, which might not be enough to
+   * represent silence in all channels */
+  bps = MIN (bps, 32);
   for (i = 0, j = 0; i < segsize; i++) {
     buf->empty_seg[i] = buf->spec.silence_sample[j];
     j = (j + 1) % bps;
@@ -721,6 +762,7 @@ gst_ring_buffer_is_acquired (GstRingBuffer * buf)
 /**
  * gst_ring_buffer_set_flushing:
  * @buf: the #GstRingBuffer to flush
+ * @flushing: the new mode
  *
  * Set the ringbuffer to flushing mode or normal mode.
  *
@@ -878,6 +920,7 @@ gst_ring_buffer_pause (GstRingBuffer * buf)
 
   return res;
 
+  /* ERRORS */
 flushing:
   {
     GST_OBJECT_UNLOCK (buf);
@@ -947,6 +990,12 @@ done:
  * implementation uses another internal buffer between the audio
  * device.
  *
+ * For playback ringbuffers this is the amount of samples transfered from the
+ * ringbuffer to the device but still not played.
+ *
+ * For capture ringbuffers this is the amount of samples in the device that are
+ * not yet transfered to the ringbuffer.
+ *
  * Returns: The number of samples queued in the audio device.
  *
  * MT safe.
@@ -978,7 +1027,8 @@ done:
  * @buf: the #GstRingBuffer to query
  *
  * Get the number of samples that were processed by the ringbuffer
- * since it was last started.
+ * since it was last started. This does not include the number of samples not
+ * yet processed (see gst_ring_buffer_delay()).
  *
  * Returns: The number of samples processed by the ringbuffer.
  *
@@ -988,24 +1038,15 @@ guint64
 gst_ring_buffer_samples_done (GstRingBuffer * buf)
 {
   gint segdone;
-  guint64 raw, samples;
-  guint delay;
+  guint64 samples;
 
   g_return_val_if_fail (GST_IS_RING_BUFFER (buf), 0);
 
   /* get the amount of segments we processed */
   segdone = g_atomic_int_get (&buf->segdone);
 
-  /* and the number of samples not yet processed */
-  delay = gst_ring_buffer_delay (buf);
-
-  raw = samples = ((guint64) segdone) * buf->samples_per_seg;
-
-  if (G_LIKELY (samples >= delay))
-    samples -= delay;
-
-  GST_DEBUG_OBJECT (buf, "processed samples: raw %llu, delay %u, real %llu",
-      raw, delay, samples);
+  /* convert to samples */
+  samples = ((guint64) segdone) * buf->samples_per_seg;
 
   return samples;
 }
@@ -1133,33 +1174,136 @@ no_start:
   }
 }
 
+#define FWD_SAMPLES(s,se,d,de)		 	\
+G_STMT_START {					\
+  /* no rate conversion */			\
+  guint towrite = MIN (se + bps - s, de - d);	\
+  /* simple copy */				\
+  if (!skip)					\
+    memcpy (d, s, towrite);			\
+  in_samples -= towrite / bps;			\
+  out_samples -= towrite / bps;			\
+  s += towrite;					\
+  GST_DEBUG ("copy %u bytes", towrite);		\
+} G_STMT_END
+
+/* in_samples >= out_samples, rate > 1.0 */
+#define FWD_UP_SAMPLES(s,se,d,de) 	 	\
+G_STMT_START {					\
+  guint8 *sb = s, *db = d;			\
+  while (s <= se && d < de) {			\
+    if (!skip)					\
+      memcpy (d, s, bps);			\
+    s += bps;					\
+    *accum += outr;				\
+    if ((*accum << 1) >= inr) {			\
+      *accum -= inr;				\
+      d += bps;					\
+    }						\
+  }						\
+  in_samples -= (s - sb)/bps;			\
+  out_samples -= (d - db)/bps;			\
+  GST_DEBUG ("fwd_up end %d/%d",*accum,*toprocess);	\
+} G_STMT_END
+
+/* out_samples > in_samples, for rates smaller than 1.0 */
+#define FWD_DOWN_SAMPLES(s,se,d,de) 	 	\
+G_STMT_START {					\
+  guint8 *sb = s, *db = d;			\
+  while (s <= se && d < de) {			\
+    if (!skip)					\
+      memcpy (d, s, bps);			\
+    d += bps;					\
+    *accum += inr;				\
+    if ((*accum << 1) >= outr) {		\
+      *accum -= outr;				\
+      s += bps;					\
+    }						\
+  }						\
+  in_samples -= (s - sb)/bps;			\
+  out_samples -= (d - db)/bps;			\
+  GST_DEBUG ("fwd_down end %d/%d",*accum,*toprocess);	\
+} G_STMT_END
+
+#define REV_UP_SAMPLES(s,se,d,de) 	 	\
+G_STMT_START {					\
+  guint8 *sb = se, *db = d;			\
+  while (s <= se && d < de) {			\
+    if (!skip)					\
+      memcpy (d, se, bps);			\
+    se -= bps;					\
+    *accum += outr;				\
+    while ((*accum << 1) >= inr) {		\
+      *accum -= inr;				\
+      d += bps;					\
+    }						\
+  }						\
+  in_samples -= (sb - se)/bps;			\
+  out_samples -= (d - db)/bps;			\
+  GST_DEBUG ("rev_up end %d/%d",*accum,*toprocess);	\
+} G_STMT_END
+
+#define REV_DOWN_SAMPLES(s,se,d,de) 	 	\
+G_STMT_START {					\
+  guint8 *sb = se, *db = d;			\
+  while (s <= se && d < de) {			\
+    if (!skip)					\
+      memcpy (d, se, bps);			\
+    d += bps;					\
+    *accum += inr;				\
+    while ((*accum << 1) >= outr) {		\
+      *accum -= outr;				\
+      se -= bps;				\
+    }						\
+  }						\
+  in_samples -= (sb - se)/bps;			\
+  out_samples -= (d - db)/bps;			\
+  GST_DEBUG ("rev_down end %d/%d",*accum,*toprocess);	\
+} G_STMT_END
+
 /**
- * gst_ring_buffer_commit:
+ * gst_ring_buffer_commit_full:
  * @buf: the #GstRingBuffer to commit
  * @sample: the sample position of the data
  * @data: the data to commit
- * @len: the number of samples in the data to commit
+ * @in_samples: the number of samples in the data to commit
+ * @out_sampled: the number of samples to write to the ringbuffer
+ * @accum: accumulator for rate conversion.
  *
- * Commit @len samples pointed to by @data to the ringbuffer
- * @buf. The first sample should be written at position @sample in
- * the ringbuffer.
+ * Commit @len samples pointed to by @data to the ringbuffer @buf. 
  *
- * @len does not need to be a multiple of the segment size of the ringbuffer
- * although it is recommended for optimal performance.
+ * @num and @denom define the rate conversion to perform on the the samples in
+ * @data. For negative rates, @num must be negative and @denom positive.
  *
- * Returns: The number of samples written to the ringbuffer or -1 on
- * error.
+ * When @num is positive, the first sample will be written at position @sample
+ * in the ringbuffer. When @num is negative, the last sample will be written to
+ * @sample in reverse order.
+ *
+ * @len does not need to be a multiple of the segment size of the ringbuffer (if
+ * @num and @denom are both 1) although it is recommended for optimal performance. 
+ *
+ * Returns: The number of samples written to the ringbuffer or -1 on error. The
+ * number of samples written can be less than @len when @buf was interrupted
+ * with a flush or stop.
+ *
+ * Since: 0.10.11.
  *
  * MT safe.
  */
 guint
-gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
-    guint len)
+gst_ring_buffer_commit_full (GstRingBuffer * buf, guint64 * sample,
+    guchar * data, gint in_samples, gint out_samples, gint * accum)
 {
   gint segdone;
   gint segsize, segtotal, bps, sps;
-  guint8 *dest;
-  guint to_write;
+  guint8 *dest, *data_end;
+  gint writeseg, sampleoff;
+  gint *toprocess;
+  gint inr, outr;
+  gboolean reverse;
+
+  if (G_UNLIKELY (in_samples == 0 || out_samples == 0))
+    return 0;
 
   g_return_val_if_fail (GST_IS_RING_BUFFER (buf), -1);
   g_return_val_if_fail (buf->data != NULL, -1);
@@ -1171,16 +1315,32 @@ gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
   bps = buf->spec.bytes_per_sample;
   sps = buf->samples_per_seg;
 
-  to_write = len;
-  /* write out all samples */
-  while (to_write > 0) {
-    gint sampleslen;
-    gint writeseg, sampleoff;
+  reverse = out_samples < 0;
+  out_samples = ABS (out_samples);
 
-    /* figure out the segment and the offset inside the segment where
-     * the sample should be written. */
-    writeseg = sample / sps;
-    sampleoff = (sample % sps);
+  if (in_samples >= out_samples)
+    toprocess = &in_samples;
+  else
+    toprocess = &out_samples;
+
+  inr = in_samples - 1;
+  outr = out_samples - 1;
+
+  /* data_end points to the last sample we have to write, not past it. This is
+   * needed to properly handle reverse playback: it points to the last sample. */
+  data_end = data + (bps * inr);
+
+  /* figure out the segment and the offset inside the segment where
+   * the first sample should be written. */
+  writeseg = *sample / sps;
+  sampleoff = (*sample % sps) * bps;
+
+  /* write out all samples */
+  while (*toprocess > 0) {
+    gint avail;
+    guint8 *d, *d_end;
+    gint ws;
+    gboolean skip;
 
     while (TRUE) {
       gint diff;
@@ -1192,21 +1352,23 @@ gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
       diff = writeseg - segdone;
 
       GST_DEBUG
-          ("pointer at %d, sample %llu, write to %d-%d, to_write %d, diff %d, segtotal %d, segsize %d",
-          segdone, sample, writeseg, sampleoff, to_write, diff, segtotal, sps);
+          ("pointer at %d, write to %d-%d, diff %d, segtotal %d, segsize %d",
+          segdone, writeseg, sampleoff, diff, segtotal, segsize);
 
-      /* segment too far ahead, we need to drop, hopefully UNLIKELY */
+      /* segment too far ahead, writer too slow, we need to drop, hopefully UNLIKELY */
       if (G_UNLIKELY (diff < 0)) {
         /* we need to drop one segment at a time, pretend we wrote a
          * segment. */
-        sampleslen = MIN (sps, to_write);
-        goto next;
+        skip = TRUE;
+        break;
       }
 
       /* write segment is within writable range, we can break the loop and
        * start writing the data. */
-      if (diff < segtotal)
+      if (diff < segtotal) {
+        skip = FALSE;
         break;
+      }
 
       /* else we need to wait for the segment to become writable. */
       if (!wait_segment (buf))
@@ -1214,31 +1376,80 @@ gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
     }
 
     /* we can write now */
-    writeseg = writeseg % segtotal;
-    sampleslen = MIN (sps - sampleoff, to_write);
+    ws = writeseg % segtotal;
+    avail = MIN (segsize - sampleoff, bps * out_samples);
 
-    GST_DEBUG_OBJECT (buf, "write @%p seg %d, off %d, sampleslen %d",
-        dest + writeseg * segsize, writeseg, sampleoff, sampleslen);
+    d = dest + (ws * segsize) + sampleoff;
+    d_end = d + avail;
+    *sample += avail / bps;
 
-    memcpy (dest + (writeseg * segsize) + (sampleoff * bps), data,
-        (sampleslen * bps));
+    GST_DEBUG_OBJECT (buf, "write @%p seg %d, sps %d, off %d, avail %d",
+        dest + ws * segsize, ws, sps, sampleoff, avail);
 
-  next:
-    to_write -= sampleslen;
-    sample += sampleslen;
-    data += sampleslen * bps;
+    if (G_LIKELY (inr == outr && !reverse)) {
+      /* no rate conversion, simply copy samples */
+      FWD_SAMPLES (data, data_end, d, d_end);
+    } else if (!reverse) {
+      if (inr >= outr)
+        /* forward speed up */
+        FWD_UP_SAMPLES (data, data_end, d, d_end);
+      else
+        /* forward slow down */
+        FWD_DOWN_SAMPLES (data, data_end, d, d_end);
+    } else {
+      if (inr >= outr)
+        /* reverse speed up */
+        REV_UP_SAMPLES (data, data_end, d, d_end);
+      else
+        /* reverse slow down */
+        REV_DOWN_SAMPLES (data, data_end, d, d_end);
+    }
+
+    /* for the next iteration we write to the next segment at the beginning. */
+    writeseg++;
+    sampleoff = 0;
   }
+  /* we consumed all samples here */
+  data = data_end + bps;
 
-  return len - to_write;
+done:
+  return inr - ((data_end - data) / bps);
 
   /* ERRORS */
 not_started:
   {
     GST_DEBUG_OBJECT (buf, "stopped processing");
-    return len - to_write;
+    goto done;
   }
 }
 
+/**
+ * gst_ring_buffer_commit:
+ * @buf: the #GstRingBuffer to commit
+ * @sample: the sample position of the data
+ * @data: the data to commit
+ * @len: the number of samples in the data to commit
+ *
+ * Same as gst_ring_buffer_commit_full() but with a num, denom of 1, ignoring
+ * accum.
+ *
+ * Returns: The number of samples written to the ringbuffer or -1 on
+ * error.
+ *
+ * MT safe.
+ */
+guint
+gst_ring_buffer_commit (GstRingBuffer * buf, guint64 sample, guchar * data,
+    guint len)
+{
+  guint res;
+  guint64 samplep = sample;
+
+  res = gst_ring_buffer_commit_full (buf, &samplep, data, len, len, NULL);
+
+  return res;
+}
+
 /**
  * gst_ring_buffer_read:
  * @buf: the #GstRingBuffer to read from
@@ -1266,6 +1477,7 @@ gst_ring_buffer_read (GstRingBuffer * buf, guint64 sample, guchar * data,
   gint segdone;
   gint segsize, segtotal, bps, sps;
   guint8 *dest;
+  guint to_read;
 
   g_return_val_if_fail (GST_IS_RING_BUFFER (buf), -1);
   g_return_val_if_fail (buf->data != NULL, -1);
@@ -1277,13 +1489,14 @@ gst_ring_buffer_read (GstRingBuffer * buf, guint64 sample, guchar * data,
   bps = buf->spec.bytes_per_sample;
   sps = buf->samples_per_seg;
 
+  to_read = len;
   /* read enough samples */
-  while (len > 0) {
+  while (to_read > 0) {
     gint sampleslen;
     gint readseg, sampleoff;
 
     /* figure out the segment and the offset inside the segment where
-     * the sample should be written. */
+     * the sample should be read from. */
     readseg = sample / sps;
     sampleoff = (sample % sps);
 
@@ -1293,33 +1506,29 @@ gst_ring_buffer_read (GstRingBuffer * buf, guint64 sample, guchar * data,
       /* get the currently processed segment */
       segdone = g_atomic_int_get (&buf->segdone) - buf->segbase;
 
-      /* see how far away it is from the read segment */
+      /* see how far away it is from the read segment, normally segdone (where
+       * the hardware is writing) is bigger than readseg (where software is
+       * reading) */
       diff = segdone - readseg;
 
       GST_DEBUG
-          ("pointer at %d, sample %llu, read from %d-%d, len %d, diff %d, segtotal %d, segsize %d",
-          segdone, sample, readseg, sampleoff, len, diff, segtotal, segsize);
+          ("pointer at %d, sample %llu, read from %d-%d, to_read %d, diff %d, segtotal %d, segsize %d",
+          segdone, sample, readseg, sampleoff, to_read, diff, segtotal,
+          segsize);
 
-      /* segment too far ahead, we need to drop */
-      if (diff < 0) {
-        /* we need to drop one segment at a time, pretend we read an
-         * empty segment. */
-        sampleslen = MIN (sps, len);
+      /* segment too far ahead, reader too slow */
+      if (G_UNLIKELY (diff >= segtotal)) {
+        /* pretend we read an empty segment. */
+        sampleslen = MIN (sps, to_read);
         memcpy (data, buf->empty_seg, sampleslen * bps);
         goto next;
       }
 
       /* read segment is within readable range, we can break the loop and
        * start reading the data. */
-      if (diff > 0 && diff < segtotal)
+      if (diff > 0)
         break;
 
-      /* flush if diff has grown bigger than ringbuffer */
-      if (diff >= segtotal) {
-        gst_ring_buffer_clear_all (buf);
-        buf->segdone = readseg;
-      }
-
       /* else we need to wait for the segment to become readable. */
       if (!wait_segment (buf))
         goto not_started;
@@ -1327,27 +1536,27 @@ gst_ring_buffer_read (GstRingBuffer * buf, guint64 sample, guchar * data,
 
     /* we can read now */
     readseg = readseg % segtotal;
-    sampleslen = MIN (sps - sampleoff, len);
+    sampleslen = MIN (sps - sampleoff, to_read);
 
-    GST_DEBUG_OBJECT (buf, "read @%p seg %d, off %d, len %d",
+    GST_DEBUG_OBJECT (buf, "read @%p seg %d, off %d, sampleslen %d",
         dest + readseg * segsize, readseg, sampleoff, sampleslen);
 
     memcpy (data, dest + (readseg * segsize) + (sampleoff * bps),
         (sampleslen * bps));
 
   next:
-    len -= sampleslen;
+    to_read -= sampleslen;
     sample += sampleslen;
     data += sampleslen * bps;
   }
 
-  return len;
+  return len - to_read;
 
   /* ERRORS */
 not_started:
   {
     GST_DEBUG_OBJECT (buf, "stopped processing");
-    return -1;
+    return len - to_read;
   }
 }
 
