@@ -38,12 +38,46 @@
 /**
  * SECTION:gstcddabasesrc
  * @short_description: Base class for CD digital audio (CDDA) sources
- * 
+ *
  * <refsect2>
  * <para>
  * Provides a base class for CDDA sources, which handles things like seeking,
  * querying, discid calculation, tags, and buffer timestamping.
  * </para>
+ * <title>Using GstCddaBaseSrc-based elements in applications</title>
+ * <para>
+ * GstCddaBaseSrc registers two #GstFormat<!-- -->s of its own, namely
+ * the "track" format and the "sector" format. Applications will usually
+ * only find the "track" format interesting. You can retrieve that #GstFormat
+ * for use in seek events or queries with gst_format_get_by_nick("track").
+ * </para>
+ * <para>
+ * In order to query the number of tracks, for example, an application would
+ * set the CDDA source element to READY or PAUSED state and then query the
+ * the number of tracks via gst_element_query_duration() using the track
+ * format acquired above. Applications can query the currently playing track
+ * in the same way.
+ * </para>
+ * <para>
+ * Alternatively, applications may retrieve the currently playing track and
+ * the total number of tracks from the taglist that will posted on the bus
+ * whenever the CD is opened or the currently playing track changes. The
+ * taglist will contain GST_TAG_TRACK_NUMBER and GST_TAG_TRACK_COUNT tags.
+ * </para>
+ * <para>
+ * Applications playing back CD audio using playbin and cdda://n URIs should
+ * issue a seek command in track format to change between tracks, rather than
+ * setting a new cdda://n+1 URI on playbin (as setting a new URI on playbin
+ * involves closing and re-opening the CD device, which is much much slower).
+ * </para>
+ * <title>Tags and meta-information</title>
+ * <para>
+ * CDDA sources will automatically emit a number of tags, details about which
+ * can be found in the libgsttag documentation. Those tags are:
+ * #GST_TAG_CDDA_CDDB_DISCID, #GST_TAG_CDDA_CDDB_DISCID_FULL,
+ * #GST_TAG_CDDA_MUSICBRAINZ_DISCID, #GST_TAG_CDDA_MUSICBRAINZ_DISCID_FULL,
+ * among others.
+ * </para>
  * </refsect2>
  */
 
@@ -52,6 +86,7 @@
 #endif
 
 #include <string.h>
+#include <stdlib.h>             /* for strtol */
 
 #include "gstcddabasesrc.h"
 #include "gst/gst-i18n-plugin.h"
@@ -129,7 +164,7 @@ static GType
 gst_cdda_base_src_mode_get_type (void)
 {
   static GType mode_type;       /* 0 */
-  static GEnumValue modes[] = {
+  static const GEnumValue modes[] = {
     {GST_CDDA_BASE_SRC_MODE_NORMAL, "Stream consists of a single track",
         "normal"},
     {GST_CDDA_BASE_SRC_MODE_CONTINUOUS, "Stream consists of the whole disc",
@@ -155,23 +190,8 @@ gst_cdda_base_src_base_init (gpointer g_class)
   track_format = gst_format_register ("track", "CD track");
   sector_format = gst_format_register ("sector", "CD sector");
 
-  /* tags */
-  gst_tag_register (GST_TAG_CDDA_CDDB_DISCID, GST_TAG_FLAG_META,
-      G_TYPE_STRING, "discid", "CDDB discid for metadata retrieval",
-      gst_tag_merge_use_first);
-
-  gst_tag_register (GST_TAG_CDDA_CDDB_DISCID_FULL, GST_TAG_FLAG_META,
-      G_TYPE_STRING, "discid full",
-      "CDDB discid for metadata retrieval (full)", gst_tag_merge_use_first);
-
-  gst_tag_register (GST_TAG_CDDA_MUSICBRAINZ_DISCID, GST_TAG_FLAG_META,
-      G_TYPE_STRING, "musicbrainz-discid",
-      "Musicbrainz discid for metadata retrieval", gst_tag_merge_use_first);
-
-  gst_tag_register (GST_TAG_CDDA_MUSICBRAINZ_DISCID_FULL, GST_TAG_FLAG_META,
-      G_TYPE_STRING, "musicbrainz-discid-full",
-      "Musicbrainz discid for metadata retrieval (full)",
-      gst_tag_merge_use_first);
+  /* register CDDA tags */
+  gst_tag_register_musicbrainz_tags ();
 
 #if 0
   ///// FIXME: what type to use here? ///////
@@ -790,7 +810,7 @@ gst_cdda_base_src_handle_track_seek (GstCddaBaseSrc * src, gdouble rate,
     return FALSE;
   }
 
-  GST_DEBUG_OBJECT (src, "seeking to track %d", start + 1);
+  GST_DEBUG_OBJECT (src, "seeking to track %" G_GINT64_FORMAT, start + 1);
 
   src->cur_sector = src->tracks[start].start;
   GST_DEBUG_OBJECT (src, "starting at sector %d", src->cur_sector);
@@ -864,7 +884,7 @@ gst_cdda_base_src_handle_event (GstBaseSrc * basesrc, GstEvent * event)
   return ret;
 }
 
-static guint
+static GstURIType
 gst_cdda_base_src_uri_get_type (void)
 {
   return GST_URI_SRC;
@@ -1461,6 +1481,7 @@ gst_cdda_base_src_stop (GstBaseSrc * basesrc)
   return TRUE;
 }
 
+
 static GstFlowReturn
 gst_cdda_base_src_create (GstPushSrc * pushsrc, GstBuffer ** buffer)
 {
@@ -1469,8 +1490,10 @@ gst_cdda_base_src_create (GstPushSrc * pushsrc, GstBuffer ** buffer)
   GstBuffer *buf;
   GstFormat format;
   gboolean eos;
-  gint64 position;
-  gint64 duration;
+
+  GstClockTime position = GST_CLOCK_TIME_NONE;
+  GstClockTime duration = GST_CLOCK_TIME_NONE;
+  gint64 qry_position;
 
   g_assert (klass->read_sector != NULL);
 
@@ -1491,7 +1514,7 @@ gst_cdda_base_src_create (GstPushSrc * pushsrc, GstBuffer ** buffer)
     src->prev_track = -1;
     GST_DEBUG_OBJECT (src, "EOS at sector %d, cur_track=%d, mode=%d",
         src->cur_sector, src->cur_track, src->mode);
-    gst_pad_push_event (GST_BASE_SRC_PAD (src), gst_event_new_eos ());
+    /* base class will send EOS for us */
     return GST_FLOW_UNEXPECTED;
   }
 
@@ -1524,13 +1547,23 @@ gst_cdda_base_src_create (GstPushSrc * pushsrc, GstBuffer ** buffer)
   }
 
   format = GST_FORMAT_TIME;
-  if (!gst_pad_query_position (GST_BASE_SRC_PAD (src), &format, &position)) {
-    position = GST_CLOCK_TIME_NONE;
+  if (gst_pad_query_position (GST_BASE_SRC_PAD (src), &format, &qry_position)) {
+    position = (GstClockTime) qry_position;
+
+    gint64 next_ts = 0;
+
+    ++src->cur_sector;
+    if (gst_pad_query_position (GST_BASE_SRC_PAD (src), &format, &next_ts)) {
+      duration = (GstClockTime) (next_ts - qry_position);
+    }
+    --src->cur_sector;
   }
 
-  /* 4 bytes per sample, 44100 samples per second */
-  duration = gst_util_uint64_scale_int (GST_BUFFER_SIZE (buf) >> 2, GST_SECOND,
-      44100);
+  /* fallback duration: 4 bytes per sample, 44100 samples per second */
+  if (duration == GST_CLOCK_TIME_NONE) {
+    duration = gst_util_uint64_scale_int (GST_BUFFER_SIZE (buf) >> 2,
+        GST_SECOND, 44100);
+  }
 
   GST_BUFFER_TIMESTAMP (buf) = position;
   GST_BUFFER_DURATION (buf) = duration;
