@@ -38,10 +38,10 @@
  * </listitem>
  * <listitem>
  * <para>
- * To force a redrawing of the latest video frame the video sink element 
+ * To force a redrawing of the latest video frame the video sink element
  * displayed on the Window. Indeed if the #GstPipeline is in #GST_STATE_PAUSED
  * state, moving the Window around will damage its content. Application
- * developers will want to handle the Expose events themselves and force the 
+ * developers will want to handle the Expose events themselves and force the
  * video sink element to refresh the Window's content.
  * </para>
  * </listitem>
@@ -49,7 +49,7 @@
  * </para>
  * <para>
  * Using the Window created by the video sink is probably the simplest scenario,
- * in some cases, though, it might not be flexible enough for application 
+ * in some cases, though, it might not be flexible enough for application
  * developers if they need to catch events such as mouse moves and button
  * clicks.
  * </para>
@@ -58,14 +58,12 @@
  * flexible solution but it has some issues. Indeed the application needs to set
  * its Window identifier at the right time to avoid internal Window creation
  * from the video sink element. To solve this issue a #GstMessage is posted on
- * the bus to inform the application that it should set the Window identifier 
+ * the bus to inform the application that it should set the Window identifier
  * immediately. Here is an example on how to do that correctly:
  * <programlisting>
  * static GstBusSyncReply
  * create_window (GstBus * bus, GstMessage * message, GstPipeline * pipeline)
  * {
- *  XGCValues values;
- *  
  *  // ignore anything but 'prepare-xwindow-id' element messages
  *  if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ELEMENT)
  *    return GST_BUS_PASS;
@@ -77,8 +75,6 @@
  *  
  *  XSetWindowBackgroundPixmap (disp, win, None);
  *  
- *  gc = XCreateGC (disp, win, 0, &amp;values);
- *  
  *  XMapRaised (disp, win);
  *  
  *  XSync (disp, FALSE);
@@ -86,6 +82,8 @@
  *  gst_x_overlay_set_xwindow_id (GST_X_OVERLAY (GST_MESSAGE_SRC (message)),
  *      win);
  *   
+ *  gst_message_unref (message);
+ *   
  *  return GST_BUS_DROP;
  * }
  * ...
@@ -158,7 +156,12 @@ gst_x_overlay_base_init (gpointer g_class)
 void
 gst_x_overlay_set_xwindow_id (GstXOverlay * overlay, gulong xwindow_id)
 {
-  GstXOverlayClass *klass = GST_X_OVERLAY_GET_CLASS (overlay);
+  GstXOverlayClass *klass;
+
+  g_return_if_fail (overlay != NULL);
+  g_return_if_fail (GST_IS_X_OVERLAY (overlay));
+
+  klass = GST_X_OVERLAY_GET_CLASS (overlay);
 
   if (klass->set_xwindow_id) {
     klass->set_xwindow_id (overlay, xwindow_id);
@@ -226,7 +229,11 @@ gst_x_overlay_prepare_xwindow_id (GstXOverlay * overlay)
 void
 gst_x_overlay_expose (GstXOverlay * overlay)
 {
-  GstXOverlayClass *klass = GST_X_OVERLAY_GET_CLASS (overlay);
+  GstXOverlayClass *klass;
+
+  g_return_if_fail (overlay != NULL);
+
+  klass = GST_X_OVERLAY_GET_CLASS (overlay);
 
   if (klass->expose) {
     klass->expose (overlay);
