@@ -20,7 +20,7 @@
 
 #include "gstbasertpdepayload.h"
 
-GST_DEBUG_CATEGORY (basertpdepayload_debug);
+GST_DEBUG_CATEGORY_STATIC (basertpdepayload_debug);
 #define GST_CAT_DEFAULT (basertpdepayload_debug)
 
 /* Filter signals and args */
@@ -44,8 +44,8 @@ static void gst_base_rtp_depayload_class_init (GstBaseRTPDepayloadClass *
 static void gst_base_rtp_depayload_init (GstBaseRTPDepayload * filter,
     gpointer g_class);
 
-static void gst_base_rtp_depayload_push (GstBaseRTPDepayload * filter,
-    GstBuffer * rtp_buf);
+static GstFlowReturn gst_base_rtp_depayload_process (GstBaseRTPDepayload *
+    filter, GstBuffer * rtp_buf);
 
 GType
 gst_base_rtp_depayload_get_type (void)
@@ -150,6 +150,7 @@ gst_base_rtp_depayload_init (GstBaseRTPDepayload * filter, gpointer g_class)
       gst_element_class_get_pad_template (GST_ELEMENT_CLASS (g_class), "src");
   g_return_if_fail (pad_template != NULL);
   filter->srcpad = gst_pad_new_from_template (pad_template, "src");
+  gst_pad_use_fixed_caps (filter->srcpad);
   gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
 
   filter->queue = g_queue_new ();
@@ -165,13 +166,13 @@ static void
 gst_base_rtp_depayload_finalize (GObject * object)
 {
   GstBuffer *buf;
+  GstBaseRTPDepayload *filter = GST_BASE_RTP_DEPAYLOAD (object);
 
-  while ((buf = g_queue_pop_head (GST_BASE_RTP_DEPAYLOAD (object)->queue)))
+  while ((buf = g_queue_pop_head (filter->queue)))
     gst_buffer_unref (buf);
-  g_queue_free (GST_BASE_RTP_DEPAYLOAD (object)->queue);
+  g_queue_free (filter->queue);
 
-  if (G_OBJECT_CLASS (parent_class)->finalize)
-    G_OBJECT_CLASS (parent_class)->finalize (object);
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static gboolean
@@ -193,6 +194,7 @@ gst_base_rtp_depayload_setcaps (GstPad * pad, GstCaps * caps)
     res = TRUE;
 
   gst_object_unref (filter);
+
   return res;
 }
 
@@ -205,18 +207,37 @@ gst_base_rtp_depayload_chain (GstPad * pad, GstBuffer * in)
 
   filter = GST_BASE_RTP_DEPAYLOAD (GST_OBJECT_PARENT (pad));
 
-  g_return_val_if_fail (filter->clock_rate > 0, GST_FLOW_ERROR);
+  if (filter->clock_rate <= 0)
+    goto not_configured;
 
   bclass = GST_BASE_RTP_DEPAYLOAD_GET_CLASS (filter);
 
   if (filter->queue_delay == 0) {
     GST_DEBUG_OBJECT (filter, "Pushing directly!");
-    gst_base_rtp_depayload_push (filter, in);
+    ret = gst_base_rtp_depayload_process (filter, in);
   } else {
     if (bclass->add_to_queue)
       ret = bclass->add_to_queue (filter, in);
+    else
+      goto no_delay;
   }
   return ret;
+
+  /* ERRORS */
+not_configured:
+  {
+    GST_ELEMENT_ERROR (filter, STREAM, FORMAT,
+        (NULL), ("no clock rate was specified, likely incomplete input caps"));
+    gst_buffer_unref (in);
+    return GST_FLOW_NOT_NEGOTIATED;
+  }
+no_delay:
+  {
+    GST_ELEMENT_ERROR (filter, STREAM, NOT_IMPLEMENTED,
+        (NULL), ("This element cannot operate with delay"));
+    gst_buffer_unref (in);
+    return GST_FLOW_NOT_SUPPORTED;
+  }
 }
 
 static gboolean
@@ -240,18 +261,33 @@ gst_base_rtp_depayload_handle_sink_event (GstPad * pad, GstEvent * event)
         gst_event_unref (event);
         break;
       } else {
+        GstFormat format;
+
+        gst_event_parse_new_segment (event, NULL, NULL, &format, NULL, NULL,
+            NULL);
+        if (format != GST_FORMAT_TIME)
+          goto wrong_format;
+
         GST_DEBUG_OBJECT (filter,
             "Upstream sent a NEWSEGMENT, passing through.");
       }
-    }
       /* note: pass through to default if no thread running */
+    }
     default:
       /* pass other events forward */
       res = gst_pad_push_event (filter->srcpad, event);
       break;
   }
-
   return res;
+
+  /* ERRORS */
+wrong_format:
+  {
+    GST_DEBUG_OBJECT (filter,
+        "Upstream sent a NEWSEGMENT in wrong format, dropping.");
+    gst_event_unref (event);
+    return TRUE;
+  }
 }
 
 static GstFlowReturn
@@ -276,7 +312,10 @@ gst_base_rtp_depayload_add_to_queue (GstBaseRTPDepayload * filter,
     /* look for right place to insert it */
     i = 0;
 
-    while (seqnum > queueseq) {
+    /* Check for seqnum wraparound.
+     * Seqnums in the lowest quadrant of the 0-65535 space are considered to
+     * be greater than seqnums in the highest quadrant of this space. */
+    while (seqnum > queueseq || (seqnum < 16384 && queueseq > 49150)) {
       gpointer data;
 
       i++;
@@ -300,62 +339,130 @@ gst_base_rtp_depayload_add_to_queue (GstBaseRTPDepayload * filter,
   return GST_FLOW_OK;
 }
 
-static void
-gst_base_rtp_depayload_push (GstBaseRTPDepayload * filter, GstBuffer * rtp_buf)
+static GstFlowReturn
+gst_base_rtp_depayload_push_full (GstBaseRTPDepayload * filter,
+    gboolean do_ts, guint32 timestamp, GstBuffer * out_buf)
 {
-  GstBaseRTPDepayloadClass *bclass = GST_BASE_RTP_DEPAYLOAD_GET_CLASS (filter);
-  GstBuffer *out_buf;
+  GstFlowReturn ret;
   GstCaps *srccaps;
+  GstBaseRTPDepayloadClass *bclass;
+
+  /* set the caps if any */
+  srccaps = GST_PAD_CAPS (filter->srcpad);
+  if (srccaps)
+    gst_buffer_set_caps (out_buf, srccaps);
+
+  bclass = GST_BASE_RTP_DEPAYLOAD_GET_CLASS (filter);
+
+  /* set the timestamp if we must and can */
+  if (bclass->set_gst_timestamp && do_ts)
+    bclass->set_gst_timestamp (filter, timestamp, out_buf);
+
+  /* push it */
+  GST_LOG_OBJECT (filter, "Pushing buffer size %d, timestamp %" GST_TIME_FORMAT,
+      GST_BUFFER_SIZE (out_buf),
+      GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (out_buf)));
+  ret = gst_pad_push (filter->srcpad, out_buf);
+  GST_LOG_OBJECT (filter, "Pushed buffer: %s", gst_flow_get_name (ret));
+
+  return ret;
+}
+
+/**
+ * gst_base_rtp_depayload_push_ts:
+ * @filter: a #GstBaseRTPDepayload
+ * @timestamp: an RTP timestamp to apply
+ * @out_buf: a #GstBuffer
+ *
+ * Push @out_buf to the peer of @filter. This function takes ownership of
+ * @out_buf.
+ *
+ * Unlike gst_base_rtp_depayload_push(), this function will apply @timestamp
+ * on the outgoing buffer, using the configured clock_rate to convert the
+ * timestamp to a valid GStreamer clock time.
+ *
+ * Return: a #GstFlowReturn.
+ */
+GstFlowReturn
+gst_base_rtp_depayload_push_ts (GstBaseRTPDepayload * filter, guint32 timestamp,
+    GstBuffer * out_buf)
+{
+  return gst_base_rtp_depayload_push_full (filter, TRUE, timestamp, out_buf);
+}
+
+/**
+ * gst_base_rtp_depayload_push:
+ * @filter: a #GstBaseRTPDepayload
+ * @out_buf: a #GstBuffer
+ *
+ * Push @out_buf to the peer of @filter. This function takes ownership of
+ * @out_buf.
+ *
+ * Unlike gst_base_rtp_depayload_push_ts(), this function will not apply
+ * any timestamp on the outgoing buffer.
+ *
+ * Return: a #GstFlowReturn.
+ */
+GstFlowReturn
+gst_base_rtp_depayload_push (GstBaseRTPDepayload * filter, GstBuffer * out_buf)
+{
+  return gst_base_rtp_depayload_push_full (filter, FALSE, 0, out_buf);
+}
+
+static GstFlowReturn
+gst_base_rtp_depayload_process (GstBaseRTPDepayload * filter,
+    GstBuffer * rtp_buf)
+{
+  GstBaseRTPDepayloadClass *bclass;
+  GstBuffer *out_buf;
+  GstFlowReturn ret = GST_FLOW_OK;
+
+  bclass = GST_BASE_RTP_DEPAYLOAD_GET_CLASS (filter);
 
   /* let's send it out to processing */
   out_buf = bclass->process (filter, rtp_buf);
   if (out_buf) {
-    /* set the caps */
-    srccaps = GST_PAD_CAPS (filter->srcpad);
+    guint32 timestamp = gst_rtp_buffer_get_timestamp (rtp_buf);
 
-    if (srccaps)
-      gst_buffer_set_caps (GST_BUFFER (out_buf), srccaps);
-
-    /* set the timestamp
-     * I am assuming here that the timestamp of the last RTP buffer
-     * is the same as the timestamp wanted on the collector
-     * maybe i should add a way to override this timestamp from the
-     * depayloader child class
+    /* push buffer with timestamp 
+     * We are assuming here that the timestamp of the last RTP buffer
+     * is the same as the timestamp wanted on the collector. If this is not a
+     * desired result, the process function should push itself with another
+     * timestamp and return NULL.
      */
-    bclass->set_gst_timestamp (filter, gst_rtp_buffer_get_timestamp (rtp_buf),
-        out_buf);
-    /* push it */
-    GST_LOG_OBJECT (filter, "Pushing buffer size %d, timestamp %u",
-        GST_BUFFER_SIZE (out_buf), GST_BUFFER_TIMESTAMP (out_buf));
-    gst_pad_push (filter->srcpad, GST_BUFFER (out_buf));
-    GST_LOG_OBJECT (filter, "Pushed buffer");
+    ret = gst_base_rtp_depayload_push_ts (filter, timestamp, out_buf);
   }
   gst_buffer_unref (rtp_buf);
+
+  return ret;
 }
 
 static void
 gst_base_rtp_depayload_set_gst_timestamp (GstBaseRTPDepayload * filter,
     guint32 timestamp, GstBuffer * buf)
 {
-  guint64 ts = ((timestamp * GST_SECOND) / filter->clock_rate);
+  GstClockTime ts, adjusted;
+
+  ts = gst_util_uint64_scale_int (timestamp, GST_SECOND, filter->clock_rate);
 
   /* rtp timestamps are based on the clock_rate
    * gst timesamps are in nanoseconds
    */
-  GST_DEBUG_OBJECT (filter, "calculating ts : timestamp : %u, clockrate : %u",
+  GST_DEBUG_OBJECT (filter, "ts : timestamp : %u, clockrate : %u",
       timestamp, filter->clock_rate);
 
   /* add delay to timestamp */
-  GST_BUFFER_TIMESTAMP (buf) = ts + (filter->queue_delay * GST_MSECOND);
+  adjusted = ts + (filter->queue_delay * GST_MSECOND);
 
-  GST_DEBUG_OBJECT (filter, "calculated ts %"
-      GST_TIME_FORMAT, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
+  GST_DEBUG_OBJECT (filter, "RTP: %u, GST: %" GST_TIME_FORMAT ", adjusted %"
+      GST_TIME_FORMAT, timestamp, GST_TIME_ARGS (ts), GST_TIME_ARGS (adjusted));
 
-  /* if this is the first buf send a discont */
+  GST_BUFFER_TIMESTAMP (buf) = adjusted;
+
+  /* if this is the first buf send a NEWSEGMENT */
   if (filter->need_newsegment) {
-    /* send discont */
     GstEvent *event = gst_event_new_new_segment (FALSE, 1.0, GST_FORMAT_TIME,
-        GST_BUFFER_TIMESTAMP (buf), GST_CLOCK_TIME_NONE, 0);
+        adjusted, GST_CLOCK_TIME_NONE, 0);
 
     gst_pad_push_event (filter->srcpad, event);
     filter->need_newsegment = FALSE;
@@ -397,7 +504,7 @@ gst_base_rtp_depayload_queue_release (GstBaseRTPDepayload * filter)
     if (bclass->process) {
       GstBuffer *in = g_queue_pop_head (queue);
 
-      gst_base_rtp_depayload_push (filter, in);
+      gst_base_rtp_depayload_process (filter, in);
     }
     headts =
         gst_rtp_buffer_get_timestamp (GST_BUFFER (g_queue_peek_head (queue)));
@@ -452,17 +559,33 @@ static void
 gst_base_rtp_depayload_wait (GstBaseRTPDepayload * filter, GstClockTime time)
 {
   GstClockID id;
+  GstClock *clock;
+  GstClockTime base;
 
   g_return_if_fail (GST_CLOCK_TIME_IS_VALID (time));
-  if (GST_ELEMENT_CLOCK (filter) == NULL) {
-    GST_DEBUG_OBJECT (filter, "No clock given yet");
-    return;
-  }
 
-  id = gst_clock_new_single_shot_id (GST_ELEMENT_CLOCK (filter), time);
+  GST_OBJECT_LOCK (filter);
+  if ((clock = GST_ELEMENT_CLOCK (filter)) == NULL)
+    goto no_clock;
+  gst_object_ref (clock);
+  GST_OBJECT_UNLOCK (filter);
+
+  base = gst_clock_get_time (clock);
+  id = gst_clock_new_single_shot_id (clock, base + time);
+
+  gst_object_unref (clock);
 
   gst_clock_id_wait (id, NULL);
   gst_clock_id_unref (id);
+
+  return;
+
+no_clock:
+  {
+    GST_DEBUG_OBJECT (filter, "No clock given yet");
+    GST_OBJECT_UNLOCK (filter);
+    return;
+  }
 }
 
 static GstStateChangeReturn
@@ -470,15 +593,13 @@ gst_base_rtp_depayload_change_state (GstElement * element,
     GstStateChange transition)
 {
   GstBaseRTPDepayload *filter;
+  GstStateChangeReturn ret;
 
-  g_return_val_if_fail (GST_IS_BASE_RTP_DEPAYLOAD (element),
-      GST_STATE_CHANGE_FAILURE);
   filter = GST_BASE_RTP_DEPAYLOAD (element);
 
   /* we disallow changing the state from the thread */
   if (g_thread_self () == filter->thread)
-    return GST_STATE_CHANGE_FAILURE;
-
+    goto wrong_thread;
 
   switch (transition) {
     case GST_STATE_CHANGE_NULL_TO_READY:
@@ -493,6 +614,8 @@ gst_base_rtp_depayload_change_state (GstElement * element,
       break;
   }
 
+  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
+
   switch (transition) {
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       break;
@@ -504,11 +627,18 @@ gst_base_rtp_depayload_change_state (GstElement * element,
     default:
       break;
   }
-  return GST_STATE_CHANGE_SUCCESS;
+  return ret;
 
   /* ERRORS */
+wrong_thread:
+  {
+    GST_ELEMENT_ERROR (filter, CORE, STATE_CHANGE,
+        (NULL), ("cannot perform a state change from this thread"));
+    return GST_STATE_CHANGE_FAILURE;
+  }
 start_failed:
   {
+    /* start method should have posted an error message */
     return GST_STATE_CHANGE_FAILURE;
   }
 }
@@ -519,7 +649,6 @@ gst_base_rtp_depayload_set_property (GObject * object, guint prop_id,
 {
   GstBaseRTPDepayload *filter;
 
-  g_return_if_fail (GST_IS_BASE_RTP_DEPAYLOAD (object));
   filter = GST_BASE_RTP_DEPAYLOAD (object);
 
   switch (prop_id) {
@@ -538,7 +667,6 @@ gst_base_rtp_depayload_get_property (GObject * object, guint prop_id,
 {
   GstBaseRTPDepayload *filter;
 
-  g_return_if_fail (GST_IS_BASE_RTP_DEPAYLOAD (object));
   filter = GST_BASE_RTP_DEPAYLOAD (object);
 
   switch (prop_id) {
