@@ -55,7 +55,7 @@ static GType
 gst_stream_type_get_type (void)
 {
   static GType stream_type_type = 0;
-  static GEnumValue stream_type[] = {
+  static const GEnumValue stream_type[] = {
     {GST_STREAM_TYPE_UNKNOWN, "Unknown stream", "unknown"},
     {GST_STREAM_TYPE_AUDIO, "Audio stream", "audio"},
     {GST_STREAM_TYPE_VIDEO, "Video stream", "video"},
@@ -120,7 +120,7 @@ gst_stream_info_class_init (GstStreamInfoClass * klass)
 
   gobject_klass = (GObjectClass *) klass;
 
-  parent_class = g_type_class_ref (G_TYPE_OBJECT);
+  parent_class = g_type_class_peek_parent (klass);
 
   gobject_klass->set_property = gst_stream_info_set_property;
   gobject_klass->get_property = gst_stream_info_get_property;
@@ -187,15 +187,23 @@ cb_probe (GstPad * pad, GstEvent * e, gpointer user_data)
     if (gst_tag_list_get_string (list, GST_TAG_VIDEO_CODEC, &codec)) {
       g_free (info->codec);
       info->codec = codec;
+      GST_LOG_OBJECT (pad, "codec = %s (video)", codec);
       g_object_notify (G_OBJECT (info), "codec");
     } else if (gst_tag_list_get_string (list, GST_TAG_AUDIO_CODEC, &codec)) {
       g_free (info->codec);
       info->codec = codec;
+      GST_LOG_OBJECT (pad, "codec = %s (audio)", codec);
+      g_object_notify (G_OBJECT (info), "codec");
+    } else if (gst_tag_list_get_string (list, GST_TAG_CODEC, &codec)) {
+      g_free (info->codec);
+      info->codec = codec;
+      GST_LOG_OBJECT (pad, "codec = %s (generic)", codec);
       g_object_notify (G_OBJECT (info), "codec");
     }
     if (gst_tag_list_get_string (list, GST_TAG_LANGUAGE_CODE, &lang)) {
       g_free (info->langcode);
       info->langcode = lang;
+      GST_LOG_OBJECT (pad, "language-code = %s", lang);
       g_object_notify (G_OBJECT (info), "language-code");
     }
   }
@@ -241,7 +249,7 @@ gst_stream_info_dispose (GObject * object)
         GST_PAD_CAST (stream_info->object));
     if (parent != NULL) {
       g_signal_handlers_disconnect_by_func (parent,
-          G_CALLBACK (stream_info_change_state), stream_info);
+          (gpointer) stream_info_change_state, stream_info);
       gst_object_unref (parent);
     }
 
