@@ -37,7 +37,7 @@
 GST_DEBUG_CATEGORY_STATIC (stream_selector_debug);
 #define GST_CAT_DEFAULT stream_selector_debug
 
-static GstElementDetails gst_stream_selector_details =
+static const GstElementDetails gst_stream_selector_details =
 GST_ELEMENT_DETAILS ("StreamSelector",
     "Generic",
     "N-to-1 input stream_selectoring",
@@ -136,7 +136,7 @@ gst_stream_selector_class_init (GstStreamSelectorClass * klass)
   gobject_class->get_property =
       GST_DEBUG_FUNCPTR (gst_stream_selector_get_property);
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_ACTIVE_PAD,
+  g_object_class_install_property (gobject_class, PROP_ACTIVE_PAD,
       g_param_spec_string ("active-pad", "Active pad", "Name of the currently"
           " active sink pad", NULL, G_PARAM_READWRITE));
 
@@ -167,7 +167,10 @@ gst_stream_selector_dispose (GObject * object)
 {
   GstStreamSelector *sel = GST_STREAM_SELECTOR (object);
 
-  gst_object_unref (sel->active_sinkpad);
+  if (sel->active_sinkpad) {
+    gst_object_unref (sel->active_sinkpad);
+    sel->active_sinkpad = NULL;
+  }
 
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
@@ -354,11 +357,11 @@ gst_stream_selector_request_new_pad (GstElement * element,
 
   GST_LOG_OBJECT (sel, "Creating new pad %d", sel->nb_sinkpads);
 
+  GST_OBJECT_LOCK (sel);
   name = g_strdup_printf ("sink%d", sel->nb_sinkpads++);
   sinkpad = gst_pad_new_from_template (templ, name);
   g_free (name);
 
-  GST_OBJECT_LOCK (sel);
   if (sel->active_sinkpad == NULL)
     sel->active_sinkpad = gst_object_ref (sinkpad);
   GST_OBJECT_UNLOCK (sel);
@@ -372,6 +375,7 @@ gst_stream_selector_request_new_pad (GstElement * element,
   gst_pad_set_bufferalloc_function (sinkpad,
       GST_DEBUG_FUNCPTR (gst_stream_selector_bufferalloc));
 
+  gst_pad_set_active (sinkpad, TRUE);
   gst_element_add_pad (GST_ELEMENT (sel), sinkpad);
 
   return sinkpad;
