@@ -1,6 +1,7 @@
 /* GStreamer
  * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
- * Copyright (c) 2004 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
+ * Copyright (C) 2004 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
+ * Copyright (C) 2006 Tim-Philipp Müller <tim centricular net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -29,9 +30,26 @@
 
 #include "gstsubparse.h"
 #include "gstssaparse.h"
+#include "samiparse.h"
+#include "tmplayerparse.h"
+
+GST_DEBUG_CATEGORY (sub_parse_debug);
+
+#define DEFAULT_ENCODING   NULL
+
+enum
+{
+  PROP_0,
+  PROP_ENCODING
+};
+
+static void
+gst_sub_parse_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec);
+static void
+gst_sub_parse_get_property (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec);
 
-GST_DEBUG_CATEGORY_STATIC (sub_parse_debug);
-#define GST_CAT_DEFAULT sub_parse_debug
 
 static const GstElementDetails sub_parse_details =
 GST_ELEMENT_DETAILS ("Subtitle parser",
@@ -40,11 +58,20 @@ GST_ELEMENT_DETAILS ("Subtitle parser",
     "Gustavo J. A. M. Carneiro <gjc@inescporto.pt>\n"
     "Ronald S. Bultje <rbultje@ronald.bitfreak.net>");
 
+#ifndef GST_DISABLE_LOADSAVE_REGISTRY
+static GstStaticPadTemplate sink_templ = GST_STATIC_PAD_TEMPLATE ("sink",
+    GST_PAD_SINK,
+    GST_PAD_ALWAYS,
+    GST_STATIC_CAPS ("application/x-subtitle; application/x-subtitle-sami; "
+        "application/x-subtitle-tmplayer")
+    );
+#else
 static GstStaticPadTemplate sink_templ = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("application/x-subtitle")
     );
+#endif
 
 static GstStaticPadTemplate src_templ = GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
@@ -114,6 +141,15 @@ gst_sub_parse_dispose (GObject * object)
     gst_segment_free (subparse->segment);
     subparse->segment = NULL;
   }
+  if (subparse->encoding) {
+    g_free (subparse->encoding);
+    subparse->encoding = NULL;
+  }
+  if (subparse->textbuf) {
+    g_string_free (subparse->textbuf, TRUE);
+    subparse->textbuf = NULL;
+  }
+  sami_context_deinit (&subparse->state);
 
   GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
 }
@@ -127,8 +163,17 @@ gst_sub_parse_class_init (GstSubParseClass * klass)
   parent_class = g_type_class_peek_parent (klass);
 
   object_class->dispose = gst_sub_parse_dispose;
+  object_class->set_property = gst_sub_parse_set_property;
+  object_class->get_property = gst_sub_parse_get_property;
 
   element_class->change_state = gst_sub_parse_change_state;
+
+  g_object_class_install_property (object_class, PROP_ENCODING,
+      g_param_spec_string ("subtitle-encoding", "subtitle charset encoding",
+          "Encoding to assume if input subtitles are not in UTF-8 encoding. "
+          "If not set, the GST_SUBTITLE_ENCODING environment variable will "
+          "be checked for an encoding to use. If that is not set either, "
+          "ISO-8859-15 will be assumed.", DEFAULT_ENCODING, G_PARAM_READWRITE));
 }
 
 static void
@@ -157,6 +202,7 @@ gst_sub_parse_init (GstSubParse * subparse)
     GST_WARNING_OBJECT (subparse, "segment creation failed");
     g_assert_not_reached ();
   }
+  subparse->encoding = g_strdup (DEFAULT_ENCODING);
 }
 
 /*
@@ -227,6 +273,45 @@ beach:
   return ret;
 }
 
+static void
+gst_sub_parse_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec)
+{
+  GstSubParse *subparse = GST_SUBPARSE (object);
+
+  GST_OBJECT_LOCK (subparse);
+  switch (prop_id) {
+    case PROP_ENCODING:
+      g_free (subparse->encoding);
+      subparse->encoding = g_value_dup_string (value);
+      GST_LOG_OBJECT (object, "subtitle encoding set to %s",
+          GST_STR_NULL (subparse->encoding));
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+  GST_OBJECT_UNLOCK (subparse);
+}
+
+static void
+gst_sub_parse_get_property (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec)
+{
+  GstSubParse *subparse = GST_SUBPARSE (object);
+
+  GST_OBJECT_LOCK (subparse);
+  switch (prop_id) {
+    case PROP_ENCODING:
+      g_value_set_string (value, subparse->encoding);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+  GST_OBJECT_UNLOCK (subparse);
+}
+
 static gchar *
 convert_encoding (GstSubParse * self, const gchar * str, gsize len)
 {
@@ -243,7 +328,10 @@ convert_encoding (GstSubParse * self, const gchar * str, gsize len)
     self->valid_utf8 = FALSE;
   }
 
-  encoding = g_getenv ("GST_SUBTITLE_ENCODING");
+  encoding = self->encoding;
+  if (encoding == NULL || *encoding == '\0') {
+    encoding = g_getenv ("GST_SUBTITLE_ENCODING");
+  }
   if (encoding == NULL || *encoding == '\0') {
     /* if local encoding is UTF-8 and no encoding specified
      * via the environment variable, assume ISO-8859-15 */
@@ -265,7 +353,8 @@ convert_encoding (GstSubParse * self, const gchar * str, gsize len)
         NULL, NULL, NULL);
   }
 
-  GST_LOG_OBJECT (self, "successfully converted %d characters from %s to UTF-8"
+  GST_LOG_OBJECT (self,
+      "successfully converted %" G_GSIZE_FORMAT " characters from %s to UTF-8"
       "%s", len, encoding, (err) ? " , using ISO-8859-15 as fallback" : "");
 
   return ret;
@@ -390,6 +479,20 @@ parse_mdvdsub (ParserState * state, const gchar * line)
   return ret;
 }
 
+static void
+strip_trailing_newlines (gchar * txt)
+{
+  if (txt) {
+    guint len;
+
+    len = strlen (txt);
+    while (len > 1 && txt[len - 1] == '\n') {
+      txt[len - 1] = '\0';
+      --len;
+    }
+  }
+}
+
 /* we want to escape text in general, but retain basic markup like
  * <i></i>, <u></u>, and <b></b>. The easiest and safest way is to
  * just unescape a white list of allowed markups again after
@@ -428,6 +531,71 @@ subrip_unescape_formatting (gchar * txt)
   }
 }
 
+/* we only allow <i>, <u> and <b>, so let's take a simple approach. This code
+ * assumes the input has been escaped and subrip_unescape_formatting() has then
+ * been run over the input! This function adds missing closing markup tags and
+ * removes broken closing tags for tags that have never been opened. */
+static void
+subrip_fix_up_markup (gchar ** p_txt)
+{
+  gchar *cur, *next_tag;
+  gchar open_tags[32];
+  guint num_open_tags = 0;
+
+  g_assert (*p_txt != NULL);
+
+  cur = *p_txt;
+  while (*cur != '\0') {
+    next_tag = strchr (cur, '<');
+    if (next_tag == NULL)
+      break;
+    ++next_tag;
+    switch (*next_tag) {
+      case '/':{
+        ++next_tag;
+        if (num_open_tags == 0 || open_tags[num_open_tags - 1] != *next_tag) {
+          GST_LOG ("broken input, closing tag '%c' is not open", *next_tag);
+          g_memmove (next_tag - 2, next_tag + 2, strlen (next_tag + 2) + 1);
+          next_tag -= 2;
+        } else {
+          /* it's all good, closing tag which is open */
+          --num_open_tags;
+        }
+        break;
+      }
+      case 'i':
+      case 'b':
+      case 'u':
+        if (num_open_tags == G_N_ELEMENTS (open_tags))
+          return;               /* something dodgy is going on, stop parsing */
+        open_tags[num_open_tags] = *next_tag;
+        ++num_open_tags;
+        break;
+      default:
+        GST_ERROR ("unexpected tag '%c' (%s)", *next_tag, next_tag);
+        g_assert_not_reached ();
+        break;
+    }
+    cur = next_tag;
+  }
+
+  if (num_open_tags > 0) {
+    GString *s;
+
+    s = g_string_new (*p_txt);
+    while (num_open_tags > 0) {
+      GST_LOG ("adding missing closing tag '%c'", open_tags[num_open_tags - 1]);
+      g_string_append_c (s, '<');
+      g_string_append_c (s, '/');
+      g_string_append_c (s, open_tags[num_open_tags - 1]);
+      g_string_append_c (s, '>');
+      --num_open_tags;
+    }
+    g_free (*p_txt);
+    *p_txt = g_string_free (s, FALSE);
+  }
+}
+
 static gchar *
 parse_subrip (ParserState * state, const gchar * line)
 {
@@ -486,6 +654,8 @@ parse_subrip (ParserState * state, const gchar * line)
         g_string_truncate (state->buf, 0);
         state->state = 0;
         subrip_unescape_formatting (ret);
+        strip_trailing_newlines (ret);
+        subrip_fix_up_markup (&ret);
         return ret;
       }
       return NULL;
@@ -569,6 +739,9 @@ parser_state_dispose (ParserState * state)
     g_string_free (state->buf, TRUE);
     state->buf = NULL;
   }
+  if (state->user_data) {
+    sami_context_reset (state);
+  }
 }
 
 /*
@@ -583,6 +756,7 @@ gst_sub_parse_data_format_autodetect (gchar * match_str)
   static gboolean need_init_regexps = TRUE;
   static regex_t mdvd_rx;
   static regex_t subrip_rx;
+  guint n1, n2, n3;
 
   /* initialize the regexps used the first time around */
   if (need_init_regexps) {
@@ -592,7 +766,7 @@ gst_sub_parse_data_format_autodetect (gchar * match_str)
     need_init_regexps = FALSE;
     if ((err = regcomp (&mdvd_rx, "^\\{[0-9]+\\}\\{[0-9]+\\}",
                 REG_EXTENDED | REG_NEWLINE | REG_NOSUB) != 0) ||
-        (err = regcomp (&subrip_rx, "^1(\x0d)?\x0a"
+        (err = regcomp (&subrip_rx, "^[1-9]([0-9]){0,3}(\x0d)?\x0a"
                 "[0-9][0-9]:[0-9][0-9]:[0-9][0-9],[0-9]{3}"
                 " --> [0-9][0-9]:[0-9][0-9]:[0-9][0-9],[0-9]{3}",
                 REG_EXTENDED | REG_NEWLINE | REG_NOSUB)) != 0) {
@@ -602,19 +776,33 @@ gst_sub_parse_data_format_autodetect (gchar * match_str)
   }
 
   if (regexec (&mdvd_rx, match_str, 0, NULL, 0) == 0) {
-    GST_LOG ("subparse: MicroDVD (frame based) format detected");
+    GST_LOG ("MicroDVD (frame based) format detected");
     return GST_SUB_PARSE_FORMAT_MDVDSUB;
   }
   if (regexec (&subrip_rx, match_str, 0, NULL, 0) == 0) {
-    GST_LOG ("subparse: SubRip (time based) format detected");
+    GST_LOG ("SubRip (time based) format detected");
     return GST_SUB_PARSE_FORMAT_SUBRIP;
   }
   if (!strncmp (match_str, "FORMAT=TIME", 11)) {
-    GST_LOG ("subparse: MPSub (time based) format detected");
+    GST_LOG ("MPSub (time based) format detected");
     return GST_SUB_PARSE_FORMAT_MPSUB;
   }
+  if (strstr (match_str, "<SAMI>") != NULL ||
+      strstr (match_str, "<sami>") != NULL) {
+    GST_LOG ("SAMI (time based) format detected");
+    return GST_SUB_PARSE_FORMAT_SAMI;
+  }
+  /* we're boldly assuming the first subtitle appears within the first hour */
+  if (sscanf (match_str, "0:%02u:%02u:", &n1, &n2) == 2 ||
+      sscanf (match_str, "0:%02u:%02u=", &n1, &n2) == 2 ||
+      sscanf (match_str, "00:%02u:%02u:", &n1, &n2) == 2 ||
+      sscanf (match_str, "00:%02u:%02u=", &n1, &n2) == 2 ||
+      sscanf (match_str, "00:%02u:%02u,%u=", &n1, &n2, &n3) == 3) {
+    GST_LOG ("TMPlayer (time based) format detected");
+    return GST_SUB_PARSE_FORMAT_TMPLAYER;
+  }
 
-  GST_WARNING ("subparse: subtitle format autodetection failed!");
+  GST_DEBUG ("no subtitle format detected");
   return GST_SUB_PARSE_FORMAT_UNKNOWN;
 }
 
@@ -646,6 +834,13 @@ gst_sub_parse_format_autodetect (GstSubParse * self)
     case GST_SUB_PARSE_FORMAT_MPSUB:
       self->parse_line = parse_mpsub;
       return gst_caps_new_simple ("text/plain", NULL);
+    case GST_SUB_PARSE_FORMAT_SAMI:
+      self->parse_line = parse_sami;
+      sami_context_init (&self->state);
+      return gst_caps_new_simple ("text/x-pango-markup", NULL);
+    case GST_SUB_PARSE_FORMAT_TMPLAYER:
+      self->parse_line = parse_tmplayer;
+      return gst_caps_new_simple ("text/plain", NULL);
     case GST_SUB_PARSE_FORMAT_UNKNOWN:
     default:
       GST_DEBUG ("no subtitle format detected");
@@ -662,6 +857,7 @@ feed_textbuf (GstSubParse * self, GstBuffer * buf)
     /* flush the parser state */
     parser_state_init (&self->state);
     g_string_truncate (self->textbuf, 0);
+    sami_context_reset (&self->state);
   }
 
   self->textbuf = g_string_append_len (self->textbuf,
@@ -697,26 +893,29 @@ handle_buffer (GstSubParse * self, GstBuffer * buf)
     /* Set segment on our parser state machine */
     self->state.segment = self->segment;
     /* Now parse the line, out of segment lines will just return NULL */
-    GST_DEBUG ("Parsing line '%s'", line);
+    GST_LOG_OBJECT (self, "Parsing line '%s'", line);
     subtitle = self->parse_line (&self->state, line);
     g_free (line);
 
     if (subtitle) {
       guint subtitle_len = strlen (subtitle);
 
+      /* +1 for terminating NUL character */
       ret = gst_pad_alloc_buffer_and_set_caps (self->srcpad,
-          GST_BUFFER_OFFSET_NONE, subtitle_len,
+          GST_BUFFER_OFFSET_NONE, subtitle_len + 1,
           GST_PAD_CAPS (self->srcpad), &buf);
 
       if (ret == GST_FLOW_OK) {
-        memcpy (GST_BUFFER_DATA (buf), subtitle, subtitle_len);
+        /* copy terminating NUL character as well */
+        memcpy (GST_BUFFER_DATA (buf), subtitle, subtitle_len + 1);
+        GST_BUFFER_SIZE (buf) = subtitle_len;
         GST_BUFFER_TIMESTAMP (buf) = self->state.start_time;
         GST_BUFFER_DURATION (buf) = self->state.duration;
 
         gst_segment_set_last_stop (self->segment, GST_FORMAT_TIME,
             self->state.start_time);
 
-        GST_DEBUG ("Sending text '%s', %" GST_TIME_FORMAT " + %"
+        GST_DEBUG_OBJECT (self, "Sending text '%s', %" GST_TIME_FORMAT " + %"
             GST_TIME_FORMAT, subtitle, GST_TIME_ARGS (self->state.start_time),
             GST_TIME_ARGS (self->state.duration));
 
@@ -740,8 +939,7 @@ gst_sub_parse_chain (GstPad * sinkpad, GstBuffer * buf)
   GstFlowReturn ret;
   GstSubParse *self;
 
-  GST_DEBUG ("gst_sub_parse_chain");
-  self = GST_SUBPARSE (gst_pad_get_parent (sinkpad));
+  self = GST_SUBPARSE (GST_PAD_PARENT (sinkpad));
 
   /* Push newsegment if needed */
   if (self->need_segment) {
@@ -754,8 +952,6 @@ gst_sub_parse_chain (GstPad * sinkpad, GstBuffer * buf)
 
   ret = handle_buffer (self, buf);
 
-  gst_object_unref (self);
-
   return ret;
 }
 
@@ -838,9 +1034,11 @@ gst_sub_parse_change_state (GstElement * element, GstStateChange transition)
   switch (transition) {
     case GST_STATE_CHANGE_READY_TO_PAUSED:
       /* format detection will init the parser state */
-      self->offset = self->next_offset = 0;
+      self->offset = 0;
+      self->next_offset = 0;
       self->parser_type = GST_SUB_PARSE_FORMAT_UNKNOWN;
       self->valid_utf8 = TRUE;
+      g_string_truncate (self->textbuf, 0);
       break;
     default:
       break;
@@ -866,15 +1064,23 @@ gst_sub_parse_change_state (GstElement * element, GstStateChange transition)
  * Typefind support.
  */
 
+/* FIXME 0.11: these caps are ugly, use app/x-subtitle + type field or so;
+ * also, give different  subtitle formats really different types */
+static GstStaticCaps tmp_caps =
+GST_STATIC_CAPS ("application/x-subtitle-tmplayer");
+static GstStaticCaps smi_caps = GST_STATIC_CAPS ("application/x-subtitle-sami");
 static GstStaticCaps sub_caps = GST_STATIC_CAPS ("application/x-subtitle");
 
 #define SUB_CAPS (gst_static_caps_get (&sub_caps))
+#define SAMI_CAPS (gst_static_caps_get (&smi_caps))
+#define TMP_CAPS (gst_static_caps_get (&tmp_caps))
 
 static void
 gst_subparse_type_find (GstTypeFind * tf, gpointer private)
 {
-  const guint8 *data;
   GstSubParseFormat format;
+  const guint8 *data;
+  GstCaps *caps;
   gchar *str;
 
   if (!(data = gst_type_find_peek (tf, 0, 36)))
@@ -888,26 +1094,40 @@ gst_subparse_type_find (GstTypeFind * tf, gpointer private)
   switch (format) {
     case GST_SUB_PARSE_FORMAT_MDVDSUB:
       GST_DEBUG ("MicroDVD format detected");
+      caps = SUB_CAPS;
       break;
     case GST_SUB_PARSE_FORMAT_SUBRIP:
       GST_DEBUG ("SubRip format detected");
+      caps = SUB_CAPS;
       break;
     case GST_SUB_PARSE_FORMAT_MPSUB:
       GST_DEBUG ("MPSub format detected");
+      caps = SUB_CAPS;
       break;
+    case GST_SUB_PARSE_FORMAT_SAMI:
+      GST_DEBUG ("SAMI (time-based) format detected");
+      caps = SAMI_CAPS;
+      break;
+    case GST_SUB_PARSE_FORMAT_TMPLAYER:
+      GST_DEBUG ("TMPlayer (time based) format detected");
+      caps = TMP_CAPS;
+      break;
+    default:
     case GST_SUB_PARSE_FORMAT_UNKNOWN:
       GST_DEBUG ("no subtitle format detected");
       return;
   }
 
   /* if we're here, it's ok */
-  gst_type_find_suggest (tf, GST_TYPE_FIND_MAXIMUM, SUB_CAPS);
+  gst_type_find_suggest (tf, GST_TYPE_FIND_MAXIMUM, caps);
 }
 
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
-  static gchar *sub_exts[] = { "srt", "sub", "mpsub", "mdvd", NULL };
+  static gchar *sub_exts[] = { "srt", "sub", "mpsub", "mdvd", "smi", "txt",
+    NULL
+  };
 
   GST_DEBUG_CATEGORY_INIT (sub_parse_debug, "subparse", 0, ".sub parser");
 
