@@ -19,21 +19,28 @@
 
 /**
  * SECTION:element-videorate
- * @short_description: adjusts the framerate of video
+ * @short_description: retimestamps and drops/duplicates video frames to
+ *  match the source pad's framerate and create a perfect stream
  *
  * <refsect2>
  * <para>
- * This element converts video from one framerate to another. This operation
- * is performed by dropping and duplicating frames, no fancy algorithm is
- * used to interpolate frames (yet).
+ * This element takes an incoming stream of timestamped video frames.
+ * It will produce a perfect stream that matches the source pad's framerate.
+ *
+ * The correction is performed by dropping and duplicating frames, no fancy
+ * algorithm is used to interpolate frames (yet).
  * </para>
  * <para>
  * By default the element will simply negotiate the same framerate on its
- * source and sink pad and will adjust timestamps/insert/drop frames in case
- * the input stream is not respecting that framerate.
+ * source and sink pad.
  * </para>
  * <para>
- * A conversion to another framerate can be forced by using filtered caps on
+ * This operation is useful to link to elements that require a perfect stream.
+ * Typical examples are formats that do not store timestamps for video frames,
+ * but only store a framerate, like Ogg and AVI.
+ * </para>
+ * <para>
+ * A conversion to a specific framerate can be forced by using filtered caps on
  * the source pad.
  * </para>
  * <para>
@@ -59,9 +66,16 @@
  * Decode an Ogg/Theora file and adjust the framerate to 15 fps before playing.
  * To create the test Ogg/Theora file refer to the documentation of theoraenc.
  * </para>
- * </refsect2>
+ * <para>
+ * <programlisting>
+ * gst-launch -v v4lsrc ! videorate ! video/x-raw-yuv,framerate=25/2 ! theoraenc ! oggmux ! filesink location=v4l.ogg
+ * </programlisting>
+ * Capture video from a V4L device, and adjust the stream to 12.5 fps before
+ * encoding to Ogg/Theora.
+ * </para>
+  * </refsect2>
  *
- * Last reviewed on 2006-03-02 (0.10.4)
+ * Last reviewed on 2006-09-02 (0.10.11)
  */
 
 #ifdef HAVE_CONFIG_H
@@ -70,7 +84,7 @@
 
 #include "gstvideorate.h"
 
-GST_DEBUG_CATEGORY (video_rate_debug);
+GST_DEBUG_CATEGORY_STATIC (video_rate_debug);
 #define GST_CAT_DEFAULT video_rate_debug
 
 /* elementfactory information */
@@ -421,19 +435,20 @@ gst_video_rate_flush_prev (GstVideoRate * videorate)
 
   /* this is the timestamp we put on the buffer */
   push_ts = videorate->next_ts;
-  GST_BUFFER_TIMESTAMP (outbuf) = push_ts;
 
   videorate->out++;
   if (videorate->to_rate_numerator) {
-    videorate->next_ts =
+    /* interpolate next expected timestamp in the segment */
+    videorate->next_ts = videorate->segment.accum + videorate->segment.start +
         gst_util_uint64_scale (videorate->out,
         videorate->to_rate_denominator * GST_SECOND,
         videorate->to_rate_numerator);
-    GST_BUFFER_DURATION (outbuf) =
-        videorate->next_ts - GST_BUFFER_TIMESTAMP (outbuf);
+    GST_BUFFER_DURATION (outbuf) = videorate->next_ts - push_ts;
   }
+
   /* adapt for looping, bring back to time in current segment. */
-  GST_BUFFER_TIMESTAMP (outbuf) -= videorate->segment.accum;
+  GST_BUFFER_TIMESTAMP (outbuf) = push_ts - videorate->segment.accum;
+
   gst_buffer_set_caps (outbuf, GST_PAD_CAPS (videorate->srcpad));
 
   GST_LOG_OBJECT (videorate,
@@ -463,7 +478,7 @@ gst_video_rate_swap_prev (GstVideoRate * videorate, GstBuffer * buffer,
   videorate->prev_ts = time;
 }
 
-static GstFlowReturn
+static gboolean
 gst_video_rate_event (GstPad * pad, GstEvent * event)
 {
   GstVideoRate *videorate;
@@ -475,25 +490,24 @@ gst_video_rate_event (GstPad * pad, GstEvent * event)
     case GST_EVENT_NEWSEGMENT:
     {
       gint64 start, stop, time;
-      gdouble rate;
+      gdouble rate, arate;
       gboolean update;
       GstFormat format;
 
-      gst_event_parse_new_segment (event, &update, &rate, &format, &start,
-          &stop, &time);
+      gst_event_parse_new_segment_full (event, &update, &rate, &arate, &format,
+          &start, &stop, &time);
 
-      if (format != videorate->segment.format)
+      if (format != GST_FORMAT_TIME)
         goto format_error;
 
+      GST_DEBUG_OBJECT (videorate, "handle NEWSEGMENT");
+
       /* We just want to update the accumulated stream_time  */
-      gst_segment_set_newsegment (&videorate->segment, update, rate,
+      gst_segment_set_newsegment_full (&videorate->segment, update, rate, arate,
           format, start, stop, time);
 
-      GST_DEBUG_OBJECT (videorate, "Updated segment.accum:%" GST_TIME_FORMAT
-          " segment.start:%" GST_TIME_FORMAT " segment.stop:%"
-          GST_TIME_FORMAT, GST_TIME_ARGS (videorate->segment.accum),
-          GST_TIME_ARGS (videorate->segment.start),
-          GST_TIME_ARGS (videorate->segment.stop));
+      GST_DEBUG_OBJECT (videorate, "updated segment: %" GST_SEGMENT_FORMAT,
+          &videorate->segment);
       break;
     }
     case GST_EVENT_EOS:
@@ -547,14 +561,17 @@ gst_video_rate_chain (GstPad * pad, GstBuffer * buffer)
   GST_DEBUG_OBJECT (videorate, "got buffer with timestamp %" GST_TIME_FORMAT,
       GST_TIME_ARGS (in_ts));
 
-  intime = gst_segment_to_running_time (&videorate->segment,
-      GST_FORMAT_TIME, in_ts);
+  /* the input time is the time in the segment + all previously accumulated
+   * segments */
+  intime = in_ts + videorate->segment.accum;
 
   /* we need to have two buffers to compare */
   if (videorate->prevbuf == NULL) {
     gst_video_rate_swap_prev (videorate, buffer, intime);
     videorate->in++;
-    videorate->next_ts = 0;
+    /* new buffer, we expect to output a buffer that matches the first
+     * timestamp in the segment */
+    videorate->next_ts = videorate->segment.start;
   } else {
     GstClockTime prevtime;
     gint count = 0;
