@@ -67,6 +67,7 @@
 
 #include "gstvideoscale.h"
 #include "vs_image.h"
+#include "vs_4tap.h"
 
 
 /* debug variable definition */
@@ -93,6 +94,10 @@ static GstStaticCaps gst_video_scale_format_caps[] = {
   GST_STATIC_CAPS (GST_VIDEO_CAPS_xRGB),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRx),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_xBGR),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_ARGB),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRA),
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_ABGR),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_RGB),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_BGR),
   GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ("AYUV")),
@@ -112,6 +117,10 @@ enum
   GST_VIDEO_SCALE_xRGB,
   GST_VIDEO_SCALE_BGRx,
   GST_VIDEO_SCALE_xBGR,
+  GST_VIDEO_SCALE_RGBA,
+  GST_VIDEO_SCALE_ARGB,
+  GST_VIDEO_SCALE_BGRA,
+  GST_VIDEO_SCALE_ABGR,
   GST_VIDEO_SCALE_RGB,
   GST_VIDEO_SCALE_BGR,
   GST_VIDEO_SCALE_AYUV,
@@ -133,6 +142,7 @@ gst_video_scale_method_get_type (void)
   static const GEnumValue video_scale_methods[] = {
     {GST_VIDEO_SCALE_NEAREST, "Nearest Neighbour", "nearest-neighbour"},
     {GST_VIDEO_SCALE_BILINEAR, "Bilinear", "bilinear"},
+    {GST_VIDEO_SCALE_4TAP, "4-tap", "4-tap"},
     {0, NULL, NULL},
   };
 
@@ -179,6 +189,7 @@ gst_video_scale_sink_template_factory (void)
 static void gst_video_scale_base_init (gpointer g_class);
 static void gst_video_scale_class_init (GstVideoScaleClass * klass);
 static void gst_video_scale_init (GstVideoScale * videoscale);
+static void gst_video_scale_finalize (GstVideoScale * videoscale);
 static gboolean gst_video_scale_src_event (GstBaseTransform * trans,
     GstEvent * event);
 
@@ -249,10 +260,11 @@ gst_video_scale_class_init (GstVideoScaleClass * klass)
   gobject_class = (GObjectClass *) klass;
   trans_class = (GstBaseTransformClass *) klass;
 
+  gobject_class->finalize = (GObjectFinalizeFunc) gst_video_scale_finalize;
   gobject_class->set_property = gst_video_scale_set_property;
   gobject_class->get_property = gst_video_scale_get_property;
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_METHOD,
+  g_object_class_install_property (gobject_class, PROP_METHOD,
       g_param_spec_enum ("method", "method", "method",
           GST_TYPE_VIDEO_SCALE_METHOD, DEFAULT_PROP_METHOD, G_PARAM_READWRITE));
 
@@ -278,6 +290,14 @@ gst_video_scale_init (GstVideoScale * videoscale)
   videoscale->method = DEFAULT_PROP_METHOD;
 }
 
+static void
+gst_video_scale_finalize (GstVideoScale * videoscale)
+{
+  if (videoscale->tmp_buf)
+    g_free (videoscale->tmp_buf);
+
+  G_OBJECT_CLASS (parent_class)->finalize (G_OBJECT (videoscale));
+}
 
 static void
 gst_video_scale_set_property (GObject * object, guint prop_id,
@@ -330,8 +350,8 @@ gst_video_scale_transform_caps (GstBaseTransform * trans,
   structure = gst_caps_get_structure (ret, 0);
 
   gst_structure_set (structure,
-      "width", GST_TYPE_INT_RANGE, 16, 4096,
-      "height", GST_TYPE_INT_RANGE, 16, 4096, NULL);
+      "width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
+      "height", GST_TYPE_INT_RANGE, 1, G_MAXINT, NULL);
 
   /* if pixel aspect ratio, make a range of it */
   if ((par = gst_structure_get_value (structure, "pixel-aspect-ratio"))) {
@@ -390,6 +410,10 @@ gst_video_scale_prepare_size (GstVideoScale * videoscale, gint format,
     case GST_VIDEO_SCALE_xRGB:
     case GST_VIDEO_SCALE_BGRx:
     case GST_VIDEO_SCALE_xBGR:
+    case GST_VIDEO_SCALE_RGBA:
+    case GST_VIDEO_SCALE_ARGB:
+    case GST_VIDEO_SCALE_BGRA:
+    case GST_VIDEO_SCALE_ABGR:
     case GST_VIDEO_SCALE_AYUV:
       img->stride = img->width * 4;
       *size = img->stride * img->height;
@@ -491,7 +515,7 @@ gst_video_scale_set_caps (GstBaseTransform * trans, GstCaps * in, GstCaps * out)
   if (videoscale->tmp_buf)
     g_free (videoscale->tmp_buf);
 
-  videoscale->tmp_buf = g_malloc (videoscale->dest.stride * 2);
+  videoscale->tmp_buf = g_malloc (videoscale->dest.stride * 4);
 
   /* FIXME: par */
   GST_DEBUG_OBJECT (videoscale, "from=%dx%d, size %d -> to=%dx%d, size %d",
@@ -510,7 +534,7 @@ gst_video_scale_get_unit_size (GstBaseTransform * trans, GstCaps * caps,
   gint format, width, height;
   VSImage img;
 
-  g_return_val_if_fail (size, FALSE);
+  g_assert (size);
 
   videoscale = GST_VIDEO_SCALE (trans);
 
@@ -581,7 +605,7 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
     gst_structure_get_int (ins, "height", &from_h);
 
     if (!gst_video_calculate_display_ratio (&num, &den, from_w, from_h,
-        from_par_n, from_par_d, to_par_n, to_par_d)) {
+            from_par_n, from_par_d, to_par_n, to_par_d)) {
       GST_ELEMENT_ERROR (base, CORE, NEGOTIATION, (NULL),
           ("Error calculating the output scaled size - integer overflow"));
       return;
@@ -600,10 +624,10 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
     /* if one of the output width or height is fixed, we work from there */
     if (h) {
       GST_DEBUG_OBJECT (base, "height is fixed,scaling width");
-      w = h * num / den;
+      w = (guint) gst_util_uint64_scale_int (h, num, den);
     } else if (w) {
-      GST_DEBUG_OBJECT (base, "width is fixes, scaling height");
-      h = w * den / num;
+      GST_DEBUG_OBJECT (base, "width is fixed, scaling height");
+      h = (guint) gst_util_uint64_scale_int (w, den, num);
     } else {
       /* none of width or height is fixed, figure out both of them based only on
        * the input width and height */
@@ -611,15 +635,15 @@ gst_video_scale_fixate_caps (GstBaseTransform * base, GstPadDirection direction,
       if (from_h % den == 0) {
         GST_DEBUG_OBJECT (base, "keeping video height");
         h = from_h;
-        w = h * num / den;
+        w = (guint) gst_util_uint64_scale_int (h, num, den);
       } else if (from_w % num == 0) {
         GST_DEBUG_OBJECT (base, "keeping video width");
         w = from_w;
-        h = w * den / num;
+        h = (guint) gst_util_uint64_scale_int (w, den, num);
       } else {
         GST_DEBUG_OBJECT (base, "approximating but keeping video height");
         h = from_h;
-        w = h * num / den;
+        w = (guint) gst_util_uint64_scale_int (h, num, den);
       }
     }
     GST_DEBUG_OBJECT (base, "scaling to %dx%d", w, h);
@@ -659,7 +683,7 @@ gst_video_scale_prepare_image (gint format, GstBuffer * buf,
       img_u->pixels = img->pixels + GST_ROUND_UP_2 (img->height) * img->stride;
       img_u->height = GST_ROUND_UP_2 (img->height) / 2;
       img_u->width = GST_ROUND_UP_2 (img->width) / 2;
-      img_u->stride = GST_ROUND_UP_4 (img->stride / 2);
+      img_u->stride = GST_ROUND_UP_4 (img_u->width);
       memcpy (img_v, img_u, sizeof (*img_v));
       img_v->pixels = img_u->pixels + img_u->height * img_u->stride;
       break;
@@ -700,6 +724,10 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
         case GST_VIDEO_SCALE_xRGB:
         case GST_VIDEO_SCALE_BGRx:
         case GST_VIDEO_SCALE_xBGR:
+        case GST_VIDEO_SCALE_RGBA:
+        case GST_VIDEO_SCALE_ARGB:
+        case GST_VIDEO_SCALE_BGRA:
+        case GST_VIDEO_SCALE_ABGR:
         case GST_VIDEO_SCALE_AYUV:
           vs_image_scale_nearest_RGBA (dest, src, videoscale->tmp_buf);
           break;
@@ -739,6 +767,10 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
         case GST_VIDEO_SCALE_xRGB:
         case GST_VIDEO_SCALE_BGRx:
         case GST_VIDEO_SCALE_xBGR:
+        case GST_VIDEO_SCALE_RGBA:
+        case GST_VIDEO_SCALE_ARGB:
+        case GST_VIDEO_SCALE_BGRA:
+        case GST_VIDEO_SCALE_ABGR:
         case GST_VIDEO_SCALE_AYUV:
           vs_image_scale_linear_RGBA (dest, src, videoscale->tmp_buf);
           break;
@@ -759,8 +791,6 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
         case GST_VIDEO_SCALE_I420:
         case GST_VIDEO_SCALE_YV12:
           vs_image_scale_linear_Y (dest, src, videoscale->tmp_buf);
-          //memset (dest_u.pixels, 128, dest_u.stride * dest_u.height);
-          //memset (dest_v.pixels, 128, dest_v.stride * dest_v.height);
           vs_image_scale_linear_Y (&dest_u, &src_u, videoscale->tmp_buf);
           vs_image_scale_linear_Y (&dest_v, &src_v, videoscale->tmp_buf);
           break;
@@ -774,6 +804,18 @@ gst_video_scale_transform (GstBaseTransform * trans, GstBuffer * in,
           goto unsupported;
       }
       break;
+    case GST_VIDEO_SCALE_4TAP:
+      switch (videoscale->format) {
+        case GST_VIDEO_SCALE_I420:
+        case GST_VIDEO_SCALE_YV12:
+          vs_image_scale_4tap_Y (dest, src, videoscale->tmp_buf);
+          vs_image_scale_4tap_Y (&dest_u, &src_u, videoscale->tmp_buf);
+          vs_image_scale_4tap_Y (&dest_v, &src_v, videoscale->tmp_buf);
+          break;
+        default:
+          goto unsupported;
+      }
+      break;
     default:
       goto unknown_mode;
   }
@@ -850,6 +892,8 @@ plugin_init (GstPlugin * plugin)
   GST_DEBUG_CATEGORY_INIT (video_scale_debug, "videoscale", 0,
       "videoscale element");
 
+  vs_4tap_init ();
+
   return TRUE;
 }
 
