@@ -40,7 +40,6 @@
 
 #include <string.h>
 #include <math.h>
-#include <stdlib.h>
 
 G_BEGIN_DECLS
 
@@ -64,7 +63,7 @@ typedef struct _GstXvImageBufferClass GstXvImageBufferClass;
 typedef struct _GstXvImageSink GstXvImageSink;
 typedef struct _GstXvImageSinkClass GstXvImageSinkClass;
 
-/*
+/**
  * GstXContext:
  * @disp: the X11 Display of this context
  * @screen: the default Screen of Display @disp
@@ -80,7 +79,7 @@ typedef struct _GstXvImageSinkClass GstXvImageSinkClass;
  * @height: the height in pixels of Display @disp
  * @widthmm: the width in millimeters of Display @disp
  * @heightmm: the height in millimeters of Display @disp
- * @par: the pixel aspect ratio calculated from @width, @widthmm and @height,
+ * @par: the pixel aspect ratio calculated from @width, @widthmm and @height, 
  * @heightmm ratio
  * @use_xshm: used to known wether of not XShm extension is usable or not even
  * if the Extension is present
@@ -116,8 +115,6 @@ struct _GstXContext {
   gboolean use_xshm;
 
   XvPortID xv_port_id;
-  guint nb_adaptors;
-  gchar ** adaptors;
   gint im_format;
 
   GList *formats_list;
@@ -128,11 +125,9 @@ struct _GstXContext {
   /* Optimisation storage for buffer_alloc return */
   GstCaps *last_caps;
   gint last_format;
-  gint last_width;
-  gint last_height;
 };
 
-/*
+/**
  * GstXWindow:
  * @win: the Window ID of this X11 window
  * @width: the width in pixels of Window @win
@@ -210,11 +205,10 @@ struct _GstXvImageBuffer {
  * @pool_lock: used to protect the buffer pool
  * @image_pool: a list of #GstXvImageBuffer that could be reused at next buffer
  * allocation call
- * @synchronous: used to store if XSynchronous should be used or not (for
+ * @synchronous: used to store if XSynchronous should be used or not (for 
  * debugging purpose only)
  * @keep_aspect: used to remember if reverse negotiation scaling should respect
  * aspect ratio
- * @handle_events: used to know if we should handle select XEvents or not
  * @brightness: used to store the user settings for color balance brightness
  * @contrast: used to store the user settings for color balance contrast
  * @hue: used to store the user settings for color balance hue
@@ -230,66 +224,38 @@ struct _GstXvImageSink {
   GstVideoSink videosink;
 
   char *display_name;
-  guint adaptor_no;
 
   GstXContext *xcontext;
   GstXWindow *xwindow;
   GstXvImageBuffer *xvimage;
   GstXvImageBuffer *cur_image;
-
+  
   GThread *event_thread;
   gboolean running;
 
   gint fps_n;
   gint fps_d;
-
+  
   GMutex *x_lock;
   GMutex *flow_lock;
 
   /* object-set pixel aspect ratio */
   GValue *par;
-
+  
   GMutex *pool_lock;
-  gboolean pool_invalid;
   GSList *image_pool;
-
+  
   gboolean synchronous;
-  gboolean double_buffer;
   gboolean keep_aspect;
-  gboolean redraw_border;
-  gboolean handle_events;
-  gboolean handle_expose;
-
+  
   gint brightness;
   gint contrast;
   gint hue;
   gint saturation;
   gboolean cb_changed;
 
-  /* size of incoming video, used as the size for XvImage */
-  guint video_width, video_height;
-
-  /* display sizes, used for clipping the image */
-  gint disp_x, disp_y;
-  gint disp_width, disp_height;
-
-  /* port attributes */
-  gboolean autopaint_colorkey;
-  gint colorkey;
-  
-  gboolean draw_borders;
-  
-  /* port features */
-  gboolean have_autopaint_colorkey;
-  gboolean have_colorkey;
-  gboolean have_double_buffer;
-  
-  /* stream metadata */
-  gchar *media_title;
-
-  /* target video rectangle */
-  GstVideoRectangle render_rect;
-  gboolean have_render_rect;
+  guint video_width, video_height;     /* size of incoming video;
+                                        * used as the size for XvImage */
 };
 
 struct _GstXvImageSinkClass {
