@@ -35,6 +35,8 @@
 #define WIDTH 240
 #define HEIGHT 120
 
+gboolean have_eos = FALSE;
+
 /* For ease of programming we use globals to keep refs for our floating
  * src and sink pads we create; otherwise we always have to do get_pad,
  * get_peer, and then remove references in every test function */
@@ -86,8 +88,7 @@ notgst_check_setup_src_pad2 (GstElement * element,
   fail_if (srcpad == NULL, "Could not create a srcpad");
   ASSERT_OBJECT_REFCOUNT (srcpad, "srcpad", 1);
 
-  if (!(sinkpad = gst_element_get_static_pad (element, sink_template_name)))
-    sinkpad = gst_element_get_request_pad (element, sink_template_name);
+  sinkpad = gst_element_get_pad (element, sink_template_name);
   fail_if (sinkpad == NULL, "Could not get sink pad from %s",
       GST_ELEMENT_NAME (element));
   ASSERT_OBJECT_REFCOUNT (sinkpad, "sinkpad", 2);
@@ -111,8 +112,7 @@ notgst_check_teardown_src_pad2 (GstElement * element,
     sink_template_name = "sink";
 
   /* clean up floating src pad */
-  if (!(sinkpad = gst_element_get_static_pad (element, sink_template_name)))
-    sinkpad = gst_element_get_request_pad (element, sink_template_name);
+  sinkpad = gst_element_get_pad (element, sink_template_name);
   ASSERT_OBJECT_REFCOUNT (sinkpad, "sinkpad", 2);
   srcpad = gst_pad_get_peer (sinkpad);
 
@@ -255,8 +255,6 @@ cleanup_textoverlay (GstElement * textoverlay)
 
   gst_element_set_state (textoverlay, GST_STATE_NULL);
   gst_element_get_state (textoverlay, NULL, NULL, GST_CLOCK_TIME_NONE);
-  gst_pad_set_active (myvideosrcpad, FALSE);
-  gst_pad_set_active (mysinkpad, FALSE);
   notgst_check_teardown_src_pad2 (textoverlay, "video_sink");
   if (mytextsrcpad) {
     notgst_check_teardown_src_pad2 (textoverlay, "text_sink");
@@ -354,7 +352,7 @@ GST_START_TEST (test_video_passthrough)
   /* pushing gives away one of the two references we have ... */
   fail_unless (gst_pad_push (myvideosrcpad, inbuffer) == GST_FLOW_OK);
 
-  /* should have been discarded as out-of-segment since it has no timestamp */
+  /* should have been discareded as out-of-segment since it has no timestamp */
   ASSERT_BUFFER_REFCOUNT (inbuffer, "inbuffer", 1);
   fail_unless_equals_int (g_list_length (buffers), 0);
 
@@ -569,8 +567,8 @@ GST_START_TEST (test_video_waits_for_text)
   fail_unless_equals_int (g_list_length (buffers), 3);
 
   /* ... and there should not be any text rendered on it */
-  fail_unless (buffer_is_all_black (GST_BUFFER_CAST (buffers->next->
-              next->data)));
+  fail_unless (buffer_is_all_black (GST_BUFFER_CAST (buffers->next->next->
+              data)));
 
   /* a fourth video buffer */
   inbuffer = gst_buffer_make_metadata_writable (inbuffer);
