@@ -24,7 +24,6 @@
 
 #include <gst/check/gstcheck.h>
 
-GList *buffers = NULL;
 gboolean have_eos = FALSE;
 
 /* For ease of programming we use globals to keep refs for our floating
@@ -119,7 +118,7 @@ cleanup_videorate (GstElement * videorate)
 GST_START_TEST (test_one)
 {
   GstElement *videorate;
-  GstBuffer *inbuffer, *outbuffer;
+  GstBuffer *inbuffer;
   GstCaps *caps;
 
   videorate = setup_videorate ();
@@ -260,7 +259,6 @@ GST_START_TEST (test_wrong_order_from_zero)
   GstElement *videorate;
   GstBuffer *first, *second, *third, *outbuffer;
   GstCaps *caps;
-  guint64 in, out, dropped, duplicated;
 
   videorate = setup_videorate ();
   fail_unless (gst_element_set_state (videorate,
@@ -350,7 +348,6 @@ GST_START_TEST (test_wrong_order)
   GstElement *videorate;
   GstBuffer *first, *second, *third, *fourth, *outbuffer;
   GstCaps *caps;
-  guint64 in, out, dropped, duplicated;
 
   videorate = setup_videorate ();
   fail_unless (gst_element_set_state (videorate,
@@ -459,7 +456,7 @@ GST_END_TEST;
 GST_START_TEST (test_no_framerate)
 {
   GstElement *videorate;
-  GstBuffer *inbuffer, *outbuffer;
+  GstBuffer *inbuffer;
   GstCaps *caps;
 
   videorate = setup_videorate ();
