@@ -24,8 +24,6 @@
 
 #include <gst/check/gstcheck.h>
 
-GList *buffers = NULL;
-
 /* For ease of programming we use globals to keep refs for our floating
  * src and sink pads we create; otherwise we always have to do get_pad,
  * get_peer, and then remove references in every test function */
@@ -92,7 +90,7 @@ cleanup_vorbisdec (GstElement * vorbisdec)
 GST_START_TEST (test_wrong_channels_identification_header)
 {
   GstElement *vorbisdec;
-  GstBuffer *inbuffer, *outbuffer;
+  GstBuffer *inbuffer;
   GstBus *bus;
   GstMessage *message;
 
@@ -133,7 +131,7 @@ GST_END_TEST;
 GST_START_TEST (test_empty_identification_header)
 {
   GstElement *vorbisdec;
-  GstBuffer *inbuffer, *outbuffer;
+  GstBuffer *inbuffer;
   GstBus *bus;
   GstMessage *message;
 
@@ -170,7 +168,7 @@ GST_END_TEST;
 GST_START_TEST (test_identification_header)
 {
   GstElement *vorbisdec;
-  GstBuffer *inbuffer, *outbuffer;
+  GstBuffer *inbuffer;
   GstBus *bus;
   GstMessage *message;
   GstTagList *tag_list;
