@@ -4,7 +4,7 @@
 #include <gst/gst.h>
 #include <string.h>
 
-GST_DEBUG_CATEGORY (seek_debug);
+GST_DEBUG_CATEGORY_STATIC (seek_debug);
 #define GST_CAT_DEFAULT (seek_debug)
 
 static GList *seekable_pads = NULL;
@@ -15,7 +15,9 @@ static gboolean accurate_seek = FALSE;
 static gboolean keyframe_seek = FALSE;
 static gboolean loop_seek = FALSE;
 static gboolean flush_seek = TRUE;
-static gboolean scrub = FALSE;
+static gboolean scrub = TRUE;
+static gboolean play_scrub = FALSE;
+static gdouble rate = 1.0;
 
 static GstElement *pipeline;
 static gint64 position;
@@ -813,6 +815,7 @@ make_playerbin_pipeline (const gchar * location)
   return player;
 }
 
+#ifndef GST_DISABLE_PARSE
 static GstElement *
 make_parselaunch_pipeline (const gchar * description)
 {
@@ -827,6 +830,7 @@ make_parselaunch_pipeline (const gchar * description)
 
   return pipeline;
 }
+#endif
 
 static gchar *
 format_value (GtkScale * scale, gdouble value)
@@ -1067,8 +1071,13 @@ do_seek (GtkWidget * widget)
   if (loop_seek)
     flags |= GST_SEEK_FLAG_SEGMENT;
 
-  s_event = gst_event_new_seek (1.0,
-      GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, real, GST_SEEK_TYPE_NONE, 0);
+  if (rate >= 0) {
+    s_event = gst_event_new_seek (rate,
+        GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, real, GST_SEEK_TYPE_SET, -1);
+  } else {
+    s_event = gst_event_new_seek (rate,
+        GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, 0, GST_SEEK_TYPE_SET, real);
+  }
 
   GST_DEBUG ("seek to %" GST_TIME_FORMAT, GST_TIME_ARGS (real));
 
@@ -1087,7 +1096,7 @@ static void
 seek_cb (GtkWidget * widget)
 {
   /* If the timer hasn't expired yet, then the pipeline is running */
-  if (scrub && seek_timeout_id != 0) {
+  if (play_scrub && seek_timeout_id != 0) {
     GST_DEBUG ("do scrub seek, PAUSED");
     gst_element_set_state (pipeline, GST_STATE_PAUSED);
   }
@@ -1095,7 +1104,7 @@ seek_cb (GtkWidget * widget)
   GST_DEBUG ("do seek");
   do_seek (widget);
 
-  if (scrub) {
+  if (play_scrub) {
     GST_DEBUG ("do scrub seek, PLAYING");
     gst_element_set_state (pipeline, GST_STATE_PLAYING);
 
@@ -1128,14 +1137,14 @@ start_seek (GtkWidget * widget, GdkEventButton * event, gpointer user_data)
   if (event->type != GDK_BUTTON_PRESS)
     return FALSE;
 
-  if (state == GST_STATE_PLAYING && flush_seek) {
+  set_update_scale (FALSE);
+
+  if (state == GST_STATE_PLAYING && flush_seek && scrub) {
     GST_DEBUG ("start scrub seek, PAUSE");
     gst_element_set_state (pipeline, GST_STATE_PAUSED);
   }
 
-  set_update_scale (FALSE);
-
-  if (changed_id == 0 && flush_seek) {
+  if (changed_id == 0 && flush_seek && scrub) {
     changed_id = gtk_signal_connect (GTK_OBJECT (hscale),
         "value_changed", G_CALLBACK (seek_cb), pipeline);
   }
@@ -1151,7 +1160,7 @@ stop_seek (GtkWidget * widget, gpointer user_data)
     changed_id = 0;
   }
 
-  if (!flush_seek) {
+  if (!flush_seek || !scrub) {
     GST_DEBUG ("do final seek");
     do_seek (widget);
   }
@@ -1166,7 +1175,6 @@ stop_seek (GtkWidget * widget, gpointer user_data)
       gst_element_set_state (pipeline, GST_STATE_PLAYING);
     }
   }
-  set_update_scale (TRUE);
 
   return FALSE;
 }
@@ -1182,7 +1190,6 @@ play_cb (GtkButton * button, gpointer data)
     if (ret == GST_STATE_CHANGE_FAILURE)
       goto failed;
 
-    set_update_scale (TRUE);
     state = GST_STATE_PLAYING;
   }
   return;
@@ -1204,7 +1211,6 @@ pause_cb (GtkButton * button, gpointer data)
     if (ret == GST_STATE_CHANGE_FAILURE)
       goto failed;
 
-    set_update_scale (FALSE);
     state = GST_STATE_PAUSED;
   }
   return;
@@ -1227,7 +1233,6 @@ stop_cb (GtkButton * button, gpointer data)
       goto failed;
 
     gtk_adjustment_set_value (adjustment, 0.0);
-    set_update_scale (FALSE);
 
     state = GST_STATE_READY;
   }
@@ -1269,6 +1274,48 @@ scrub_toggle_cb (GtkToggleButton * button, GstPipeline * pipeline)
   scrub = gtk_toggle_button_get_active (button);
 }
 
+static void
+play_scrub_toggle_cb (GtkToggleButton * button, GstPipeline * pipeline)
+{
+  play_scrub = gtk_toggle_button_get_active (button);
+}
+
+static void
+rate_spinbutton_changed_cb (GtkSpinButton * button, GstPipeline * pipeline)
+{
+  gboolean res = FALSE;
+  GstEvent *s_event;
+  GstSeekFlags flags;
+
+  rate = gtk_spin_button_get_value (button);
+
+  flags = 0;
+  if (flush_seek)
+    flags |= GST_SEEK_FLAG_FLUSH;
+  if (loop_seek)
+    flags |= GST_SEEK_FLAG_SEGMENT;
+
+  if (rate >= 0) {
+    s_event = gst_event_new_seek (rate,
+        GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, position,
+        GST_SEEK_TYPE_SET, -1);
+  } else {
+    s_event = gst_event_new_seek (rate,
+        GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, 0, GST_SEEK_TYPE_SET,
+        position);
+  }
+
+  res = send_event (s_event);
+
+  if (res) {
+    if (flush_seek) {
+      gst_pipeline_set_new_stream_time (GST_PIPELINE (pipeline), 0);
+      gst_element_get_state (GST_ELEMENT (pipeline), NULL, NULL, SEEK_TIMEOUT);
+    }
+  } else
+    g_print ("seek failed\n");
+}
+
 static void
 segment_done (GstBus * bus, GstMessage * message, GstPipeline * pipeline)
 {
@@ -1280,10 +1327,8 @@ segment_done (GstBus * bus, GstMessage * message, GstPipeline * pipeline)
   if (loop_seek)
     flags |= GST_SEEK_FLAG_SEGMENT;
 
-  event = gst_event_new_seek (1.0,
-      GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, 0, GST_SEEK_TYPE_NONE, 0);
-
-  GST_DEBUG ("segmeent seek to start");
+  event = gst_event_new_seek (rate,
+      GST_FORMAT_TIME, flags, GST_SEEK_TYPE_SET, 0, GST_SEEK_TYPE_SET, -1);
 
   res = send_event (event);
   if (!res) {
@@ -1311,6 +1356,27 @@ message_received (GstBus * bus, GstMessage * message, GstPipeline * pipeline)
   }
 }
 
+static void
+msg_state_changed (GstBus * bus, GstMessage * message, GstPipeline * pipeline)
+{
+  const GstStructure *s;
+
+  s = gst_message_get_structure (message);
+
+  /* We only care about state changed on the pipeline */
+  if (s && GST_MESSAGE_SRC (message) == GST_OBJECT_CAST (pipeline)) {
+    GstState old, new, pending;
+
+    gst_message_parse_state_changed (message, &old, &new, &pending);
+
+    /* When state of the pipeline changes to playing we start updating scale */
+    if (new == GST_STATE_PLAYING) {
+      set_update_scale (TRUE);
+    } else {
+      set_update_scale (FALSE);
+    }
+  }
+}
 
 typedef struct
 {
@@ -1335,7 +1401,9 @@ static Pipeline pipelines[] = {
   {"dv", make_dv_pipeline},
   {"mpeg1nothreads", make_mpegnt_pipeline},
   {"playerbin", make_playerbin_pipeline},
+#ifndef GST_DISABLE_PARSE
   {"parse-launch", make_parselaunch_pipeline},
+#endif
   {NULL, NULL},
 };
 
@@ -1357,9 +1425,12 @@ print_usage (int argc, char **argv)
 int
 main (int argc, char **argv)
 {
-  GtkWidget *window, *hbox, *vbox, *play_button, *pause_button, *stop_button;
+  GtkWidget *window, *hbox, *vbox, *flagtable;
+  GtkWidget *play_button, *pause_button, *stop_button;
   GtkWidget *accurate_checkbox, *key_checkbox, *loop_checkbox, *flush_checkbox;
-  GtkWidget *scrub_checkbox;
+  GtkWidget *scrub_checkbox, *play_scrub_checkbox, *rate_spinbutton;
+  GtkWidget *rate_label;
+  GtkTooltips *tips;
   GOptionEntry options[] = {
     {"stats", 's', 0, G_OPTION_ARG_NONE, &stats,
         "Show pad stats", NULL},
@@ -1373,7 +1444,7 @@ main (int argc, char **argv)
   GOptionContext *ctx;
   GError *err = NULL;
 
-  ctx = g_option_context_new ("seek");
+  ctx = g_option_context_new ("- test seeking in gsteamer");
   g_option_context_add_main_entries (ctx, options, NULL);
   g_option_context_add_group (ctx, gst_init_get_option_group ());
 
@@ -1402,21 +1473,51 @@ main (int argc, char **argv)
   g_assert (pipeline);
 
   /* initialize gui elements ... */
+  tips = gtk_tooltips_new ();
   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   hbox = gtk_hbox_new (FALSE, 0);
   vbox = gtk_vbox_new (FALSE, 0);
-  play_button = gtk_button_new_with_label ("play");
-  pause_button = gtk_button_new_with_label ("pause");
-  stop_button = gtk_button_new_with_label ("stop");
+  flagtable = gtk_table_new (4, 2, FALSE);
+  gtk_container_set_border_width (GTK_CONTAINER (vbox), 3);
 
+  /* media controls */
+  play_button = gtk_button_new_from_stock (GTK_STOCK_MEDIA_PLAY);
+  pause_button = gtk_button_new_from_stock (GTK_STOCK_MEDIA_PAUSE);
+  stop_button = gtk_button_new_from_stock (GTK_STOCK_MEDIA_STOP);
+
+  /* seek flags */
   accurate_checkbox = gtk_check_button_new_with_label ("Accurate Seek");
-  key_checkbox = gtk_check_button_new_with_label ("Key_unit Seek");
+  key_checkbox = gtk_check_button_new_with_label ("Key-unit Seek");
   loop_checkbox = gtk_check_button_new_with_label ("Loop");
   flush_checkbox = gtk_check_button_new_with_label ("Flush");
   scrub_checkbox = gtk_check_button_new_with_label ("Scrub");
+  play_scrub_checkbox = gtk_check_button_new_with_label ("Play Scrub");
+  rate_spinbutton = gtk_spin_button_new_with_range (-10, 10, 0.1);
+  rate_label = gtk_label_new ("Rate");
+
+  gtk_tooltips_set_tip (tips, accurate_checkbox,
+      "accurate position is requested, this might be considerably slower for some formats",
+      NULL);
+  gtk_tooltips_set_tip (tips, key_checkbox,
+      "seek to the nearest keyframe. This might be faster but less accurate",
+      NULL);
+  gtk_tooltips_set_tip (tips, loop_checkbox, "loop playback", NULL);
+  gtk_tooltips_set_tip (tips, flush_checkbox, "flush pipeline after seeking",
+      NULL);
+  gtk_tooltips_set_tip (tips, rate_spinbutton, "define the playback rate, "
+      "negative value trigger reverse playback", NULL);
+  /* FIXME: describe these */
+  gtk_tooltips_set_tip (tips, scrub_checkbox, "show images while seeking",
+      NULL);
+  gtk_tooltips_set_tip (tips, play_scrub_checkbox, "play video while seeking",
+      NULL);
 
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (flush_checkbox), TRUE);
+  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (scrub_checkbox), TRUE);
 
+  gtk_spin_button_set_value (GTK_SPIN_BUTTON (rate_spinbutton), rate);
+
+  /* seek bar */
   adjustment =
       GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.00, 100.0, 0.1, 1.0, 1.0));
   hscale = gtk_hscale_new (adjustment);
@@ -1437,11 +1538,18 @@ main (int argc, char **argv)
   gtk_box_pack_start (GTK_BOX (hbox), play_button, FALSE, FALSE, 2);
   gtk_box_pack_start (GTK_BOX (hbox), pause_button, FALSE, FALSE, 2);
   gtk_box_pack_start (GTK_BOX (hbox), stop_button, FALSE, FALSE, 2);
-  gtk_box_pack_start (GTK_BOX (hbox), accurate_checkbox, FALSE, FALSE, 2);
-  gtk_box_pack_start (GTK_BOX (hbox), key_checkbox, FALSE, FALSE, 2);
-  gtk_box_pack_start (GTK_BOX (hbox), loop_checkbox, FALSE, FALSE, 2);
-  gtk_box_pack_start (GTK_BOX (hbox), flush_checkbox, FALSE, FALSE, 2);
-  gtk_box_pack_start (GTK_BOX (hbox), scrub_checkbox, FALSE, FALSE, 2);
+  gtk_box_pack_start (GTK_BOX (hbox), flagtable, FALSE, FALSE, 2);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), accurate_checkbox, 0, 1, 0,
+      1);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), flush_checkbox, 1, 2, 0, 1);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), loop_checkbox, 2, 3, 0, 1);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), key_checkbox, 0, 1, 1, 2);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), scrub_checkbox, 1, 2, 1, 2);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), play_scrub_checkbox, 2, 3,
+      1, 2);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), rate_label, 3, 4, 0, 1);
+  gtk_table_attach_defaults (GTK_TABLE (flagtable), rate_spinbutton, 3, 4, 1,
+      2);
   gtk_box_pack_start (GTK_BOX (vbox), hscale, TRUE, TRUE, 2);
 
   /* connect things ... */
@@ -1461,6 +1569,10 @@ main (int argc, char **argv)
       G_CALLBACK (flush_toggle_cb), pipeline);
   g_signal_connect (G_OBJECT (scrub_checkbox), "toggled",
       G_CALLBACK (scrub_toggle_cb), pipeline);
+  g_signal_connect (G_OBJECT (play_scrub_checkbox), "toggled",
+      G_CALLBACK (play_scrub_toggle_cb), pipeline);
+  g_signal_connect (G_OBJECT (rate_spinbutton), "value_changed",
+      G_CALLBACK (rate_spinbutton_changed_cb), pipeline);
 
   g_signal_connect (G_OBJECT (window), "destroy", gtk_main_quit, NULL);
 
@@ -1492,6 +1604,8 @@ main (int argc, char **argv)
         pipeline);
     g_signal_connect (bus, "message::segment-done",
         (GCallback) message_received, pipeline);
+    g_signal_connect (bus, "message::state-changed",
+        (GCallback) msg_state_changed, pipeline);
     g_signal_connect (bus, "message::segment-done", (GCallback) segment_done,
         pipeline);
   }
